
/*
 *  Author:    Dave Rodgers  (a.k.a. trepan)
 *  Modified:  July 11, 2003
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "bmp2bzw.h"
#include "loadbmp.h"
#include "utils.h"


typedef struct {
  char height;
  char state;
} bit_t;

#define BITHEIGHT(map,x,y)	((map + (y*SizeX + x))->height)
#define BITSTATE(map,x,y)	((map + (y*SizeX + x))->state)


static int SizeX;
static int SizeY;
static int Maxval;

enum BitState {
  bit_off,
  bit_on,
  bit_boxed                     /* already covered */
};


/*
 * Function Prototypes
 */

static int find_maxval (u8 * image, int w, int h);

static int bitmerge (bit_t * bitmap, box_t ** boxes);
static int bitmerge_vert (bit_t * bitmap, box_t * r, int count, int len);
static int bitmerge_horiz (bit_t * bitmap, box_t * r, int count, int len);

static int comp_vert (const void *a, const void *b);
static int comp_horiz (const void *a, const void *b);
static void shift_boxes (int pos, box_t * r, int count);
static int boxmerge_vert (box_t * r, int count);
static int boxmerge_horiz (box_t * r, int count);

static void clear_bitmarks (bit_t * bitmap);
static void mark_bits (bit_t * bitmap, int c0, int c1, int r0, int r1);


/****************************************************************************/

int
bmp2bzw (const char *filename, box_t ** boxes, int *count,
         int *width, int *height, int levels, int invert)
{
  int x, y;
  u8 *image;
  bit_t *bitmap;


  if (load_bmp (filename, &SizeX, &SizeY, &image) != 0) {
    return -1;
  }

  *width = SizeX;
  *height = SizeY;

  bitmap = (bit_t *) malloc (sizeof (bit_t) * SizeX * SizeY);
  if (bitmap == NULL) {
    perror ("malloc");
    free (image);
    return -1;
  }

  Maxval = find_maxval (image, SizeX, SizeY);

  DEBUG7 ("Image  = %i X %i\n", SizeX, SizeY);
  DEBUG7 ("Levels = %i\n", levels);
  DEBUG7 ("Maxval = %i\n", Maxval);

  for (y = 0; y < SizeY; y++) {
    for (x = 0; x < SizeX; x++) {
      BITHEIGHT (bitmap, x, y) =
          ((*(image + y * SizeX + x) * levels) + (Maxval / 2)) / Maxval;

      if (invert)
        BITHEIGHT (bitmap, x, y) = levels - BITHEIGHT (bitmap, x, y);

      if (BITHEIGHT (bitmap, x, y) != 0)
        BITSTATE (bitmap, x, y) = bit_on;
      else
        BITSTATE (bitmap, x, y) = bit_off;
    }
  }

  free (image);

  *count = bitmerge (bitmap, boxes);    /* do the grind work */

  if (*count == -1) {
    free (bitmap);
    free (boxes);
    return -1;
  }

  free (bitmap);

  return 0;
}

/****************************************************************************/

static int
find_maxval (u8 * image, int w, int h)
{
  int x, y;
  int maxval = 0;

  for (x = 0; x < w; x++) {
    for (y = 0; y < h; y++) {
      if (*(image + (y * w + x)) > maxval)
        maxval = *(image + (y * w + x));
    }
  }

  return maxval;
}

/****************************************************************************/

static int
bitmerge (bit_t * bitmap, box_t ** boxes)
{
  int vcount;
  int hcount;
  box_t *vboxes;
  box_t *hboxes;

  vboxes = (box_t *) malloc (sizeof (box_t) * SizeX * SizeY);
  if (vboxes == NULL) {
    perror ("vboxes(malloc)");
    return -1;
  }

  hboxes = (box_t *) malloc (sizeof (box_t) * SizeX * SizeY);
  if (hboxes == NULL) {
    perror ("hboxes(malloc)");
    return -1;
  }


  vcount = bitmerge_vert (bitmap, vboxes, 0, 3);
  vcount = bitmerge_horiz (bitmap, vboxes, vcount, 2);
  vcount = bitmerge_vert (bitmap, vboxes, vcount, 1);
  vcount = boxmerge_horiz (vboxes, vcount);
  vcount = boxmerge_vert (vboxes, vcount);

  clear_bitmarks (bitmap);

  hcount = bitmerge_horiz (bitmap, hboxes, 0, 3);
  hcount = bitmerge_vert (bitmap, hboxes, hcount, 2);
  hcount = bitmerge_horiz (bitmap, hboxes, hcount, 1);
  hcount = boxmerge_horiz (hboxes, hcount);
  hcount = boxmerge_vert (hboxes, hcount);

  DEBUG8 ("bitmerge: vcount = %i\n", vcount);
  DEBUG8 ("bitmerge: hcount = %i\n", hcount);

  if (vcount > hcount) {
    free (vboxes);
    *boxes = (box_t *) realloc (hboxes, hcount * sizeof (box_t));
    if ((*boxes == NULL) && (hcount != 0)) {
      perror ("bitmerge(realloc)");
      free (hboxes);
      return -1;
    }
    return hcount;
  }
  else {
    free (hboxes);
    *boxes = (box_t *) realloc (vboxes, hcount * sizeof (box_t));
    if ((*boxes == NULL) && (vcount != 0)) {
      perror ("bitmerge(realloc)");
      free (vboxes);
      return -1;
    }
    return vcount;
  }
}

/****************************************************************************/

static int
bitmerge_vert (bit_t * bitmap, box_t * boxes, int count, int len)
{
  int row, col;
  int length;
  int sz;
  int bitsz;


  for (col = 0; col < SizeX; col++) {

    length = 0;
    sz = -1;

    DEBUG7 ("bg_vert: col %i\n", col);

    for (row = 0; row < SizeY; row++) {


      bitsz = BITHEIGHT (bitmap, col, row);

      if ((BITSTATE (bitmap, col, row) == bit_on) &&
          ((sz == -1) || (bitsz == sz))) {

        length++;

        sz = bitsz;
      }
      else if (length >= len) {
        boxes[count].x = col;
        boxes[count].y = row - length;
        boxes[count].sx = 1;
        boxes[count].sy = length;
        boxes[count].sz = sz;

        mark_bits (bitmap, col, col, row - length, row - 1);

        count++;

        if (BITSTATE (bitmap, col, row) == bit_on) {
          length = 1;
          sz = bitsz;
        }
        else {
          length = 0;
          sz = -1;
        }
      }
      else {
        if (BITSTATE (bitmap, col, row) == bit_on) {
          length = 1;
          sz = bitsz;
        }
        else {
          length = 0;
          sz = -1;
        }
      }

      if ((row == (SizeY - 1)) && (length >= len)) {
        boxes[count].x = col;
        boxes[count].y = row + 1 - length;
        boxes[count].sx = 1;
        boxes[count].sy = length;
        boxes[count].sz = sz;

        mark_bits (bitmap, col, col, SizeY - length, SizeY - 1);

        count++;
      }
    }
  }

  return count;
}

/****************************************************************************/

static int
bitmerge_horiz (bit_t * bitmap, box_t * boxes, int count, int len)
{
  int row, col;
  int length;
  int sz;
  int bitsz;


  for (row = 0; row < SizeY; row++) {

    DEBUG7 ("bg_horiz: row %i\n", row);

    length = 0;
    sz = -1;

    for (col = 0; col < SizeX; col++) {

      bitsz = BITHEIGHT (bitmap, col, row);

      if ((BITSTATE (bitmap, col, row) == bit_on) &&
          ((sz == -1) || (bitsz == sz))) {

        length++;

        sz = bitsz;
      }
      else if (length >= len) {
        boxes[count].x = col - length;
        boxes[count].y = row;
        boxes[count].sx = length;
        boxes[count].sy = 1;
        boxes[count].sz = sz;

        mark_bits (bitmap, col - length, col - 1, row, row);

        count++;

        if (BITSTATE (bitmap, col, row) == bit_on) {
          length = 1;
          sz = bitsz;
        }
        else {
          length = 0;
          sz = -1;
        }
      }
      else {
        if (BITSTATE (bitmap, col, row) == bit_on) {
          length = 1;
          sz = bitsz;
        }
        else {
          length = 0;
          sz = -1;
        }
      }

      if ((col == (SizeX - 1)) && (length >= len)) {
        boxes[count].x = col + 1 - length;
        boxes[count].y = row;
        boxes[count].sx = length;
        boxes[count].sy = 1;
        boxes[count].sz = sz;

        mark_bits (bitmap, SizeX - length, SizeX - 1, row, row);

        count++;
      }
    }
  }

  return count;
}

/****************************************************************************/

static int
comp_vert (const void *a, const void *b)
{
  if (((const box_t *) a)->y == ((const box_t *) b)->y)
    return 0;
  else if (((const box_t *) a)->y > ((const box_t *) b)->y)
    return +1;
  else
    return -1;
}

/****************************************************************************/

static int
comp_horiz (const void *a, const void *b)
{
  if (((const box_t *) a)->x == ((const box_t *) b)->x)
    return 0;
  else if (((const box_t *) a)->x > ((const box_t *) b)->x)
    return +1;
  else
    return -1;
}

/****************************************************************************/

static void
shift_boxes (int pos, box_t * r, int count)
{
  if (pos < (count - 1)) {
    memmove (&r[pos], &r[pos + 1], (count - pos - 1) * sizeof (box_t));
  }

  return;
}

/****************************************************************************/

static int
boxmerge_horiz (box_t * r, int count)
{
  int i, j;

  DEBUG8 ("boxmerge_horiz() COUNT = %i\n", count);

  /*
   * Merge horizontal strips 
   */

  qsort (r, (unsigned int) count, sizeof (box_t), comp_vert);

  for (i = 0; i < (count - 1); i++) {

    if ((get_debug () >= 9) && ((i % 100) == 0)) {
      printf ("bg_mergeh: %i -> %i\n", i, count);
    }

    for (j = (i + 1); j < count; j++) {

      DEBUG9 ("Horiz Comparing %i and %i:  "
              "(%i, %i, %i, %i, %i) vs (%i, %i, %i, %i, %i) ", i, j,
              r[i].x, r[i].y, r[i].sx, r[i].sy, r[i].sz,
              r[j].x, r[j].y, r[j].sx, r[j].sy, r[j].sz);

      if ((r[i].sz == r[j].sz) &&
          (r[i].x == r[j].x) && (r[i].sx == r[j].sx) &&
          ((r[i].y == (r[j].y + r[j].sy)) ||
           ((r[i].y + r[i].sy) == r[j].y))) {

        DEBUG9 ("match\n");

        if (r[i].y > r[j].y)
          r[i].y = r[j].y;

        r[i].sy = r[i].sy + r[j].sy;

        shift_boxes (j, r, count);

        j--;
        count--;
      }
      else {
        DEBUG9 ("no match\n");
      }
    }
  }

  return count;
}

/****************************************************************************/

static int
boxmerge_vert (box_t * r, int count)
{
  int i, j;

  DEBUG8 ("boxmerge_vert () COUNT = %i\n", count);

  /*
   * Merge vertical strips 
   */

  qsort (r, (unsigned int) count, sizeof (box_t), comp_horiz);

  for (i = 0; i < (count - 1); i++) {

    if ((get_debug () >= 9) && ((i % 100) == 0)) {
      printf ("bg_mergev: %i -> %i\n", i, count);
    }

    for (j = (i + 1); j < count; j++) {

      DEBUG9 ("Vert Comparing %i and %i:  "
              "(%i, %i, %i, %i, %i) vs (%i, %i, %i, %i, %i) ", i, j,
              r[i].x, r[i].y, r[i].sx, r[i].sy, r[i].sz,
              r[j].x, r[j].y, r[j].sx, r[j].sy, r[j].sz);

      if ((r[i].sz == r[j].sz) &&
          (r[i].y == r[j].y) && (r[i].sy == r[j].sy) &&
          ((r[i].x == (r[j].x + r[j].sx)) ||
           ((r[i].x + r[i].sx) == r[j].x))) {

        DEBUG9 ("match\n");

        if (r[i].x > r[j].x)
          r[i].x = r[j].x;

        r[i].sx = r[i].sx + r[j].sx;

        shift_boxes (j, r, count);

        j--;
        count--;
      }
      else {
        DEBUG9 ("no match\n");
      }
    }
  }

  return count;
}

/****************************************************************************/

static void
clear_bitmarks (bit_t * bitmap)
{
  int row, col;

  for (row = 0; row < SizeY; row++) {
    for (col = 0; col < SizeX; col++) {
      if (BITSTATE (bitmap, col, row) != bit_off) {
        BITSTATE (bitmap, col, row) = bit_on;
      }
    }
  }

  return;
}

/****************************************************************************/

static void
mark_bits (bit_t * bitmap, int c0, int c1, int r0, int r1)
{
  int row, col;

  for (row = r0; row <= r1; row++) {
    for (col = c0; col <= c1; col++) {
      BITSTATE (bitmap, col, row) = bit_boxed;
    }
  }

  return;
}

/****************************************************************************/
/****************************************************************************/
