
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "boxfont.h"
#include "utils.h"

#include "deffont.h"


/*
 * #define INVERTED_FONT 
 */


/*
 * PSF1 FONTS 
 */

#define PSF1_MAGIC0     (unsigned char)0x36
#define PSF1_MAGIC1     (unsigned char)0x04

#define PSF1_MODE512    (unsigned char)0x01
#define PSF1_MODEHASTAB (unsigned char)0x02
#define PSF1_MODEHASSEQ (unsigned char)0x04
#define PSF1_MAXMODE    (unsigned char)0x05

#define PSF1_SEPARATOR  0xFFFF
#define PSF1_STARTSEQ   0xFFFE

typedef struct psf1_header {
  unsigned char magic[2];       /* Magic number */
  unsigned char mode;           /* PSF font mode */
  unsigned char charsize;       /* Character size */
} psf1_header;


enum BitState {
  bit_off,
  bit_on,
  bit_boxed                     /* already covered */
};



/*
 * Function Prototypes
 */

static int create_glyph (FILE * file, int w, int h, glyph_t * g);

static int bitgrind (char *bitmap, int w, int h, rect_t ** rects);
static int bitgrind_vert
    (char *bitmap, int w, int h, rect_t * r, int count, int len);
static int bitgrind_horiz
    (char *bitmap, int w, int h, rect_t * r, int count, int len);
static int comp_vert (const void *a, const void *b);
static int comp_horiz (const void *a, const void *b);
static void shift_rects (int pos, rect_t * r, int count);
static int bitgrind_merge (rect_t * r, int count);

static void clear_bitmarks (char *bitmap, int w, int h);
static void mark_bits (char *bitmap, int w, int c0, int c1, int r0,
                       int r1);



/*
 * Static Variables
 */

static glyph_t *Glyphs = NULL;
static int GlyphCount = 0;

static int Bits = 0;
static int Rects = 0;

#ifdef INVERTED_FONT
static int Invert = 1;
#else
static int Invert = 0;
#endif


/****************************************************************************/

int
init_boxfont (const char *filename)
{
  int i;
  FILE *file;
  unsigned char buffer[4];
  struct psf1_header header1;

  if (filename == NULL) {
    Glyphs = DefGlyphs;
    GlyphCount = DEF_GLYPH_COUNT;
    return 0;
  }

  file = fopen (filename, "rb");    /* the 'b' is for Windows */
  if (file == NULL) {
    error ("init_boxfont(%s)\n", filename);
    return -1;
  }

  fread (buffer, 1, 4, file);

  if ((buffer[0] == PSF1_MAGIC0) && (buffer[1] == PSF1_MAGIC1)) {
    DEBUG1 ("BOXFONT: (%s) PSF1 type font file.\n", filename);
  }
  else {
    error ("Unknown font file type (%s)\n", filename);
    fclose (file);
    return -1;
  }

  header1.mode = buffer[2];
  header1.charsize = buffer[3];

  if ((header1.mode & PSF1_MODE512) == 0) {
    GlyphCount = 256;
    DEBUG3 ("BOXFONT: Glyphs = 256.\n");
  }
  else {
    GlyphCount = 512;
    DEBUG3 ("BOXFONT: Glyphs = 512.\n");
  }

  Glyphs = (glyph_t *) malloc (sizeof (glyph_t) * GlyphCount);
  if (Glyphs == NULL) {
    perror ("init_boxfont(malloc)");
    fclose (file);
    return -1;
  }

  Bits = 0;
  Rects = 0;

  for (i = 0; i < GlyphCount; i++) {
    if (create_glyph (file, 8, header1.charsize, &Glyphs[i]) == -1) {
      fclose (file);
      free (Glyphs);
      return -1;
    }
  }

  DEBUG8 ("BOXFONT: BitCount    = %i\n", Bits);
  DEBUG8 ("BOXFONT: RectCount   = %i\n", Rects);
  DEBUG8 ("BOXFONT: Compression = %.2f%%\n",
          100.0f * (double) (Bits - Rects) / (double) Bits);

  fclose (file);

  return 0;
}

/****************************************************************************/

int
boxfont_char (int c, glyph_t * g)
{
  if (c < GlyphCount) {
    g->rectcount = Glyphs[c].rectcount;
    g->rects = Glyphs[c].rects;
  }
  else {
    return -1;
  }

  return 0;
}

/****************************************************************************/

void
free_boxfont (void)
{
  int i;

  if (Glyphs == DefGlyphs) {
    Glyphs = NULL;
    GlyphCount = 0;
    return;
  }

  for (i = 0; i < GlyphCount; i++) {
    free (Glyphs[i].rects);
  }

  free (Glyphs);

  return;
}

/****************************************************************************/

static int
create_glyph (FILE * file, int width, int height, glyph_t * g)
{
  int row, col, tmp;
  int bitcount = 0;
  int rectcount = 0;
  unsigned char c;
  char *bitmap;
  rect_t *rects;

  bitmap = (char *)
      malloc (sizeof (char) * width * height);
  if (bitmap == NULL) {
    perror ("malloc");
    return -1;
  }

  for (row = 0; row < height; row++) {

    fread (&c, 1, 1, file);
    tmp = 1 << (width - 1);

    for (col = 0; col < width; col++) {

      /*
       * the bitmap starts at (0,0) in the bottom left hand
       * corner, and then sweeps across horizontally, and
       * scans upwards vertically.
       */
      if (((tmp & c) == 0) ^ (Invert)) {
        *(bitmap + ((height - row - 1) * width + col)) = bit_off;
      }
      else {
        *(bitmap + ((height - row - 1) * width + col)) = bit_on;
        bitcount++;
      }

      tmp = tmp >> 1;
    }
  }

  rectcount = bitgrind (bitmap, width, height, &rects);

  free (bitmap);

  if (rectcount == -1) {
    return -1;
  }

  g->rectcount = rectcount;
  g->rects = rects;

  Bits = Bits + bitcount;
  Rects = Rects + rectcount;


  return 0;
}

/****************************************************************************/

static int
bitgrind (char *bitmap, int w, int h, rect_t ** rects)
{
  int vcount;
  int hcount;
  rect_t *vrects;
  rect_t *hrects;

  vrects = (rect_t *) malloc (sizeof (rect_t) * w * h);
  if (vrects == NULL) {
    perror ("vrects(malloc)");
    return -1;
  }

  hrects = (rect_t *) malloc (sizeof (rect_t) * w * h);
  if (hrects == NULL) {
    free (vrects);
    perror ("hrects(malloc)");
    return -1;
  }


  vcount = bitgrind_vert (bitmap, w, h, vrects, 0, 3);
  vcount = bitgrind_horiz (bitmap, w, h, vrects, vcount, 2);
  vcount = bitgrind_vert (bitmap, w, h, vrects, vcount, 1);
  vcount = bitgrind_merge (vrects, vcount);

  clear_bitmarks (bitmap, w, h);

  hcount = bitgrind_horiz (bitmap, w, h, hrects, 0, 3);
  hcount = bitgrind_vert (bitmap, w, h, hrects, hcount, 2);
  hcount = bitgrind_horiz (bitmap, w, h, hrects, hcount, 1);
  hcount = bitgrind_merge (hrects, hcount);

  DEBUG9 ("bitgrind: vcount = %i\n", vcount);
  DEBUG9 ("bitgrind: hcount = %i\n", hcount);

  if (vcount > hcount) {
    free (vrects);
    *rects = (rect_t *) realloc (hrects, hcount * sizeof (rect_t));
    if ((*rects == NULL) && (hcount != 0)) {
      perror ("bitgrind(realloc)");
      return -1;
    }
    return hcount;
  }
  else {
    free (hrects);
    *rects = vrects;
    *rects = (rect_t *) realloc (vrects, hcount * sizeof (rect_t));
    if ((*rects == NULL) && (vcount != 0)) {
      perror ("bitgrind(realloc)");
      return -1;
    }
    return vcount;
  }
}

/****************************************************************************/

/*
 * Kinda like a run-length encoding
 */
static int
bitgrind_vert (char *bitmap, int w, int h, rect_t * rects, int count,
               int len)
{
  int row, col;
  int length;


  for (col = 0; col < w; col++) {

    length = 0;

    for (row = 0; row < h; row++) {
      if (*(bitmap + (row * w + col)) == bit_on) {
        length++;

        if ((row == (h - 1)) && (length >= len)) {
          rects[count].x = col;
          rects[count].y = row + 1 - length;
          rects[count].width = 1;
          rects[count].height = length;

          mark_bits (bitmap, w, col, col, h - length, h - 1);

          length = 0;
          count++;
        }
      }
      else if (length >= len) {
        rects[count].x = col;
        rects[count].y = row - length;
        rects[count].width = 1;
        rects[count].height = length;

        mark_bits (bitmap, w, col, col, row - length, row - 1);

        length = 0;
        count++;
      }
      else {
        length = 0;
      }
    }
  }

  return count;
}

/****************************************************************************/

/*
 * Kinda like a run-length encoding
 */
static int
bitgrind_horiz (char *bitmap, int w, int h, rect_t * rects, int count,
                int len)
{
  int row, col;
  int length;


  for (row = 0; row < h; row++) {

    length = 0;

    for (col = 0; col < w; col++) {
      if (*(bitmap + (row * w + col)) == bit_on) {
        length++;

        if ((col == (w - 1)) && (length >= len)) {
          rects[count].x = col + 1 - length;
          rects[count].y = row;
          rects[count].width = length;
          rects[count].height = 1;

          mark_bits (bitmap, w, w - length, w - 1, row, row);

          length = 0;
          count++;
        }
      }
      else if (length >= len) {
        rects[count].x = col - length;
        rects[count].y = row;
        rects[count].width = length;
        rects[count].height = 1;

        mark_bits (bitmap, w, col - length, col - 1, row, row);

        length = 0;
        count++;
      }
      else {
        length = 0;
      }
    }
  }

  return count;
}

/****************************************************************************/

static int
comp_vert (const void *a, const void *b)
{
  if (((const rect_t *) a)->y == ((const rect_t *) b)->y)
    return 0;
  else if (((const rect_t *) a)->y > ((const rect_t *) b)->y)
    return +1;
  else
    return -1;
}

/****************************************************************************/

static int
comp_horiz (const void *a, const void *b)
{
  if (((const rect_t *) a)->x == ((const rect_t *) b)->x)
    return 0;
  else if (((const rect_t *) a)->x > ((const rect_t *) b)->x)
    return +1;
  else
    return -1;
}

/****************************************************************************/

static void
shift_rects (int pos, rect_t * r, int count)
{
  if (pos < (count - 1))
    memmove (&r[pos], &r[pos + 1], (count - pos - 1) * sizeof (rect_t));
  return;
}

/****************************************************************************/

static int
bitgrind_merge (rect_t * r, int count)
{
  int i, j;

  /*
   * Merge horizontal strips 
   */

  qsort (r, (unsigned int) count, sizeof (rect_t), comp_horiz);

  for (i = 0; i < (count - 1); i++) {
    for (j = (i + 1); j < count; j++) {

      if ((r[i].x == r[j].x) && (r[i].width == r[j].width) &&
          ((r[i].y == (r[j].y + r[j].height)) ||
           ((r[i].y + r[i].height) == r[j].y))) {

        if (r[i].y > r[j].y)
          r[i].y = r[j].y;

        r[i].height = r[i].height + r[j].height;

        shift_rects (j, r, count);

        j--;
        count--;
      }
    }
  }

  /*
   * Merge vertical strips 
   */

  qsort (r, (unsigned int) count, sizeof (rect_t), comp_vert);

  for (i = 0; i < (count - 1); i++) {
    for (j = (i + 1); j < count; j++) {

      if ((r[i].y == r[j].y) && (r[i].height == r[j].height) &&
          ((r[i].x == (r[j].x + r[j].width)) ||
           ((r[i].x + r[i].width) == r[j].x))) {

        if (r[i].x > r[j].x)
          r[i].x = r[j].x;

        r[i].width = r[i].width + r[j].width;

        shift_rects (j, r, count);

        j--;
        count--;
      }
    }
  }

  return count;
}

/****************************************************************************/

static void
clear_bitmarks (char *bitmap, int w, int h)
{
  int row, col;

  for (row = 0; row < h; row++) {
    for (col = 0; col < w; col++) {
      if (*(bitmap + (row * w + col)) != bit_off) {
        *(bitmap + (row * w + col)) = bit_on;
      }
    }
  }

  return;
}

/****************************************************************************/

static void
mark_bits (char *bitmap, int w, int c0, int c1, int r0, int r1)
{
  int row, col;

  for (row = r0; row <= r1; row++) {
    for (col = c0; col <= c1; col++) {
      *(bitmap + (row * w + col)) = bit_boxed;
    }
  }

  return;
}

/****************************************************************************/
