/*
 *  B Z M A P P E R . C
 *
 *  Author:     Dave R (trepan)
 *  Date:       Feb 2003
 *
 *
 *  An enhanced parser for generating 3D world files for the
 *  BZFlag game. It is backwards compatible with the standard
 *  BZFlag world file format, but adds many improvements. See
 *  the "world/tutorial.bzg' for more information.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#include "bzmapper.h"
#include "keywords.h"
#include "escan.h"
#include "flowctrl.h"
#include "intersct.h"
#include "linebuf.h"
#include "boxfont.h"
#include "bmp2bzw.h"

#include "utils.h"

#ifdef __cplusplus
#  include "global.h"
#  warning "using global.h"
#else
#  include "cglobal.h"
#endif


/***********
 * Defines *
 ***********/

#define WORLD_NAME          "*WorldDef*"
#define SHADOW_NAME         "*ShadowWorld*"

#define MIN_DELAY           (0.1)   // FIXME - wrong place for these
#define DEF_DELAY           (10.0)
#define DEF_INITDELAY       (10.0)

#define RANDOM_LINK         (-1)

#define DEF_PREFIX          '/' /* for all default names */
#define DELIMITER           ':' /* for seperating groups */
#define MATCH_MULTI         '*' /* matches any number of characters */
#define MATCH_SINGLE        '?' /* matches a single character */

#define DEF_BOX_CHAR        'b' /* for default box names */
#define DEF_PYR_CHAR        'p' /* for default pyramid names */
#define DEF_TETRA_CHAR      'f' /* for default tetra (four) names */
#define DEF_BASE_CHAR       'B' /* for default base names */
#define DEF_TELE_CHAR       't' /* for default teleporter names */
#define DEF_LINK_CHAR       'l' /* for default link names */
#define DEF_WEAPON_CHAR     'w' /* for default weapon names */
#define DEF_ZONE_CHAR       'z' /* for default zone names */
#define DEF_ARC_CHAR        'A' /* for default arc names */
#define DEF_TEXT_CHAR       'T' /* for default text names */
#define DEF_HOLE_CHAR       'H' /* for default hole names */
#define DEF_IMAGE_CHAR      'I' /* for default image names */
#define DEF_PATHS_CHAR      'P' /* for default paths names */
#define DEF_STEPS_CHAR      'S' /* for default steps names */

#define GAP_STRING          "g" /* for the end of gap box names */

#define DEF_IGNORE_CHAR     '<' /* ignore the name tag */

#define LINK_FRONT          'F' /* both cases checked */
#define LINK_BACK           'B' /* both cases checked */

#define PARSE_EOF           (2) /* End of File */
#define PARSE_END           (1) /* Finished with an END */
#define PARSE_OK            (0) /* Normal command execution */
#define PARSE_ERROR         (-1)    /* Something has gone terribly wrong */


#define MAXZ_TYPE_GROUND    (+0.0)
#define MAXZ_TYPE_BOX       (+1.0)
#define MAXZ_TYPE_PYR       (+2.0)
#define MAXZ_TYPE_BASE      (+3.0)
#define MAXZ_TYPE_TELE      (+4.0)

#define MAXZ_SUPPORT_SLOPE      (-1.0)
#define MAXZ_SUPPORT_PARTIAL    (+1.0)
#define MAXZ_SUPPORT_CENTER     (+2.0)
#define MAXZ_SUPPORT_FULL       (+3.0)

#define MAXZ_INVALID        (-9.0)

#define COPY_VECTOR(dest, src)  \
 dest[0] = src[0];              \
 dest[1] = src[1];              \
 dest[2] = src[2]


/********************
 * Type Definitions *
 ********************/

typedef struct {
  char *name;
  int objs_count;
  struct _object *objs;
  int group_count;
  int box_count;
  int pyr_count;
  int tetra_count;
  int base_count;
  int tele_count;
  int link_count;
  int weapon_count;
  int zone_count;
  int arc_count;
  int text_count;
  int image_count;
  int paths_count;
  int steps_count;
  int hole_count;
} group_def;


typedef struct {
  vec3 pos;
  vec3 size;
} hole_def;

typedef hole_def listbox_def;

typedef struct {
  double x;
  double y;
} point_def;

typedef double TransformMatrix[4][4];  /* [row][col], opengl style */


typedef struct _object {
  unsigned char type;

  char *name;
  int id;                       // unique ID

  int deleted;                  // object will not export, clip, etc...

  vec3 pos;                     // basics
  vec3 size;
  double rotation;
  double tilt;                  // text and images
  int flipz;
  int drivethrough;
  int shootthrough;

  int grpdef;                   // group definition number

  int to;                       // links
  int from;                     // links
  int to_gate;                  // links
  int from_gate;                // links
  char *to_name;                // links and intersects (targets)
  char *from_name;              // links and intersects (clippers)

  int color;                    // bases
  double border;                // teles
  
  int vertex_count;             // tetras
  vec3 vertices[4];             // tetras
  double planes[4][4];          // tetras
  unsigned char visible;        // tetras
  unsigned char colored;        // tetras
  unsigned char colors[4][4];
  
  

  double initdelay;             // weapons
  double *delays;               // weapons
  int delay_count;              // weapons

  int teams;                    // zones
  int safety;                   // zones

  double angle;                 // arcs
  int segments;                 // arcs, paths, and steps
  int closedloop;               // paths and steps
  int convert;                  // modify
  char *text;                   // depends on the object type
  // text - the message
  // image - filename
  // weapon - type
  // zones and modify - flags list

  hole_def *holes;              // holes in boxes
  int holes_count;

  point_def *points;            // holes in paths and steps
  int points_count;

  int srcline;                  // for error messages
  linebuf_fd srcfile;

} object;


/*****************************
 * Local Function Prototypes *
 *****************************/

static void free_all (void);

static int setup_static_vars (void);
static int setup_expr_vars (void);
static int setup_expr_funcs (void);

static int parse_command (int grpdef);
static int parse_define_recurse (int grpdef, const char *c);
static int parse_name (int grpdef, object * obj, const char *c);
static int parse_group_name (int grpdef, object * obj, const char *c);
static int parse_text_name (int grpdef, object * obj, const char *c);
static int parse_image_name (int grpdef, object * obj, const char *c);
static int parse_raw_name (object * obj, const char *c);
static int parse_intersect_name (object * obj, const char *c);
static int parse_object_name (int grpdef, object * obj, const char *c);
static int parse_object (int grpdef, object * group);
static int parse_seed (const char *c);
static int parse_option (const char *c);
static int parse_base_color (const char *c);
static int parse_delays (const char *c, object * obj);
static int include_font (const char *c);
static int post_parse (object * obj);
static int validate_arc (object * arc);
static int validate_tetra (object * tetra);
static int validate_points (object * obj);

static int invalid_name_chars (const char *name);
static int invalid_define_name (const char *name);
static int invalid_line_end (const char *c);
static int find_dup_name (int grpdef, const char *name);

static int generate_object (object * obj);
static int generate_core_object (object * obj);
static int generate_link_name (char **name, int *num, int *gate, int from);
static int generate_group (object * group);
static int generate_boxes (object * box);
static int generate_arcs (object * obj);
static int generate_text (object * obj);
static int generate_image (object * obj);
static int generate_path (object * obj, int *indices);
static int generate_paths_steps (object * obj);
static int generate_modify (object * obj);
static int generate_delete (object * obj);
static int generate_undelete (object * obj);
static int generate_intersect (object * obj);
static int generate_hole (object * obj);
static int match_multi (const char **string, const char **objname);
static int match_object_name (const char *string, const char *objname);
static void calc_position (vec3 gpos, vec3 gsize, double grot, double tilt,
                           vec3 ipos, vec3 opos);
static void calc_rotation (const vec3 size, const double objrot,
                           const double staterot, double *outrot);

static int matrix_pop (void);
static int matrix_push (void);
static void matrix_ident (void);
static void matrix_shift (vec3 pos);
static void matrix_scale (vec3 size);
static void matrix_rotate (double angle, vec3 normal);
static void matrix_mult (vec3 output, const vec3 input);

static inline int core_object (object * obj);
static inline int solid_core_object (object * obj);
static inline int composite_object (object * obj);
static int last_object_solid (void);
static int intersect_objs (object * obj1, object * obj2);
static double clip_last (parameter * params);
static double clip_against (parameter * params);
static double shiftinc_last (parameter * params);
static double shiftinc_against (parameter * params);
static double ztop (parameter * params);
static double ztop_against (parameter * params);
static double ztop_type (parameter * params);
static double ztop_support (parameter * params);

static int resolve_links (void);
static int resolve_link (char *name, int *num);
static int pre_order_link (void);
static int post_order_link (void);
static int link_pos_to_id (int *linknum);
static int link_id_to_pos (int *linknum);

static int remove_pyr_overlaps (void);
static int remove_box_overlaps (void);
static int merge_boxes (void);
static int merge_boxes2 (int direction);
static int positize_boxes (void);

static int calc_statistics (void);

static int world_order (const void *a, const void *b);
static int export_world (int grpdef);
static int export_adjust (object * obj);

static int append_name (const char *name);
static int remove_name (const char *name);

static int add_group_def (const char *name);
static int find_group_def (const char *name);

static int add_object (int grpdef, object * obj);
static void init_object (object * obj, int type);
static void free_object (object * obj);
static void nuke_objects (int last_id);
static void nuke_shadow (void);

static int add_hole (object * obj, hole_def * hole);
static int parse_hole (hole_def * hole);

static int add_point (object * obj, point_def * point);

static int append_string (char **head, const char *tail);

static void print_dbgobj (object * obj);
static void print_hole (hole_def * hole, int count);
static void print_point (point_def * point, int count);
static void print_objects (void);
static void print_tele_names (void);
static void print_object_names (void);


/*******************
 * Local variables *
 *******************/

static const vec3 Ones = { 1.0, 1.0, 1.0 };
static const vec3 Zeroes = { 0.0, 0.0, 0.0 };

static int WorldDef = -1;
static int ShadowDef = -1;

static int GroupDefsCount = 0;
static group_def *GroupDefs = NULL;

static char *ConfigOptions = NULL;

// static group_list GroupDefs = {0, 64, NULL};

static char GroupDepthName[LINE_LENGTH];
static int LastObjectNumber = -1;

static int FoundWorld = 0;
static double WorldDimension = WorldSize / 2.0;
static double FlagHeight = FlagAltitude;

static double TopType = MAXZ_INVALID;
static double TopSupport = MAXZ_INVALID;

static int MatrixCount = 0;
static TransformMatrix* MatrixStack = NULL;
static TransformMatrix TheMatrix = {
  { 1.0, 0.0, 0.0, 0.0 },
  { 0.0, 1.0, 0.0, 0.0 },
  { 0.0, 0.0, 1.0, 0.0 },
  { 0.0, 0.0, 0.0, 1.0 }
};

static vec3 Position = { 0.0, 0.0, 0.0 };
static vec3 Size = { 1.0, 1.0, 1.0 };
static double Rotation = 0.0;
static double Tilt = 0.0;
static int DriveThrough = 0;
static int ShootThrough = 0;

static int InSizeStyle = BZMAPPER_NEGSIZE;
static int OutSizeStyle = BZMAPPER_FLIPZ;
static int OrderZ = BZMAPPER_ASCENDING;
static int IgnoreName = 1;
static int Overlaps = 0;
static int MergeBoxes = 0;
static int ConvertOld = 0;

static int StatsBoxes = -1;
static int StatsPyrs = -1;
static int StatsTetras = -1;
static int StatsBases = -1;
static int StatsTeles = -1;
static int StatsLinks = -1;
static int StatsWeapons = -1;
static int StatsZones = -1;

const char Delimiter[2] = { DELIMITER, '\0' };
const char DefPrefix[2] = { DEF_PREFIX, '\0' };

static const char *TeamNames[] = {
  "ROGUE",                      // 0
  "RED",                        // 1
  "GREEN",                      // 2
  "BLUE",                       // 3
  "PURPLE"                      // 4
};
#define TEAM_NAMES  (int)(sizeof(TeamNames)/sizeof(*TeamNames))


/******************************************************************************/
/******************************************************************************/

static int
parse_command (int grpdef)
{
  int keyword;
  char line[LINE_LENGTH];
  const char *c;
  object obj;

  keyword = get_flowed_line (line, LINE_LENGTH, &c);

  if (keyword == LINEBUF_EOF) {
    if (grpdef == WorldDef) {
      return PARSE_EOF;
    }
    else {
      error ("missing end\n");
      return PARSE_ERROR;
    }
  }
  else if (keyword == LINEBUF_ERROR) {
    return PARSE_ERROR;
  }
  else if (keyword == end_key) {
    return PARSE_END;
  }
  else if (keyword == defgroup_key) {
    return parse_define_recurse (grpdef, c);
  }
  else if (keyword == loadfont_key) {
    return include_font (c);
  }
  else if (keyword == seed_key) {
    return parse_seed (c);
  }
  else if (keyword == options_key) {
    return parse_option (c);
  }
  else if (keyword == flipzstyle_key) {
    InSizeStyle = BZMAPPER_FLIPZ;
    return PARSE_OK;
  }
  else if (keyword == world_key) {

    if (FoundWorld != 0)
      error ("there is only one world\n");  // not really an error
    else
      FoundWorld = 1;

    obj.type = world_key;

    return parse_object (grpdef, &obj);
  }
  else if (keyword_flow_control (keyword)) {
    error ("internal error, control leak (%s)\n",
           keyword_string (keyword));
    return PARSE_ERROR;
  }
  else if (!keyword_top_level (keyword)) {
    error ("misplaced keyword: %s (%i)\n", keyword_string (keyword),
           keyword);
    return PARSE_ERROR;
  }


  // initialize the object

  init_object (&obj, ((unsigned char) keyword));
  obj.srcline = current_line_number ();
  obj.srcfile = current_file ();

  if (parse_name (grpdef, &obj, c) != PARSE_OK) {
    free (&obj);
    return PARSE_ERROR;
  }

  if (keyword != group_key) {
    DEBUG2 ("  adding to (%s): %s\n", GroupDefs[grpdef].name,
            keyword_string (keyword));
  }


  // parse the object parameters

  if ((keyword != delete_key) && (keyword != undelete_key) &&
      (keyword != intersect_key)) {
    if (parse_object (grpdef, &obj) != 0) {
      free_object (&obj);
      return PARSE_ERROR;
    }
  }


  // Post parsing checks

  if (post_parse (&obj) != PARSE_OK) {
    free_object (&obj);
    return PARSE_ERROR;
  }


  /*
   * Add the object to the group definition, or to the list
   * of instantiated objects and the Shadow definition if we
   * are adding to the World definition.
   */

  if (grpdef == WorldDef) {

    LastObjectNumber = GroupDefs[grpdef].objs_count;

    if (generate_object (&obj) != 0) {
      free_object (&obj);
      return PARSE_ERROR;
    }
    if (add_object (ShadowDef, &obj) != 0) {
      free_object (&obj);
      return PARSE_ERROR;
    }
  }
  else {
    if (add_object (grpdef, &obj) != 0) {
      free_object (&obj);
      return PARSE_ERROR;
    }
  }

  return PARSE_OK;
}

/******************************************************************************/

static int
parse_define_recurse (int grpdef, const char *c)
{
  int i;
  int tmp;
  char name[LINE_LENGTH];
  int newdef;

  // Make sure that groupdefs only occur in the World group.

  if (grpdef != WorldDef) {
    error ("definition within definition\n");
    return -1;
  }

  if (escan (c, "sw", name) != 1) {
    error ("missing group definition name\n");
    return -1;
  }

  if (invalid_define_name (name)) {
    return -1;
  }

  for (i = 0; i < GroupDefsCount; i++) {
    if (strcmp (name, GroupDefs[i].name) == 0) {
      error ("duplicate group definition name: %s\n", name);
      return -1;
    }
  }

  DEBUG1 ("creating group definition: (%s)\n", name);

  newdef = add_group_def (name);

  if (newdef == -1)
    return -1;

  while ((tmp = parse_command (newdef)) == PARSE_OK) {
    // add until the "end" for the group definition has been found
  }

  if (tmp == PARSE_END)
    return PARSE_OK;            // found an associated "end"
  else
    return PARSE_ERROR;
}

/******************************************************************************/

static int
parse_name (int grpdef, object * obj, const char *c)
{
  int keyword;

  keyword = obj->type;

  if (keyword == group_key) {
    if (parse_group_name (grpdef, obj, c) != 0) {
      return PARSE_ERROR;
    }
  }
  else if (keyword == text_key) {
    if (parse_text_name (grpdef, obj, c) != 0) {
      return PARSE_ERROR;
    }
  }
  else if (keyword == image_key) {
    if (parse_image_name (grpdef, obj, c) != 0) {
      return PARSE_ERROR;
    }
  }
  else if ((keyword == modify_key) ||
           (keyword == delete_key) || (keyword == undelete_key)) {
    if (parse_raw_name (obj, c) != 0) {
      return PARSE_ERROR;
    }
  }
  else if (keyword == intersect_key) {
    if (parse_intersect_name (obj, c) != 0) {
      return PARSE_ERROR;
    }
  }
  else if (parse_object_name (grpdef, obj, c) != 0) {
    // boxes, pyramids, bases, teles, links, arcs, paths, and steps
    return PARSE_ERROR;
  }

  return PARSE_OK;
}

/******************************************************************************/

static int
parse_group_name (int grpdef, object * obj, const char *c)
{
  int tmp;
  int grpn;
  char group_name[LINE_LENGTH];
  char grpdef_name[LINE_LENGTH];


  tmp = escan (c, "ssw", grpdef_name, group_name);

  if (tmp < 1) {
    error ("groups require a group definition name\n");
    return -1;
  }
  else if (tmp == 1) {
    int i;
    int count = 0;
    int wrkdef;

    grpn = find_group_def (grpdef_name);

    if (grpdef == WorldDef) {
      wrkdef = ShadowDef;
    }
    else {
      wrkdef = grpdef;
    }

    for (i = 0; i < GroupDefs[wrkdef].objs_count; i++) {
      if (GroupDefs[wrkdef].objs[i].grpdef == grpn)
        count++;
    }

    sprintf (group_name, "%c%s%c%i", DEF_PREFIX, grpdef_name, DEF_PREFIX,
             count);

    obj->name = strdup (group_name);
    if (obj->name == NULL) {
      perror ("parse_object_name(strdup)");
      return -1;
    }
  }
  else {
    if (invalid_name_chars (group_name)) {
      error ("bad group name: %s\n", group_name);
      return -1;
    }

    obj->name = strdup (group_name);
    if (obj->name == NULL) {
      perror ("parse_object_name(strdup)");
      return -1;
    }

    grpn = find_group_def (grpdef_name);
  }

  if (grpn == -1) {
    error ("unknown group definition: %s\n", grpdef_name);
    return -1;
  }
  else if (grpn == grpdef) {    // avoid group recursion
    error ("recursive grouping: group %s\n", GroupDefs[grpdef].name);
    return -1;
  }
  else {
    obj->grpdef = grpn;
  }

  if (find_dup_name (grpdef, obj->name) != 0) {
    error ("duplicate name tag: %s\n", obj->name);
    return -1;
  }

  DEBUG2 ("  adding to (%s): groupdef (%s) tag [%s]\n",
          GroupDefs[grpdef].name, GroupDefs[obj->grpdef].name, obj->name);

  return 0;
}

/******************************************************************************/

static int
parse_text_name (int grpdef, object * obj, const char *c)
{
  int tmp;
  int wrkdef;
  char text[LINE_LENGTH];
  char name[LINE_LENGTH];

  if (grpdef == WorldDef) {
    wrkdef = ShadowDef;
  }
  else {
    wrkdef = grpdef;
  }

  tmp = escan (c, "Qsw", text, name);

  if (tmp < 1) {
    error ("text: expecting \"text\"\n");
    return -1;
  }

  obj->text = strdup (text);
  if (obj->text == NULL) {
    perror ("parse_text_name(strdup)");
    return -1;
  }

  if (tmp == 1) {
    sprintf (name, "%c%c%i", DEF_PREFIX, DEF_TEXT_CHAR,
             GroupDefs[wrkdef].text_count);
  }
  else {
    if (find_dup_name (grpdef, name) != 0) {
      error ("duplicate name tag: %s\n", name);
      return -1;
    }
    else if (invalid_name_chars (name)) {
      error ("bad text name: %s\n", name);
      return -1;
    }
  }

  obj->name = strdup (name);
  if (obj->name == NULL) {
    perror ("parse_text_name(strdup)");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
parse_image_name (int grpdef, object * obj, const char *c)
{
  int tmp;
  int wrkdef;
  char image[LINE_LENGTH];
  char name[LINE_LENGTH];

  if (grpdef == WorldDef) {
    wrkdef = ShadowDef;
  }
  else {
    wrkdef = grpdef;
  }

  tmp = escan (c, "Qsw", image, name);

  if (tmp < 1) {
    error ("image: expecting \"filename \"\n");
    return -1;
  }

  obj->text = strdup (image);
  if (obj->text == NULL) {
    perror ("parse_image_name(strdup)");
    return -1;
  }

  if (tmp == 1) {
    sprintf (name, "%c%c%i", DEF_PREFIX, DEF_IMAGE_CHAR,
             GroupDefs[wrkdef].image_count);
  }
  else {
    if (find_dup_name (grpdef, name) != 0) {
      error ("duplicate name tag: %s\n", name);
      return -1;
    }
    else if (invalid_name_chars (name)) {
      error ("bad text name: %s\n", name);
      return -1;
    }
  }

  obj->name = strdup (name);
  if (obj->name == NULL) {
    perror ("parse_image_name(strdup)");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
parse_raw_name (object * obj, const char *c)
{
  char name[LINE_LENGTH];

  if (escan (c, "sw", name) != 1) {
    error ("missing %s name\n", keyword_string (obj->type));
    return -1;
  }

  obj->name = strdup (name);
  if (obj->name == NULL) {
    perror ("strdup");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
parse_intersect_name (object * obj, const char *c)
{
  char targets[LINE_LENGTH];
  char against[LINE_LENGTH];
  char clippers[LINE_LENGTH];

  if (escan (c, "sssw", targets, against, clippers) != 3) {
    error ("missing clipper names\n");
    return -1;
  }

  if (strcmp (against, "against") != 0) {
    error ("intersect objects need the word \"against\".\n");
    return -1;
  }

  obj->to_name = strdup (targets);
  obj->from_name = strdup (clippers);
  if ((obj->to_name == NULL) || (obj->from_name == NULL)) {
    free (obj->to_name);
    free (obj->from_name);
    perror ("strdup");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
parse_object_name (int grpdef, object * obj, const char *c)
{
  int tmp;
  int wrkdef;
  char name[LINE_LENGTH];

  if (grpdef == WorldDef) {
    wrkdef = ShadowDef;
  }
  else {
    wrkdef = grpdef;
  }

  tmp = escan (c, "sw", name);

  if (tmp < 0) {                // invalid line ending
    error ("invalid object name (%s)\n", c);
    return -1;
  }
  else if (tmp == 0) {
    switch (obj->type) {
      case box_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_BOX_CHAR,
                 GroupDefs[wrkdef].box_count);
        break;
      case pyr_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_PYR_CHAR,
                 GroupDefs[wrkdef].pyr_count);
        break;
      case tetra_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_TETRA_CHAR,
                 GroupDefs[wrkdef].tetra_count);
        break;
      case base_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_BASE_CHAR,
                 GroupDefs[wrkdef].base_count);
        break;
      case tele_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_TELE_CHAR,
                 GroupDefs[wrkdef].tele_count);
        break;
      case link_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_LINK_CHAR,
                 GroupDefs[wrkdef].link_count);
        break;
      case weapon_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_WEAPON_CHAR,
                 GroupDefs[wrkdef].weapon_count);
        break;
      case zone_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_ZONE_CHAR,
                 GroupDefs[wrkdef].zone_count);
        break;
      case arc_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_ARC_CHAR,
                 GroupDefs[wrkdef].arc_count);
        break;
      case paths_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_PATHS_CHAR,
                 GroupDefs[wrkdef].paths_count);
        break;
      case steps_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_STEPS_CHAR,
                 GroupDefs[wrkdef].steps_count);
        break;
      case hole_key:
      case tophole_key:
        sprintf (name, "%c%c%i", DEF_PREFIX, DEF_HOLE_CHAR,
                 GroupDefs[wrkdef].hole_count);
        break;
      default:
        error ("internal error, parse_object_name()\n");
        return -1;
    }
  }
  else {
    if (find_dup_name (grpdef, name) != 0) {
      error ("duplicate name tag: %s\n", name);
      return -1;
    }
    else if (invalid_name_chars (name)) {
      error ("bad %s name: %s\n", keyword_string (obj->type), name);
      return -1;
    }
  }

  obj->name = strdup (name);
  if (obj->name == NULL) {
    perror ("parse_object_name(strdup)");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
parse_object (int grpdef, object * obj)
{
  int keyword;
  const char *c;
  char line[LINE_LENGTH];
  char name[LINE_LENGTH];


  while ((keyword = get_flowed_line (line, LINE_LENGTH, &c)) >= LINEBUF_OK) {

    if (keyword == end_key) {
      return PARSE_OK;
    }
    else if (keyword == name_key) {
      if (IgnoreName) {
        continue;
      }

      if (obj->type == modify_key) {
        error ("modify objects cannot be named\n");
        return -1;
      }

      if (escan (c, "sw", name) != 1) {
        error ("name parameter\n");
        return -1;
      }

      if (name[0] == DEF_IGNORE_CHAR) {
        continue;               // silently ignore
      }

      if (obj->name[0] != DEF_PREFIX) {
        error ("two names defined (%s & %s)\n", obj->name, name);
        return -1;
      }

      if (find_dup_name (grpdef, name)) {
        error ("duplicate name tag: %s\n", name);
        return -1;
      }

      if (invalid_name_chars (name)) {
        error ("bad %s name: %s\n", keyword_string (obj->type), name);
        return -1;
      }

      free (obj->name);
      obj->name = strdup (name);
      if (obj->name == NULL) {
        perror ("strdup");
        return -1;
      }
    }

    // LINKS

    else if (obj->type == link_key) {
      if (keyword == to_key) {
        if (escan (c, "sw", name) != 1) {
          error ("link-to has no value\n");
          return -1;
        }
        else {
          obj->to_name = strdup (name);
          if (obj->to_name == NULL) {
            perror ("parse_object(strdup)");
            return -1;
          }
        }
      }
      else if (keyword == from_key) {
        if (escan (c, "sw", name) != 1) {
          error ("link-from has no value\n");
          return -1;
        }
        else {
          obj->from_name = strdup (name);
          if (obj->from_name == NULL) {
            perror ("parse_object(strdup)");
            return -1;
          }
        }
      }
      else {
        error ("bad keyword (%s) in link\n", keyword_string (keyword));
        return -1;
      }
    }

    // WORLD

    else if (obj->type == world_key) {
      if (keyword == size_key) {
        if (escan (c, "fw", &WorldDimension) != 1) {
          error ("world size parameter\n");
          return -1;
        }
      }
      else if (keyword == flag_height_key) {
        if (escan (c, "fw", &FlagHeight) != 1) {
          error ("flag height parameter\n");
          return -1;
        }
      }
    }

    // BASICS

    else {
      if (keyword == position_key) {
        if (escan (c, "fffw", &obj->pos[0],
                   &obj->pos[1], &obj->pos[2]) != 3) {
          error ("position parameters\n");
          return -1;
        }
      }
      else if (keyword == size_key) {
        if (escan (c, "fffw", &obj->size[0],
                   &obj->size[1], &obj->size[2]) != 3) {
          error ("size parameters\n");
          return -1;
        }
      }
      else if (keyword == rotation_key) {
        if (escan (c, "fw", &obj->rotation) != 1) {
          error ("rotation parameter\n");
          return -1;
        }
      }
      else if (keyword == drivethrough_key) {
        if (invalid_line_end (c)) {
          return -1;
        }
        obj->drivethrough = 1;
      }
      else if (keyword == shootthrough_key) {
        if (invalid_line_end (c)) {
          return -1;
        }
        obj->shootthrough = 1;
      }
      else if (keyword == passable_key) {
        if (invalid_line_end (c)) {
          return -1;
        }
        obj->drivethrough = 1;
        obj->shootthrough = 1;
      }

      // MISCELLANEOUS

      else if ((keyword == vertex_key) && (obj->type == tetra_key)) {
        if (obj->vertex_count >= 4) {
          error ("too many vertex parameters\n");
          return -1;
        }
        double *vertex = obj->vertices[obj->vertex_count];
        if (escan (c, "fffw", &vertex[0], &vertex[1], &vertex[2]) != 3) {
          error ("not enough vertex parameters\n");
          return -1;
        }
        obj->vertex_count++;
      }
      else if ((keyword == color_key) && (obj->type == tetra_key)) {
        int color[4];
        if (escan (c, "iiiiw", &color[0], &color[1], 
                               &color[2], &color[3]) != 4) {
          error ("vertex color parameter\n");
          return -1;
        }
        if (obj->vertex_count == 0) {
          // change all four vertices
          int i, j;
          for (i = 0; i < 4; i++) {
            obj->colored = obj->colored | (1 << i);
            for (j = 0; j < 4; j++) {
              obj->colors[i][j] = color[j];
            }
          }
        }
        else if (obj->vertex_count < 5) {
          int i;
          int v = obj->vertex_count - 1;
          obj->colored = obj->colored | (1 << v);
          for (i = 0; i < 4; i++) {
            obj->colors[v][i] = color[i];
          }
        }
        else {
          error ("extra color parameter\n");
          return -1; // bzmapper is not a lenient tool
        }
      }
      else if ((keyword == visible_key) && (obj->type == tetra_key)) {
        int v[4];
        if (escan (c, "iiiiw", &v[0], &v[1], &v[2], &v[3]) != 4) {
          error ("not enough visible parameters\n");
          return -1;
        }
        else {
          int i;
          obj->visible = 0x00;
          for (i = 0; i < 4; i++) {
            if (v[i] > 0) {
              obj->visible = obj->visible | (1 << i);
            }
          }
        }
      }
      else if ((keyword == flipz_key) && (obj->type == pyr_key)) {
        if (invalid_line_end (c)) {
          return -1;
        }
        obj->flipz = 1;
      }
      else if ((keyword == color_key) &&
               ((obj->type == base_key) || (obj->type == modify_key))) {
        obj->color = parse_base_color (c);
        if (obj->color == -1) {
          return -1;
        }
      }
      else if ((keyword == vcolor_key) && (obj->type == modify_key)) {
        int plane;
        int color[4];
        if (escan (c, "iiiiiw",  &plane, &color[0], &color[1], 
                               &color[2], &color[3]) != 5) {
          error ("modify vcolor parameters (Face R G B A)\n");
          return -1;
        }
        if (plane == 0) {
          // change all four vertices
          int i, j;
          for (i = 0; i < 4; i++) {
            obj->colored = obj->colored | (1 << i);
            for (j = 0; j < 4; j++) {
              obj->colors[i][j] = color[j];
            }
          }
        }
        else if (plane < 5) {
          int i;
          int v = plane - 1;
          obj->colored = obj->colored | (1 << v);
          for (i = 0; i < 4; i++) {
            obj->colors[v][i] = color[i];
          }
        }
        else {
          error ("modify vcolor plane parameter\n");
          return -1; // bzmapper is not a lenient tool
        }
      }
      else if ((keyword == border_key) &&
               ((obj->type == tele_key) || (obj->type == modify_key))) {
        if (escan (c, "fw", &obj->border) != 1) {
          error ("border parameter\n");
          return -1;
        }
        if (obj->border < 0) {
          error ("border must be a positive value\n");
          return -1;
        }
      }
      else if ((keyword == weapon_type_key) && (obj->type == weapon_key)) {
        if (escan (c, "sw", name) != 1) {
          error ("weapon type parameter\n");
          return -1;
        }

        obj->text = strdup (name);
        if (obj->text == NULL) {
          perror ("strdup");
          return -1;
        }
      }
      else if ((keyword == weapon_init_delay_key) &&
               ((obj->type == weapon_key) || (obj->type == modify_key))) {
        if (escan (c, "fw", &obj->initdelay) != 1) {
          error ("weapon initdelay parameter\n");
          return -1;
        }
      }
      else if ((keyword == weapon_delay_key) && (obj->type == weapon_key)) {
        if (parse_delays (c, obj) == -1) {
          error ("weapon delay parameter\n");
          return -1;
        }
      }
      else if ((keyword == tilt_key) &&
               ((obj->type == text_key) || (obj->type == image_key) ||
                (obj->type == tetra_key) || (obj->type == group_key))) {
        if (escan (c, "fw", &obj->tilt) != 1) {
          error ("tilt parameter\n");
          return -1;
        }

        if ((obj->type == text_key) || (obj->type == image_key)) {
          if ((obj->tilt != 0.0) && (obj->tilt != 90.0)) {
            error ("tilt: only 0.0 and 90.0 are currrently"
                   " supported for text and images\n");
            return -1;
          }
        }
      }
      else if ((keyword == segments_key) && ((obj->type == arc_key) ||
                                             (obj->type == paths_key) ||
                                             (obj->type == steps_key) ||
                                             (obj->type == image_key))) {
        if (escan (c, "iw", &obj->segments) != 1) {
          error ("segments parameter\n");
          return -1;
        }
        if ((obj->segments < 0) && (obj->type != image_key)) {
          error ("segments must be positive for arcs, paths, and steps\n");
          return -1;
        }
      }
      else if ((keyword == angle_key) && (obj->type == arc_key)) {
        if (escan (c, "fw", &obj->angle) != 1) {
          error ("arc angle parameter\n");
          return -1;
        }
      }

      // HOLES, POINTS, AND ANGLES

      else if ((keyword == hole_key) &&
               ((obj->type == box_key) || (obj->type == modify_key))) {
        hole_def hole = {
          {0.0, 0.0, 0.0}       // default position
          ,
          {1.0, 1.0, 1.0}       // default size
        };

        if (parse_hole (&hole) == 0) {
          add_hole (obj, &hole);
        }
        else {
          return -1;
        }
      }
      else if ((keyword == point_key) &&
               ((obj->type == paths_key) || (obj->type == steps_key))) {
        int count;
        point_def *p;
        point_def point = { 0.0, 0.0 };

        if (escan (c, "ffw", &point.x, &point.y) != 2) {
          error ("point parameter\n");
          return -1;
        }
        else {
          if (add_point (obj, &point) == -1)
            return -1;
        }

        p = obj->points;
        count = obj->points_count;

        if ((count > 1) &&
            (IS_EQUAL (p[count - 2].x, p[count - 1].x) &&
             IS_EQUAL (p[count - 2].y, p[count - 1].y))) {
          error ("same point defined twice\n");
          return -1;
        }
      }
      else if ((keyword == angle_key) &&
               ((obj->type == paths_key) || (obj->type == steps_key))) {
        int count;
        point_def *p;
        point_def point = { 0.0, 0.0 };

        if (obj->points_count < 1) {
          error ("you need at least one point before using angles\n");
          return -1;
          // if (add_point (obj, &point) == -1) return -1;
        }

        if (escan (c, "ffw", &point.x, &point.y) != 2) {
          error ("angle parameter\n");
          return -1;
        }
        else {
          point_def tmppoint;
          point_def *oldp = &obj->points[obj->points_count - 1];

          tmppoint.x = oldp->x + cos (M_PI * point.x / 180.0) * point.y;
          tmppoint.y = oldp->y + sin (M_PI * point.x / 180.0) * point.y;
          if (add_point (obj, &tmppoint) == -1)
            return -1;
        }

        p = obj->points;
        count = obj->points_count;

        if ((count > 1) &&
            (IS_EQUAL (p[count - 2].x, p[count - 1].x)) &&
            (IS_EQUAL (p[count - 2].y, p[count - 1].y))) {
          error ("same point defined twice\n");
          return -1;
        }
      }
      else if ((keyword == closedloop_key) &&
               ((obj->type == paths_key) || (obj->type == steps_key))) {
        obj->closedloop = 1;
      }
      else if ((keyword == convert_key) && (obj->type == modify_key)) {
        const char *end;
        char type[LINE_LENGTH];

        if (escan (c, "ws", type) != 1) {
          return -1;
        }

        obj->convert = keyword_parse (type, LINE_LENGTH, &end);
        if (obj->convert == -1) {
          error ("unknown converion type: %s\n", type);
          return -1;
        }

        switch (obj->convert) {
          case box_key:
          case pyr_key:
          case tophole_key:
          case tele_key:
          case base_key:
          case zone_key:
            break;
          default:
            error ("modify convert type isn't appropriate: %s\n",
                   keyword_string (obj->convert));
            return -1;
        }
      }

      // ZONES

      else if ((keyword == flag_key) &&
               ((obj->type == zone_key) || (obj->type == modify_key))) {
        if (escan (c, "s", name) != 1) {
          error ("zone flag type parameter\n");
          return -1;
        }

        if (append_string (&obj->text, c) < 0) {
          return -1;
        }
      }
      else if ((keyword == team_key) &&
               ((obj->type == zone_key) || (obj->type == modify_key))) {
        int color;
        const char *pos = c;
        if (obj->teams < 0) {   // for modify objects
          obj->teams = 0;
        }
        while ((*pos != '\0') && (escan (pos, "in", &color, &pos) == 1)) {
          if ((color < 0) || (color >= TEAM_NAMES)) {
            error ("bad team color: %i\n", color);
            return -1;
          }
          obj->teams = obj->teams | (1 << color);
          while ((*pos != '\0') && (isspace (*pos))) {
            pos++;
          }
        }
        if (pos == c) {
          error ("zone team type parameter\n");
          return -1;
        }
      }
      else if ((keyword == safety_key) &&
               ((obj->type == zone_key) || (obj->type == modify_key))) {
        int color;
        const char *pos = c;
        if (obj->safety < 0) {  // for modify objects
          obj->safety = 0;
        }
        while ((*pos != '\0') && (escan (pos, "in", &color, &pos) == 1)) {
          if ((color < 0) || (color >= TEAM_NAMES)) {
            error ("bad safety color: %i\n", color);
            return -1;
          }
          obj->safety = obj->safety | (1 << color);
          while ((*pos != '\0') && (isspace (*pos))) {
            pos++;
          }
        }
        if (pos == c) {
          error ("zone safety type parameter\n");
          return -1;
        }
      }
      else {
        error ("misplaced keyword: %s (in %s)\n",
               keyword_string (keyword), keyword_string (obj->type));
        return -1;
      }
    }                           // 'not link or world' if
  }                             // while loop


  if (keyword == LINEBUF_EOF) {
    error ("missing end\n");
  }

  return PARSE_ERROR;
}

/******************************************************************************/

static int
parse_hole (hole_def * hole)
{
  int keyword;
  const char *c;
  char line[LINE_LENGTH];


  while ((keyword = get_flowed_line (line, LINE_LENGTH, &c)) >= LINEBUF_OK) {

    if (keyword == end_key) {
      return PARSE_OK;
    }
    else if (keyword == LINEBUF_ERROR) {
      return PARSE_ERROR;
    }
    else if (keyword == position_key) {
      if (escan (c, "fffw", &hole->pos[0],
                 &hole->pos[1], &hole->pos[2]) != 3) {
        error ("position parameters\n");
        return PARSE_ERROR;
      }
    }
    else if (keyword == size_key) {
      if (escan (c, "fffw", &hole->size[0],
                 &hole->size[1], &hole->size[2]) != 3) {
        error ("size parameters\n");
        return PARSE_ERROR;
      }
      if ((hole->size[0] < 0.0) || (hole->size[1] < 0.0) ||
          (hole->size[2] < 0.0)) {
        error ("hole sizes must be positive\n");
        return PARSE_ERROR;
      }
    }
    else {
      return PARSE_ERROR;
    }
  }

  if (keyword == LINEBUF_EOF) {
    error ("missing end\n");
  }

  return PARSE_ERROR;
}

/******************************************************************************/

static int
parse_seed (const char *c)
{
  int seedval;

  if (escan (c, "iw", &seedval) != 1) {
    error ("bad seed value\n");
    return -1;
  }
  else {
    my_srand ((unsigned int) seedval);
    DEBUG3 ("Seeded with %i\n", seedval);
  }

  return 0;
}

/******************************************************************************/

static int
parse_option (const char *c)
{
  int i = 0;
  char buf[LINE_LENGTH];

  while (*c != '\0') {
    if ((*c == '\\') && (*(c + 1) == '(')) {
      const char *pre = c + 1;
      double value;
      if (escan (pre, "bfn", &value, &c) != 1) {
        return PARSE_ERROR;
      }
      sprintf_double (buf + i, value);
      i = i + strlen (buf + i);
    }
    else {
      buf[i] = *c;
      if ((*c == '\\') && (*(c + 1) == '\\')) {
        c++;                    // skip escaped quotes
      }
      c++;
      i++;
    }
  }
  buf[i] = '\0';


  if (append_string (&ConfigOptions, INDENT) < 0) {
    return PARSE_ERROR;
  }
  if (append_string (&ConfigOptions, buf) < 0) {
    return PARSE_ERROR;
  }
  if (append_string (&ConfigOptions, "\n") < 0) {
    return PARSE_ERROR;
  }
  return PARSE_OK;
}

/******************************************************************************/

static int
parse_base_color (const char *c)
{
  int color = -1;

  if (escan (c, "iw", &color) != 1) {
    error ("color parameter\n");
    return -1;
  }

  if (color == 0) {
    DEBUG1 ("useless rogue base\n");
  }

  if ((color < 0) || (color >= TEAM_NAMES)) {
    error ("bad team color: %i\n", color);
    return -1;
  }

  return color;
}

/******************************************************************************/

static int
parse_delays (const char *c, object * obj)
{
  double delay;

  if (obj->delay_count == -1) { // remove default tag
    obj->delay_count = 0;
  }

  while ((*c != '\0') && isspace (*c)) {
    c++;
  }

  while ((*c != '\0') && (escan (c, "fn", &delay, &c) == 1)) {
    if (delay >= MIN_DELAY) {
      double *dtmp;

      obj->delay_count++;
      dtmp = (double *)
          realloc (obj->delays, obj->delay_count * sizeof (double));
      if (dtmp == NULL) {
        perror ("realloc");
        obj->delay_count--;
        return -1;
      }
      else {
        obj->delays = dtmp;
      }
      obj->delays[obj->delay_count - 1] = delay;

      while ((*c != '\0') && isspace (*c)) {
        c++;
      }
      if (*c == '\0') {
        break;
      }
    }
    else {
      error ("delays should be equal or greater then %0.3f, ignoring %f\n",
             MIN_DELAY, delay);
    }
  }

  return 0;
}

/******************************************************************************/

static int
include_font (const char *c)
{
  char filename[LINE_LENGTH];
  char *absname;

  if (escan (c, "Qw", filename) != 1) {
    error ("font filename\n");
    return -1;
  }

  if (get_absolute_name (current_filename (), filename, &absname) == -1) {
    error ("couldn't load font: %s\n", filename);
    error ("continuing to use current font\n");
    return 0;
  }

  DEBUG1 ("\ninclude font (%s)\n\n", filename);

  free_boxfont ();

  if (init_boxfont (absname) != 0) {
    error ("couldn't load font: %s\n", filename);
    error ("continuing to use current font\n");
    free (absname);
    return 0;
  }

  free (absname);

  return 0;
}

/******************************************************************************/

static int
post_parse (object * obj)
{
  int i;
  int keyword;

  keyword = obj->type;

  if (core_object (obj)) {
    if (ConvertOld) {
      if ((obj->size[0] < 0.0) || (obj->size[1] < 0.0) ||
          (obj->size[2] < 0.0)) {
        obj->drivethrough = 1;
        obj->shootthrough = 1;
        printf ("Converted %s to passable\n", obj->name);
      }
      if ((obj->pos[2] > 0.000001) && (obj->pos[2] < 0.2)) {
        obj->drivethrough = 1;
        obj->pos[2] = 0.0;
        printf ("Converted %s to drivethrough\n", obj->name);
      }
    }

    if (InSizeStyle == BZMAPPER_FLIPZ) {
      for (i = 0; i < 3; i++) {
        if (obj->size[i] < 0.0) {
          error ("size[%i] parameter is negative in flipz map\n", i);
          clear_error ();
        }
      }

      obj->size[0] = fabs (obj->size[0]);
      obj->size[1] = fabs (obj->size[1]);
      if (keyword == pyr_key) {
        if ((obj->flipz) || (obj->size[2] < 0.0)) {
          obj->size[2] = -fabs (obj->size[2]);
          obj->pos[2] = obj->pos[2] - obj->size[2]; // push pos[2] upwards
        }
      }
      else {
        obj->size[2] = fabs (obj->size[2]);
      }
    }
    else {
      if ((keyword == pyr_key) && (obj->flipz)) {
        obj->size[2] = -fabs (obj->size[2]);
        obj->pos[2] = obj->pos[2] - obj->size[2];   // push pos[2] upwards
      }
    }
  }


  if (keyword == arc_key) {
    if (validate_arc (obj) == -1) {
      return PARSE_ERROR;
    }
  }
  else if (keyword == tetra_key) {
    if (validate_tetra (obj) == -1) {
      return PARSE_ERROR;
    }
  }
  else if ((keyword == paths_key) || (keyword == steps_key)) {
    if (validate_points (obj) == -1) {
      return PARSE_ERROR;
    }
  }
  else if ((keyword == weapon_key) && (obj->delay_count == -1)) {
    obj->delays = (double *) malloc (sizeof (double));
    if (obj->delays == NULL) {
      perror ("parse_command(delays malloc)");
      return PARSE_ERROR;
    }
    obj->delays[0] = DEF_DELAY;
    obj->delay_count = 1;
  }

  return PARSE_OK;
}

/******************************************************************************/

static int
validate_arc (object * arc)
{
  double angle;

  angle = arc->angle / arc->segments;
  if (angle > 90.0) {
    error ("resulting arc angles are less the 90 degrees.\n");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
validate_tetra (object * tetra)
{
  if (tetra->vertex_count < 4) {
    error ("tetra has too few vertices\n");
    return -1;
  }
  else if (tetra->vertex_count > 4) {
    error ("tetra has too many vertices\n");
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
validate_points (object * obj)
{
  int count;
  point_def *p;

  if (obj->points_count < 2) {
    error ("%s requires at least 2 points\n", keyword_string (obj->type));
    return -1;
  }

  p = obj->points;
  count = obj->points_count;

  if (obj->closedloop) {
    if ((IS_EQUAL (p[0].x, p[count - 1].x)) &&
        (IS_EQUAL (p[0].y, p[count - 1].y))) {
      error ("closedloop %s beginning and end points are the same\n",
             keyword_string (obj->type));
      return -1;
    }
  }

  return 0;
}

/******************************************************************************/

static void
init_object (object * obj, int type)
{
  int i, j;
  
  obj->name = NULL;
  obj->id = -1;

  obj->deleted = 0;

  obj->type = type;
  obj->grpdef = -1;
  obj->pos[0] = 0.0;
  obj->pos[1] = 0.0;
  obj->pos[2] = 0.0;
  obj->size[0] = 1.0;
  obj->size[1] = 1.0;
  obj->size[2] = 1.0;
  obj->rotation = 0.0;
  obj->tilt = 0.0;

  obj->drivethrough = 0;
  obj->shootthrough = 0;
  obj->flipz = 0;

  obj->text = NULL;
  obj->color = 0;
  obj->border = 2.0;
  obj->angle = 360.0;
  obj->closedloop = 0;
  obj->convert = -1;
  obj->border = TeleWidth;
  obj->delays = NULL;
  obj->delay_count = 0;
  obj->teams = 0;
  obj->safety = 0;
  obj->initdelay = DEF_INITDELAY;

  obj->to = -2;                 // -1 'to' is a random tele
  obj->from = -1;
  obj->to_gate = -1;
  obj->from_gate = -1;
  obj->to_name = NULL;
  obj->from_name = NULL;
  
  obj->vertex_count = 0;
  obj->visible = 0x0F;          // all visible
  obj->colored = 0x00;          // no coloring
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      obj->colors[i][j] = 0xff; // all white
    }
  }

  obj->holes = NULL;
  obj->holes_count = 0;

  obj->points = NULL;
  obj->points_count = 0;

  obj->srcline = -1;
  obj->srcfile = NULL;


  switch (type) {
    case box_key:{
      obj->size[0] = BoxBase;
      obj->size[1] = BoxBase;
      obj->size[2] = BoxHeight;
      break;
    }
    case pyr_key:{
      obj->size[0] = PyrBase;
      obj->size[1] = PyrBase;
      obj->size[2] = PyrHeight;
      break;
    }
    case base_key:{
      obj->size[0] = BaseSize;
      obj->size[1] = BaseSize;
      obj->size[2] = 1.0;
      break;
    }
    case tele_key:{
      obj->size[0] = TeleWidth * 0.5;
      obj->size[1] = TeleBreadth;
      obj->size[2] = TeleHeight * 2.0;
      break;
    }
    case arc_key:{
      obj->segments = 4;
      break;
    }
    case paths_key:{
      obj->segments = 1;
      break;
    }
    case steps_key:{
      obj->segments = 1;
      break;
    }
    case image_key:{
      obj->segments = 8;
      break;
    }
    case weapon_key:{
      /*
       * Post parsing checks initialize the weapons delay to defaults
       * if this is still -1 after the object has been parsed.
       */
      obj->delay_count = -1;
      break;
    }
    case modify_key:{
      obj->color = -1;
      obj->teams = -1;
      obj->safety = -1;
      obj->border = -1.0;
      obj->initdelay = -1.0;
      break;
    }
    default:{
      break;
    }
  }

  return;
}

/******************************************************************************/

static int
invalid_name_chars (const char *name)
{
  if (!isalpha (name[0]))
    return 1;

  while (name[0] != '\0') {
    if (!isalnum (name[0]) && !(name[0] == '_')) {
      return 1;
    }
    name++;
  }

  return 0;
}

/******************************************************************************/

static int
invalid_define_name (const char *name)
{
  if (invalid_name_chars (name)) {
    error ("invalid group definition name: %s\n", name);
    return 1;
  }

  switch (*name) {
    case DEF_BOX_CHAR:
    case DEF_PYR_CHAR:
    case DEF_BASE_CHAR:
    case DEF_TELE_CHAR:
    case DEF_LINK_CHAR:
    case DEF_WEAPON_CHAR:
    case DEF_ARC_CHAR:
    case DEF_TEXT_CHAR:
    case DEF_PATHS_CHAR:
    case DEF_STEPS_CHAR:{
      name++;
      while (*name != '\0') {
        if ((*name < '0') || (*name > '9')) {
          return 0;
        }
        name++;
      }
      break;
    }
    default:{
      return 0;
    }
  }

  error ("group definition name might conflict with other default tags\n",
         name);

  return 1;
}

/******************************************************************************/

static int
invalid_line_end (const char *c)
{
  const char *tmp;

  tmp = c;

  while ((*c != '\0') && isspace (*c))
    c++;

  if ((*c == '\0') || ((*c == '#') && (tmp != c))) {
    return 0;
  }
  else if (*c == '#') {
    error ("you need a space before #\n");
    return 1;
  }
  else {
    error ("extra junk (%s)\n", c);
    return 1;
  }
}

/******************************************************************************/

static int
find_dup_name (int grpdef, const char *name)
{
  int i;

  if (name == NULL) {
    return 0;
  }

  if (grpdef == WorldDef) {
    grpdef = ShadowDef;
  }

  for (i = 0; i < GroupDefs[grpdef].objs_count; i++) {
    if (GroupDefs[grpdef].objs[i].name == NULL)
      continue;
    if (strcmp (GroupDefs[grpdef].objs[i].name, name) == 0) {
      return 1;                 // we have a match
    }
  }

  return 0;
}

/******************************************************************************/
/******************************************************************************/

static int
generate_object (object * obj)
{
  int retval;

  push_debug_info (obj->srcline, obj->srcfile);

  switch (obj->type) {
    case box_key:
      if ((obj->type == box_key) && (obj->holes_count > 0))
        retval = generate_boxes (obj);
      else
        retval = generate_core_object (obj);
      break;
    case pyr_key:
    case tetra_key:
    case base_key:
    case tele_key:
    case link_key:
    case weapon_key:
    case zone_key:
      retval = generate_core_object (obj);
      break;
    case group_key:
      retval = generate_group (obj);
      break;
    case arc_key:
      retval = generate_arcs (obj);
      break;
    case text_key:
      retval = generate_text (obj);
      break;
    case image_key:
      retval = generate_image (obj);
      break;
    case paths_key:
    case steps_key:
      retval = generate_paths_steps (obj);
      break;
    case modify_key:
      retval = generate_modify (obj);
      break;
    case delete_key:
      retval = generate_delete (obj);
      break;
    case undelete_key:
      retval = generate_undelete (obj);
      break;
    case intersect_key:
      retval = generate_intersect (obj);
      break;
    case hole_key:
    case tophole_key:
      retval = generate_hole (obj);
      break;
    default:{
      error ("generate_object: internal error (%s)\n",
             keyword_string (obj->type));
      return -1;
    }
  }

  pop_debug_info ();

  return retval;
}

/******************************************************************************/

static int
generate_core_object (object * obj_orig)
{
  vec3 pos;
  vec3 size;
  double rotation;
  char *ctmp;
  object obj;


  memcpy (&obj, obj_orig, sizeof (object)); // don't ruin the original


  calc_position (Position, Size, Rotation, Tilt, obj.pos, pos);
  calc_rotation (Size, obj.rotation, Rotation, &rotation);

  size[0] = Size[0] * obj.size[0];
  size[1] = Size[1] * obj.size[1];
  size[2] = Size[2] * obj.size[2];

  // adjust border size with the X axis size
  obj.border = obj.border * fabs (Size[0]);


  COPY_VECTOR (obj.pos, pos);
  COPY_VECTOR (obj.size, size);
  obj.rotation = rotation;

  if (DriveThrough == 1) {
    obj.drivethrough = 1;
  }
  if (ShootThrough == 1) {
    obj.shootthrough = 1;
  }
  
  // adjust tetra vertices
  if (obj.type == tetra_key) {
    int v;
    for (v = 0; v < 4; v++) {
      calc_position (Position, Size, Rotation, Tilt,
                     obj_orig->vertices[v], obj.vertices[v]);
    }
  }

  // Append and copy the strings
  append_name (obj.name);
  ctmp = strdup (GroupDepthName);
  if (ctmp == NULL) {
    perror ("strdup");
    return -1;
  }
  remove_name (obj.name);
  obj.name = ctmp;

  if (obj.type == link_key) {
    if ((generate_link_name (&obj.to_name, &obj.to,
                             &obj.to_gate, 0) == -1) ||
        (generate_link_name (&obj.from_name, &obj.from,
                             &obj.from_gate, 1) == -1)) {
      free (ctmp);
      if (obj.to_name != obj_orig->to_name)
        free (obj.to_name);
      if (obj.from_name != obj_orig->from_name)
        free (obj.from_name);
      return -1;
    }
  }

  if ((obj.type == zone_key) || (obj.type == weapon_key)) {
    if (obj.text != NULL) {
      ctmp = strdup (obj.text);
      if (ctmp == NULL) {
        perror ("strdup");
        return -1;
      }
      obj.text = ctmp;
    }
  }

  if (obj.type == weapon_key) {
    unsigned int dsize;
    double *dtmp;

    dsize = obj.delay_count * sizeof (double);
    dtmp = (double *) malloc (dsize);
    if (dtmp == NULL) {
      perror ("malloc");
      return -1;
    }
    memcpy (dtmp, obj.delays, dsize);
    obj.delays = dtmp;
  }


  add_object (WorldDef, &obj);


  return 0;
}

/******************************************************************************/

static int
generate_link_name (char **name_ptr, int *num, int *gate, int from)
{
  int len;
  int number;
  char *name;
  char *ctmp;

  name = *name_ptr;

  if (name == NULL) {
    error ("missing link direction\n");
    return -1;
  }
  else {
    len = strlen (name);
  }

  if (strchr (name, DELIMITER) == NULL) {
    // for backwards compatibility
    // even allows expressions
    
    // random 'to' links
    if (!from && (strcmp (name, "random") == 0)) {
      *name_ptr = NULL;
      *num = number;
      *gate = 0;
      return 0;
    }

    if (escan (name, "iw", &number) != 1) {
      error ("bad link: %s\n", name);
      return -1;
    }
    if (number == (unsigned short) (-1)) {
      number = -1;
    }
    if ((number < 0) && (from || (number != RANDOM_LINK))) {
      error ("bad link: %s\n", name);
      return -1;
    }
    else {
      *name_ptr = NULL;
      *num = number;
      *gate = 0;
      return 0;
    }
  }
  else if (len < 3) {
    error ("bad link: %s\n", name);
    return -1;
  }
  else if (name[len - 2] != DELIMITER) {
    error ("no delimiter for (%s)\n", name);
    return -1;
  }


  if (toupper (name[len - 1]) == LINK_FRONT) {
    *gate = 0;
  }
  else if (toupper (name[len - 1]) == LINK_BACK) {
    *gate = 1;
  }
  else {
    error ("bad direction for (%s)\n", name);
    return -1;
  }


  name[len - 2] = '\0';

  if (name[0] == DELIMITER) {   // absolute path
    ctmp = strdup (name + 1);
    *name_ptr = ctmp;
  }
  else {                        // relative path
    append_name (name);
    ctmp = strdup (GroupDepthName);
    remove_name (name);
    *name_ptr = ctmp;
  }

  if (ctmp == NULL) {
    perror ("strdup");
    return -1;
  }

  name[len - 2] = DELIMITER;

  return 0;
}

/******************************************************************************/

static int
generate_group (object * group)
{
  int code = 0;
  int i;
  vec3 tmpPosition;
  vec3 tmpSize;
  double tmpRotation;
  double tmpTilt;
  int tmpDriveThrough;
  int tmpShootThrough;
  vec3 array;
  object *obj;


  COPY_VECTOR (tmpPosition, Position);
  COPY_VECTOR (tmpSize, Size);
  tmpRotation = Rotation;
  tmpTilt = Tilt;
  tmpDriveThrough = DriveThrough;
  tmpShootThrough = ShootThrough;


  calc_position (Position, Size, Rotation, Tilt, group->pos, array);
  calc_rotation (Size, group->rotation, Rotation, &Rotation);

  COPY_VECTOR (Position, array);

  Tilt = Tilt + group->tilt;
  Size[0] = Size[0] * group->size[0];
  Size[1] = Size[1] * group->size[1];
  Size[2] = Size[2] * group->size[2];
  

  if (group->drivethrough == 1)
    DriveThrough = 1;

  if (group->shootthrough == 1)
    ShootThrough = 1;


  append_name (group->name);
  append_name (Delimiter);

  // generate the objects

  for (i = 0; i < GroupDefs[group->grpdef].objs_count; i++) {

    obj = &GroupDefs[group->grpdef].objs[i];

    if (generate_object (obj) == -1) {
      code = -1;
      break;
    }
  }

  remove_name (Delimiter);
  remove_name (group->name);

  COPY_VECTOR (Position, tmpPosition);
  COPY_VECTOR (Size, tmpSize);
  Rotation = tmpRotation;
  Tilt = tmpTilt;
  DriveThrough = tmpDriveThrough;
  ShootThrough = tmpShootThrough;

  return code;
}

/******************************************************************************/

static int
generate_arcs (object * obj)
{
  int i, squared;
  double radius, angle;
  object box;
  vec3 oldPosition;
  double oldRotation;
  int boxcount = 0;
  char namenum[LINE_LENGTH];

  append_name (obj->name);
  append_name (DefPrefix);

  init_object (&box, box_key);
  box.drivethrough = obj->drivethrough;
  box.shootthrough = obj->shootthrough;
  box.srcfile = obj->srcfile;
  box.srcline = obj->srcline;

  COPY_VECTOR (oldPosition, Position);
  oldRotation = Rotation;

  calc_position (oldPosition, Size, Rotation, Tilt, obj->pos, Position);
  calc_rotation (Size, obj->rotation, Rotation, &Rotation);


  angle = obj->angle / (double) obj->segments;

  if (IS_EQUAL (fabs (angle), 90.0))
    squared = 1;
  else
    squared = 0;


  radius = (cos (M_PI * angle / 360.0) * (obj->size[0] + obj->size[1]))
      - obj->size[0];

  box.size[0] = obj->size[0];
  box.size[2] = obj->size[2];
  box.pos[2] = 0.0;             // Position already did this one

  if (squared) {
    box.size[1] = ((obj->size[0] + obj->size[1]) * (sqrt (2.0) / 2.0))
        - obj->size[0];
  }
  else {
    box.size[1] =
        sin (M_PI * angle / 360.0) * (obj->size[1] + obj->size[0]);
  }


  for (i = 0; i < obj->segments; i++) {

    box.rotation = angle * (double) i + (angle / 2.0);

    if (squared) {
      box.pos[0] = (radius * cos (M_PI * box.rotation / 180.0)) +
          (obj->size[0] * sin (M_PI * box.rotation / 180.0));
      box.pos[1] = (radius * sin (M_PI * box.rotation / 180.0)) -
          (obj->size[0] * cos (M_PI * box.rotation / 180.0));
    }
    else {
      box.pos[0] = (radius * cos (M_PI * box.rotation / 180.0));
      box.pos[1] = (radius * sin (M_PI * box.rotation / 180.0));
    }

    sprintf (namenum, "%i", boxcount);
    append_name (namenum);
    if (generate_object (&box) != 0)
      return -1;
    remove_name (namenum);
    boxcount++;
  }

  COPY_VECTOR (Position, oldPosition);
  Rotation = oldRotation;

  remove_name (DefPrefix);
  remove_name (obj->name);

  return 0;
}

/******************************************************************************/

static int
generate_text (object * obj)
{
  int i, j;
  glyph_t g;
  object box;
  double x = 0.0;
  vec3 oldPosition;
  vec3 oldSize;
  double oldRotation;
  int boxcount = 0;
  char namenum[16];
  char charnum[16];

  append_name (obj->name);
  append_name (DefPrefix);

  init_object (&box, box_key);
  box.drivethrough = obj->drivethrough;
  box.shootthrough = obj->shootthrough;
  box.srcfile = obj->srcfile;
  box.srcline = obj->srcline;

  COPY_VECTOR (oldPosition, Position);
  COPY_VECTOR (oldSize, Size);
  oldRotation = Rotation;

  calc_position (oldPosition, Size, Rotation, Tilt, obj->pos, Position);
  calc_rotation (Size, obj->rotation, Rotation, &Rotation);
  Size[0] = Size[0] * obj->size[0];

  if (obj->tilt != 90.0) {
    Size[1] = Size[1] * obj->size[1];
    Size[2] = Size[2] * obj->size[2];
  }
  else {
    Size[1] = Size[1] * obj->size[2];
    Size[2] = Size[2] * obj->size[1];
  }

  i = 0;
  while (obj->text[i] != '\0') {

    sprintf (charnum, "%i", i);
    append_name (charnum);
    append_name (DefPrefix);

    /*
     * Watch out for the (unsigned char) cast here.
     * If the cast isn't there, then the array index
     * will go negative before we want to, and the
     * SEGVs will start to show at around 128.
     *
     * It would be nice to get this Unicoded.
     */
    if (boxfont_char ((unsigned char) obj->text[i], &g) == -1) {
      error ("invalid text char: 0x%02X (%c).\n",
             obj->text[i], obj->text[i]);
      return -1;
    }

    boxcount = 0;

    for (j = 0; j < g.rectcount; j++) {

      if (obj->tilt != 90.0) {
        box.size[1] = (double) (g.rects[j].height) / 2.0;
        box.pos[1] = (double) (g.rects[j].y * 2 + g.rects[j].height) / 2.0;
        box.size[2] = 1.0;
        box.pos[2] = 0.0;
      }
      else {
        box.size[1] = 0.5;
        box.pos[1] = 0.0;
        box.size[2] = (double) (g.rects[j].height);
        box.pos[2] = (double) (g.rects[j].y);
      }

      box.size[0] = (double) (g.rects[j].width) / 2.0;
      box.pos[0] =
          (double) (g.rects[j].x * 2 + g.rects[j].width) / 2.0 + x;

      sprintf (namenum, "%i", boxcount);
      append_name (namenum);
      if (generate_object (&box) != 0)
        return -1;
      remove_name (namenum);
      boxcount++;
    }

    remove_name (DefPrefix);
    remove_name (charnum);
    x = x + 8.0 + 1.0;          // assumes an 8-bit character width
    i++;
  }

  COPY_VECTOR (Position, oldPosition);
  COPY_VECTOR (Size, oldSize);
  Rotation = oldRotation;

  remove_name (DefPrefix);
  remove_name (obj->name);

  return 0;
}

/******************************************************************************/

static int
generate_image (object * obj)
{
  int i;
  int invert;
  char *absname;
  box_t *boxes;
  int box_count;
  int width, length;
  object box;
  vec3 oldPosition;
  vec3 oldSize;
  double oldRotation;
  char namenum[16];


  if (get_absolute_name (current_filename (), obj->text, &absname) == -1) {
    error ("could not load image: %s\n", obj->text);
    return -1;
  }

  if (obj->segments < 0)
    invert = 1;
  else
    invert = 0;

  if (bmp2bzw (absname, &boxes, &box_count, &width, &length,
               abs (obj->segments), invert) != 0) {
    error ("bmp2bzw\n");
    free (absname);
    return -1;
  }

  free (absname);


  append_name (obj->name);
  append_name (DefPrefix);

  init_object (&box, box_key);
  box.drivethrough = obj->drivethrough;
  box.shootthrough = obj->shootthrough;
  box.srcfile = obj->srcfile;
  box.srcline = obj->srcline;

  COPY_VECTOR (oldPosition, Position);
  COPY_VECTOR (oldSize, Size);
  oldRotation = Rotation;

  calc_position (oldPosition, Size, Rotation, Tilt, obj->pos, Position);
  calc_rotation (Size, obj->rotation, Rotation, &Rotation);

  Size[0] = Size[0] * obj->size[0];

  if (obj->tilt != 90.0) {
    Size[1] = Size[1] * obj->size[1];
    Size[2] = Size[2] * obj->size[2];
  }
  else {
    Size[1] = Size[1] * obj->size[2];
    Size[2] = Size[2] * obj->size[1];
  }


  for (i = 0; i < box_count; i++) {

    box.pos[0] = (double) boxes[i].x + ((double) boxes[i].sx) / 2.0;
    box.size[0] = ((double) boxes[i].sx) / 2.0;

    if (obj->tilt != 90.0) {
      box.pos[1] = (double) boxes[i].y + ((double) boxes[i].sy) / 2.0;
      box.size[1] = ((double) boxes[i].sy) / 2.0;
      box.pos[2] = 0.0;
      box.size[2] = (double) boxes[i].sz;
    }
    else {
      box.pos[1] = -((double) boxes[i].sz) / 2.0;
      box.size[1] = ((double) boxes[i].sz) / 2.0;
      box.pos[2] = (double) boxes[i].y;
      box.size[2] = (double) boxes[i].sy;
    }

    sprintf (namenum, "%i", i);
    append_name (namenum);
    if (generate_object (&box) != 0) {
      free (boxes);
      return -1;
    }
    remove_name (namenum);
  }

  COPY_VECTOR (Position, oldPosition);
  COPY_VECTOR (Size, oldSize);
  Rotation = oldRotation;

  remove_name (DefPrefix);
  remove_name (obj->name);

  free (boxes);

  return 0;
}

/******************************************************************************/

static inline double
length2d (point_def * p)
{
  return sqrt ((p->x * p->x) + (p->y * p->y));
}

/******************************************************************************/

/*
 * Calculate the angle between two vectors (in radians).
 * Both vectors point towards p1.
 */
static double
calc_angle (point_def * p0, point_def * p1, point_def * p2)
{
  double valx, valy, angle;
  point_def v01, v12;

  v01.x = p1->x - p0->x;
  v01.y = p1->y - p0->y;
  v12.x = p2->x - p1->x;
  v12.y = p2->y - p1->y;

  valy = (v01.x * v12.y) - (v01.y * v12.x);
  valx = (v01.x * v12.x) + (v01.y * v12.y);

  angle = atan2 (valy, valx);

  DEBUG9 ("CALC_ANGLE (%f, %f) (%f, %f) (%f, %f) = %f\n",
          p0->x, p0->y, p1->x, p1->y, p2->x, p2->y, angle);

  return angle;
}

/******************************************************************************/

/*
 * Calculate the point that is 'dist' away from both vectors on the
 * right side of the vector going from p0 towards p1. If you want,
 * the left side, use a negative distance.
 */
static int
path_right (point_def * p0, point_def * p1, point_def * p2,
            point_def * result, double dist)
{
  point_def v01, v21, v;
  double len01, len21, len;
  double tmp;

  v01.x = p1->x - p0->x;
  v01.y = p1->y - p0->y;
  v21.x = p1->x - p2->x;
  v21.y = p1->y - p2->y;

  len01 = length2d (&v01);
  len21 = length2d (&v21);

  if ((IS_ZERO (len01)) || (IS_ZERO (len21))) {
    error ("path_right: zero length vector\n");
    return -1;
  }

  if (IS_ZERO (dist)) {
    /*
     * We could do this earlier, but the length
     * error checking is a nice safety to have.
     */
    result->x = p1->x;
    result->y = p1->y;
    return 0;
  }

  v01.x = v01.x / len01;        // normalize v01
  v01.y = v01.y / len01;
  v21.x = v21.x / len21;        // normalize v21
  v21.y = v21.y / len21;

  v.x = v01.x + v21.x;
  v.y = v01.y + v21.y;

  len = length2d (&v);

  if (IS_ZERO (len)) {
    DEBUG7 ("path_right: going PERP\n");
    v.x = +dist * v01.y;
    v.y = -dist * v01.x;

    result->x = p1->x + v.x;
    result->y = p1->y + v.y;

    return 0;
  }

  tmp = ((v01.y * v.x) - (v01.x * v.y)) / dist;

  if (IS_ZERO (tmp)) {
    DEBUG7 ("path_right: cutback condition\n");
    result->x = v01.x * dist;
    result->y = v01.y * dist;
  }
  else {
    v.x = v.x / tmp;
    v.y = v.y / tmp;
  }

  result->x = p1->x + v.x;
  result->y = p1->y + v.y;

  return 0;
}

/******************************************************************************/

/*
 * Calculate the point that is 'dist' away from p0 on the right side of the
 * vector going from p0 towards p1, and perpendicular to it. If you want the
 * left side, use a negative distance. If you want to start from p1 instead
 * of p0 but still want the right side, swap the two points and negated the
 * distance.
 */
static int
perp_right (point_def * p0, point_def * p1, point_def * result,
            double dist)
{
  point_def v;
  double x, y, len;

  v.x = p1->x - p0->x;
  v.y = p1->y - p0->y;

  len = length2d (&v);

  if (IS_ZERO (len)) {
    error ("perp_right: zero length vector\n");
    return -1;
  }

  x = dist * (v.y / len);
  y = -dist * (v.x / len);

  result->x = p0->x + x;
  result->y = p0->y + y;

  return 0;
}

/******************************************************************************/

/*
 * Calculate the point that is 'dist' away from the p1 points that
 * along the line between both vectors . The point is away from the
 * two outside points.
 */
static int
gap_outside (point_def * p0, point_def * p1, point_def * p2,
             point_def * result, double dist)
{
  point_def v01, v21, v;
  double len01, len21, len;

  v01.x = p1->x - p0->x;
  v01.y = p1->y - p0->y;
  v21.x = p1->x - p2->x;
  v21.y = p1->y - p2->y;

  len01 = length2d (&v01);
  len21 = length2d (&v21);

  if ((IS_ZERO (len01)) || (IS_ZERO (len21))) {
    error ("gap_outside: zero length vector\n");
    return -1;
  }

  if (IS_ZERO (dist)) {
    error ("gap_outside: internal error: gap distance is 0.0\n");
    return 0;
  }

  v01.x = v01.x / len01;        // normalize v01
  v01.y = v01.y / len01;
  v21.x = v21.x / len21;        // normalize v21
  v21.y = v21.y / len21;

  v.x = v01.x + v21.x;
  v.y = v01.y + v21.y;

  len = length2d (&v);

  v.x = v.x / len;              // normalize v
  v.y = v.y / len;

  if (IS_ZERO (len)) {
    error ("gap_outside: internal error: gapping a cutback\n");
    return -1;
  }

  result->x = p1->x + (v.x * dist);
  result->y = p1->y + (v.y * dist);

  return 0;
}

/******************************************************************************/

static int
generate_path (object * obj, int *indices)
{
  int i;
  object box;
  point_def *p;
  int count, points;
  double angle;
  double midx, midy;
  double xlen, ylen, len;
  double preadj, postadj;
  vec3 oldPosition;
  double oldRotation;
  int boxcount = 0;
  char namenum[LINE_LENGTH];


  init_object (&box, box_key);
  box.drivethrough = obj->drivethrough;
  box.shootthrough = obj->shootthrough;
  box.srcfile = obj->srcfile;
  box.srcline = obj->srcline;

  COPY_VECTOR (oldPosition, Position);
  oldRotation = Rotation;

  calc_position (oldPosition, Size, Rotation, Tilt, obj->pos, Position);
  calc_rotation (Size, obj->rotation, Rotation, &Rotation);

  box.pos[2] = 0.0;             // Position has already done the job
  box.size[0] = obj->size[0];
  box.size[2] = obj->size[2];

  p = obj->points;
  points = obj->points_count;

  if (obj->closedloop) {
    count = points;
  }
  else {
    count = points - 1;
  }


  for (i = 0; i < count; i++) {

    midx = (p[i % points].x + p[(i + 1) % points].x) / 2.0;
    midy = (p[i % points].y + p[(i + 1) % points].y) / 2.0;
    xlen = p[(i + 1) % points].x - p[i % points].x;
    ylen = p[(i + 1) % points].y - p[i % points].y;
    len = sqrt (xlen * xlen + ylen * ylen);


    angle = calc_angle (&p[(i - 1 + points) % points],
                        &p[(i + 0) % points], &p[(i + 1) % points]);
    preadj = box.size[0] * fabs (tan (angle / 2.0));

    if (fabs (angle) > (M_PI / 2.0)) {
      preadj = 0.0;
      DEBUG9 ("Acute ");
    }
    DEBUG9 ("Angle: %f\n", angle * (180.0 / M_PI));


    angle = calc_angle (&p[(i + 0) % points],
                        &p[(i + 1) % points], &p[(i + 2) % points]);
    postadj = box.size[0] * fabs (tan (angle / 2.0));

    if (fabs (angle) > (M_PI / 2.0)) {
      postadj = 0.0;
      DEBUG9 ("Acute ");
    }
    DEBUG9 ("Angle: %f\n", angle * (180.0 / M_PI));


    if ((i == 0) && (!obj->closedloop)) {
      preadj = 0.0;
    }

    if ((i == (count - 1)) && (!obj->closedloop)) {
      postadj = 0.0;
    }

    // Don't overlap if we have a 90 bend

    if ((IS_EQUAL ((M_PI / 2.0), fabs (angle))) &&
        (obj->closedloop || (i != (count - 1)))) {
      postadj = -box.size[0];
    }

    DEBUG9 ("PreAdj = %f\n", preadj);
    DEBUG9 ("PostAdj = %f\n", postadj);
    DEBUG9 ("Xadj = %f\n", (postadj - preadj) * (xlen / len) / 2.0);
    DEBUG9 ("Yadj = %f\n", (postadj - preadj) * (ylen / len) / 2.0);

    box.pos[0] = midx + (postadj - preadj) * (xlen / len) / 2.0;
    box.pos[1] = midy + (postadj - preadj) * (ylen / len) / 2.0;
    box.size[1] = (len + preadj + postadj) / 2.0;
    box.rotation = (180.0 / M_PI) * atan2 (ylen, xlen) + 90.0;

    sprintf (namenum, "%i", indices[i]);
    append_name (namenum);
    if (generate_object (&box) != 0)
      return -1;

    // Add the acute angle segment

    if ((fabs (angle) > (M_PI / 2.0))
        && (!IS_EQUAL (M_PI, fabs (angle))) &&  // cutbacks
        ((obj->closedloop) || (i != (count - 1)))) {
      double postangle;
      point_def point;

      box.size[0] = cos (angle / 2.0) * obj->size[0] / 2.0;
      box.size[1] = fabs (sin (angle / 2.0)) * obj->size[0];

      if (gap_outside (&p[i % points], &p[(i + 1) % points],
                       &p[(i + 2) % points], &point, box.size[0]) == -1) {
        return -1;
      }
      box.pos[0] = point.x;
      box.pos[1] = point.y;

      postangle = (180.0 / M_PI) *
          atan2 ((p[(i + 2) % points].y - p[(i + 1) % points].y),
                 (p[(i + 2) % points].x - p[(i + 1) % points].x));

      box.rotation = ((box.rotation + 90.0) + postangle) / 2.0;

      append_name (GAP_STRING);
      if (generate_object (&box) != 0)
        return -1;
      remove_name (GAP_STRING);

      box.size[0] = obj->size[0];
    }

    remove_name (namenum);
    boxcount++;
  }

  COPY_VECTOR (Position, oldPosition);
  Rotation = oldRotation;


  return 0;
}

/******************************************************************************/

static int
clean_path (object * obj, int *indices)
{
  point_def *p;
  int i;
  int points;
  int count = 0;

  p = obj->points;
  points = obj->points_count;

  for (i = 0; i < points; i++) {
    if ((!IS_EQUAL (p[i].x, p[(i + 1) % points].x)) ||
        (!IS_EQUAL (p[i].y, p[(i + 1) % points].y)) ||
        ((i == (points - 1)) && (!obj->closedloop))) {
      if (i != count) {
        memcpy (&p[count], &p[i], sizeof (point_def));
      }
      indices[count] = i;
      count++;
    }
  }

  DEBUG8 ("CLEAN PATH (%s): old = %i, new = %i\n",
          obj->name, obj->points_count, count);

  return count;
}

/******************************************************************************/

static int
generate_paths_steps (object * obj)
{
  int i, path;
  int points;
  point_def *p;
  point_def *shifted;
  int *indices;
  double dist, height = 0.0;
  char pathnum[LINE_LENGTH];


  append_name (obj->name);
  append_name (DefPrefix);


  p = obj->points;
  points = obj->points_count;

  indices = (int *) malloc (sizeof (int) * points);
  if (indices == NULL) {
    perror ("generate_paths_steps(malloc)");
    return -1;
  }


  shifted = (point_def *) malloc (sizeof (point_def) * points);
  if (shifted == NULL) {
    perror ("generate_paths_steps(malloc)");
    return -1;
  }


  if (obj->type == paths_key)
    dist = -(obj->size[1] * (double) (obj->segments - 1)) / 2.0;
  else
    dist = obj->size[0] + obj->size[1];


  for (path = 0; path < obj->segments; path++) {

    for (i = 0; i < points; i++) {
      if ((!obj->closedloop) && (i == 0)) {
        if (perp_right (&p[0], &p[1], &shifted[0], dist) == -1) {
          free (shifted);
          free (indices);
          return -1;
        }
      }
      else if ((!obj->closedloop) && (i == (points - 1))) {
        if (perp_right (&p[i], &p[i - 1], &shifted[i], -dist) == -1) {
          free (shifted);
          free (indices);
          return -1;
        }
      }
      else {
        if (path_right (&p[(i - 1 + points) % points],
                        &p[i % points], &p[(i + 1) % points],
                        &shifted[i], dist) == -1) {
          free (shifted);
          free (indices);
          return -1;
        }
      }
    }

    sprintf (pathnum, "%i", path);
    append_name (pathnum);
    append_name (DefPrefix);

    if (obj->type == steps_key) {
      height = obj->size[2];
      obj->size[2] = height * (double) (obj->segments - path);
    }
    obj->points = shifted;
    obj->points_count = clean_path (obj, indices);

    if (generate_path (obj, indices) == -1) {
      obj->points = p;
      free (shifted);
      free (indices);
      return -1;
    }

    obj->points = p;
    obj->points_count = points;
    if (obj->type == steps_key)
      obj->size[2] = height;

    remove_name (DefPrefix);
    remove_name (pathnum);

    if (obj->type == paths_key)
      dist = dist + obj->size[1];
    else
      dist = dist + (obj->size[0] * 2.0);
  }

  free (shifted);
  free (indices);

  remove_name (DefPrefix);
  remove_name (obj->name);

  return 0;
}

/******************************************************************************/

// Holes can't have any negative size parameters

static int
generate_boxes (object * obj)
{
  int i, j;
#ifdef SAME_POLARITY_HOLES
  int negx = 0, negy = 0, negz = 0;
#endif                          // SAME_POLARITY_HOLES
  listbox_def *pre = NULL;
  int precount = 0;
  listbox_def *post = NULL;
  int postcount = 0;
  listbox_def *ptr = NULL;
  object box;
  vec3 oldPosition;
  double oldRotation;
  int boxcount = 0;
  char namenum[LINE_LENGTH];
  char *FIXME;                  // LIST
  vec3 tmpSize;                 // FIXME - LIST

  append_name (obj->name);
  append_name (DefPrefix);
  box.srcfile = obj->srcfile;
  box.srcline = obj->srcline;

  precount = 1;
  pre = (listbox_def *) malloc (sizeof (listbox_def));
  if (pre == NULL) {
    perror ("malloc");
    return -1;
  }

#ifdef SAME_POLARITY_HOLES
  if (obj->size[0] < 0.0) {
    negx = 1;
  }
  if (obj->size[1] < 0.0) {
    negy = 1;
  }
  if (obj->size[2] < 0.0) {
    negz = 1;
  }
#endif                          // SAME_POLARITY_HOLES

  pre[0].size[0] = fabs (obj->size[0]);
  pre[0].size[1] = fabs (obj->size[1]);
  pre[0].size[2] = fabs (obj->size[2]);
  pre[0].pos[0] = 0.0;
  pre[0].pos[1] = 0.0;
  pre[0].pos[2] = 0.0;
  if (obj->size[2] < 0.0) {
    pre[0].pos[2] = pre[0].pos[2] - pre[0].size[2];
  }

  for (i = 0; i < obj->holes_count; i++) {
    hole_def hole;

    COPY_VECTOR (hole.pos, obj->holes[i].pos);
    COPY_VECTOR (hole.size, obj->holes[i].size);

    post = NULL;
    postcount = 0;

    for (j = 0; j < precount; j++) {
      /*
       * is the hole outside this box?
       */
      if ((pre[j].pos[2] >= (hole.pos[2] + hole.size[2])) ||
          ((pre[j].pos[2] + pre[j].size[2]) <= hole.pos[2]) ||
          (fabs (pre[j].pos[1] - hole.pos[1]) >=
           (pre[j].size[1] + hole.size[1])) ||
          (fabs (pre[j].pos[0] - hole.pos[0]) >=
           (pre[j].size[0] + hole.size[0]))) {
        /*
         * copy the box, no changes
         */
        postcount++;
        ptr = (listbox_def *) realloc (post,
                                       sizeof (listbox_def) * postcount);
        if (ptr == NULL) {
          free (pre);
          free (post);
          return -1;
        }
        else {
          post = ptr;

          memcpy (&post[postcount - 1], &pre[j], sizeof (listbox_def));
        }
      }
      else {
        double posy;
        double sizey;
        double posz;
        double sizez;
        double miny = pre[j].pos[1] - pre[j].size[1];
        double maxy = pre[j].pos[1] + pre[j].size[1];
        double minz = pre[j].pos[2];
        double maxz = pre[j].pos[2] + pre[j].size[2];

        /*
         * bottom box
         */
        if (pre[j].pos[2] < hole.pos[2]) {

          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] = pre[j].pos[0];
            post[postcount - 1].pos[1] = pre[j].pos[1];
            post[postcount - 1].pos[2] = pre[j].pos[2];

            post[postcount - 1].size[0] = pre[j].size[0];
            post[postcount - 1].size[1] = pre[j].size[1];
            post[postcount - 1].size[2] = hole.pos[2] - pre[j].pos[2];

            minz = hole.pos[2];
          }
        }                       /* end IF bottom box */

        /*
         * top box
         */
        if ((pre[j].pos[2] + pre[j].size[2]) >
            (hole.pos[2] + hole.size[2])) {

          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] = pre[j].pos[0];
            post[postcount - 1].pos[1] = pre[j].pos[1];
            post[postcount - 1].pos[2] = hole.pos[2] + hole.size[2];

            post[postcount - 1].size[0] = pre[j].size[0];
            post[postcount - 1].size[1] = pre[j].size[1];
            post[postcount - 1].size[2] =
                (pre[j].pos[2] + pre[j].size[2]) -
                (hole.pos[2] + hole.size[2]);

            maxz = hole.pos[2] + hole.size[2];
          }
        }                       /* end IF top box */

        posz = minz;
        sizez = maxz - minz;

        /*
         * front box
         */
        if ((pre[j].pos[1] + pre[j].size[1]) >
            (hole.pos[1] + hole.size[1])) {

          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] = pre[j].pos[0];
            post[postcount - 1].pos[1] =
                ((pre[j].pos[1] + pre[j].size[1]) +
                 (hole.pos[1] + hole.size[1])) / 2.0;
            post[postcount - 1].pos[2] = posz;

            post[postcount - 1].size[0] = pre[j].size[0];
            post[postcount - 1].size[1] =
                ((pre[j].pos[1] + pre[j].size[1]) -
                 (hole.pos[1] + hole.size[1])) / 2.0;
            post[postcount - 1].size[2] = sizez;

            maxy = hole.pos[1] + hole.size[1];
          }
        }                       /* end IF front box */

        /*
         * back box
         */
        if ((pre[j].pos[1] - pre[j].size[1]) <
            (hole.pos[1] - hole.size[1])) {

          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] = pre[j].pos[0];
            post[postcount - 1].pos[1] =
                ((pre[j].pos[1] - pre[j].size[1]) +
                 (hole.pos[1] - hole.size[1])) / 2.0;
            post[postcount - 1].pos[2] = posz;

            post[postcount - 1].size[0] = pre[j].size[0];
            post[postcount - 1].size[1] =
                ((hole.pos[1] - hole.size[1]) -
                 (pre[j].pos[1] - pre[j].size[1])) / 2.0;
            post[postcount - 1].size[2] = sizez;

            miny = hole.pos[1] - hole.size[1];
          }
        }                       /* end IF back box */

        posy = (maxy + miny) / 2.0;
        sizey = (maxy - miny) / 2.0;

        /*
         * right box
         */
        if ((pre[j].pos[0] + pre[j].size[0]) >
            (hole.pos[0] + hole.size[0])) {
          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] =
                ((pre[j].pos[0] + pre[j].size[0]) +
                 (hole.pos[0] + hole.size[0])) / 2.0;
            post[postcount - 1].pos[1] = posy;
            post[postcount - 1].pos[2] = posz;

            post[postcount - 1].size[0] =
                ((pre[j].pos[0] + pre[j].size[0]) -
                 (hole.pos[0] + hole.size[0])) / 2.0;
            post[postcount - 1].size[1] = sizey;
            post[postcount - 1].size[2] = sizez;
          }
        }                       /* end IF right box */

        /*
         * left box
         */
        if ((pre[j].pos[0] - pre[j].size[0]) <
            (hole.pos[0] - hole.size[0])) {
          postcount++;
          ptr = (listbox_def *) realloc (post,
                                         sizeof (listbox_def) * postcount);
          if (ptr == NULL) {
            free (pre);
            free (post);
            return -1;
          }
          else {
            post = ptr;

            post[postcount - 1].pos[0] =
                ((pre[j].pos[0] - pre[j].size[0]) +
                 (hole.pos[0] - hole.size[0])) / 2.0;
            post[postcount - 1].pos[1] = posy;
            post[postcount - 1].pos[2] = posz;

            post[postcount - 1].size[0] =
                ((hole.pos[0] - hole.size[0]) -
                 (pre[j].pos[0] - pre[j].size[0])) / 2.0;
            post[postcount - 1].size[1] = sizey;
            post[postcount - 1].size[2] = sizez;
          }
        }                       /* end IF left box */
      }                         /* end IF (hole not in box) */
    }                           /* end FOR loop (box list) */

    free (pre);
    precount = postcount;
    pre = post;
  }                             /* end FOR loop (holes) */


  COPY_VECTOR (oldPosition, Position);
  oldRotation = Rotation;

  /*
   * generate the new boxes, the Size parameter
   * is not modified because the relative size
   * is not changed by cutting a box into pieces.
   */

  init_object (&box, box_key);

  calc_position (oldPosition, Size, Rotation, Tilt, obj->pos, Position);
  calc_rotation (Size, obj->rotation, Rotation, &Rotation);

  COPY_VECTOR (box.size, obj->size);
  box.drivethrough = obj->drivethrough;
  box.shootthrough = obj->shootthrough;
  box.rotation = 0.0;


  /*
   * Because I still have crap stuck in arrays, accesses
   * to the obj->name kill things after you've added new
   * boxes. Hence, I save the name before generating any
   * new objects.
   */
  FIXME = obj->name;            // LIST
  COPY_VECTOR (tmpSize, obj->size); // FIXME - LIST

  for (i = 0; i < postcount; i++) {

    // remove slivers
    if ((IS_ZERO (post[i].size[0]) && !IS_ZERO (tmpSize[0])) ||
        (IS_ZERO (post[i].size[1]) && !IS_ZERO (tmpSize[1])) ||
        (IS_ZERO (post[i].size[2]) && !IS_ZERO (tmpSize[2]))) {
      continue;
    }

    COPY_VECTOR (box.size, post[i].size);
    COPY_VECTOR (box.pos, post[i].pos);

#ifdef SAME_POLARITY_HOLES
    if (negx) {
      box.size[0] = -box.size[0];
    }
    if (negy) {
      box.size[1] = -box.size[1];
    }
    if (negz) {
      box.pos[2] = box.pos[2] + box.size[2];
      box.size[2] = -(box.size[2]);
    }
#endif                          // SAME_POLARITY_HOLES

    sprintf (namenum, "%i", boxcount);
    append_name (namenum);

    if (generate_object (&box) != 0) {
      free (post);
      return -1;
    }
    remove_name (namenum);
    boxcount++;
  }

  COPY_VECTOR (Position, oldPosition);
  Rotation = oldRotation;

  free (post);

  remove_name (DefPrefix);
  remove_name (FIXME);          // LIST

  return 0;
}

/******************************************************************************/

static int
generate_modify (object * obj)
{
  int i;
  int matches = 0;
  char matchName[LINE_LENGTH];
  object *target;
  int current_count;
  vec3 oldPosition;
  vec3 oldSize;
  double oldRotation;
  int oldDriveThrough;
  int oldShootThrough;
  char oldGroupDepthName[LINE_LENGTH];

  if (obj->name[0] == DELIMITER) {  /* absolute naming */
    error ("modify can't use absolute naming.\n");
    return -1;
  }
  else {
    append_name (obj->name);
    strcpy (matchName, GroupDepthName);
    remove_name (obj->name);
  }

  COPY_VECTOR (oldPosition, Position);
  COPY_VECTOR (oldSize, Size);
  oldRotation = Rotation;
  oldDriveThrough = DriveThrough;
  oldShootThrough = ShootThrough;
  strcpy (oldGroupDepthName, GroupDepthName);
  GroupDepthName[0] = '\0';

  COPY_VECTOR (Position, Zeroes);
  COPY_VECTOR (Size, Ones);
  Rotation = 0.0;
  DriveThrough = 0;
  ShootThrough = 0;

  /*
   * If holes are added, then the total number of objects increases.
   * Using GroupDefs[WorldDef].objs_count then becomes unsafe.
   */
  current_count = GroupDefs[WorldDef].objs_count;

  for (i = 0; i < current_count; i++) {

    target = &GroupDefs[WorldDef].objs[i];

    if (target->deleted) {
      continue;
    }

    if (match_object_name (matchName, target->name)) {

      matches++;
      DEBUG5 ("Modifying %s\n", GroupDefs[WorldDef].objs[i].name);

      target->srcfile = obj->srcfile;
      target->srcline = obj->srcline;

      target->pos[0] = target->pos[0] + obj->pos[0];
      target->pos[1] = target->pos[1] + obj->pos[1];
      target->pos[2] = target->pos[2] + obj->pos[2];
      target->size[0] = target->size[0] * obj->size[0];
      target->size[1] = target->size[1] * obj->size[1];
      target->size[2] = target->size[2] * obj->size[2];
      target->rotation = target->rotation + obj->rotation;

      if (obj->drivethrough == 1) {
        target->drivethrough = 1;
      }

      if (obj->shootthrough == 1) {
        target->shootthrough = 1;
      }

      // Convert the type
      if ((obj->convert != -1) &&
          (obj->convert != hole_key) && (obj->convert != tophole_key) &&
          ((target->type == box_key) || (target->type == pyr_key) ||
           (target->type == tele_key) || (target->type == base_key) ||
           (target->type == zone_key))) {
        target->type = obj->convert;
      }

      if ((target->type == base_key) && (obj->color > 0)) {
        target->color = obj->color;
      }
      else if ((target->type == tele_key) && (obj->border >= 0.0)) {
        target->border = obj->border;
      }
      else if ((target->type == weapon_key) && (obj->initdelay >= 0.0)) {
        target->initdelay = obj->initdelay;
      }
      else if (target->type == zone_key) {
        if (obj->teams > 0) {
          target->teams = obj->teams;
        }
        if (obj->safety > 0) {
          target->safety = obj->safety;
        }
        if (obj->text != NULL) {    /* zone flags */
          free (target->text);
          target->text = strdup (obj->text);
        }
      }
      else if (target->type == tetra_key) {
        int v;
        for (v = 0; v < 4; v++) {
          if (obj->colored & (1 << v)) {
            target->colored = target->colored | (1 << v);
            memcpy (target->colors, obj->colors, sizeof(char[4][4]));
          }
        }
      }


      // Make new holes in an old box

      if ((target->type == box_key) && (obj->holes_count > 0)) {
        object holy_box;

        // avoid touching the original
        memcpy (&holy_box, target, sizeof (object));
        holy_box.holes = obj->holes;
        holy_box.holes_count = obj->holes_count;

        target->deleted = 1;

        generate_boxes (&holy_box);
      }
    }                           // if match
  }                             // for loop


  // Convert boxes to topholes.
  // Note that this will affect new boxes
  // created with the generate_hole() call

  if (obj->convert == tophole_key) {
    for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
      target = &GroupDefs[WorldDef].objs[i];
      if ((target->type == box_key) && !target->deleted &&
          match_object_name (matchName, target->name)) {
        target->type = tophole_key;

        generate_hole (target);

        target = &GroupDefs[WorldDef].objs[i];  // FIXME - list
        target->type = box_key;
        target->deleted = 1;
      }
    }                           // for loop
  }                             // if (obj->convert == tophole_key)


  COPY_VECTOR (Position, oldPosition);
  COPY_VECTOR (Size, oldSize);
  Rotation = oldRotation;
  DriveThrough = oldDriveThrough;
  ShootThrough = oldShootThrough;
  strcpy (GroupDepthName, oldGroupDepthName);

  if (matches == 0) {
    print_object_names ();
    error ("modify: couldn't find %s\n", matchName);
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
generate_delete (object * obj)
{
  int i;
  int matches = 0;
  char *name;
  object *target;

  if (obj->name[0] == DELIMITER) {  // absolute naming
    error ("delete can't use absolute naming.\n");
    return -1;
  }
  else {
    append_name (obj->name);
    name = GroupDepthName;
  }

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    target = &GroupDefs[WorldDef].objs[i];
    if (match_object_name (name, target->name)) {
      matches++;
      DEBUG5 ("Deleteing %s\n", target->name);
      target->deleted = 1;
      target->srcfile = obj->srcfile;
      target->srcline = obj->srcline;
    }
  }

  if (matches == 0) {
    print_object_names ();
    error ("delete: couldn't find %s\n", name);
    return -1;
  }

  if (obj->name[0] != DELIMITER) {
    remove_name (obj->name);
  }

  return 0;
}

/******************************************************************************/

static int
generate_undelete (object * obj)
{
  int i;
  int matches = 0;
  char *name;
  object *target;

  if (obj->name[0] == DELIMITER) {  // absolute naming
    error ("undelete can't use absolute naming.\n");
    return -1;
  }
  else {
    append_name (obj->name);
    name = GroupDepthName;
  }

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    target = &GroupDefs[WorldDef].objs[i];
    if (match_object_name (name, target->name)) {
      matches++;
      DEBUG5 ("UnDeleteing %s\n", target->name);
      target->deleted = 0;
      target->srcfile = obj->srcfile;
      target->srcline = obj->srcline;
    }
  }

  if (matches == 0) {
    print_object_names ();
    error ("undelete: couldn't find %s\n", name);
    return -1;
  }

  if (obj->name[0] != DELIMITER) {
    remove_name (obj->name);
  }

  return 0;
}

/******************************************************************************/

static int
generate_intersect (object * obj)
{
  int i, j;
  int firstrun;
  int firstmatch, firstmatch_tmp;
  int lastmatch, lastmatch_tmp;
  int count = 0;
  char *targets;
  char *clippers = NULL;
  object *target, *clipper;


  if ((obj->to_name[0] == DELIMITER) && (obj->from_name[0] == DELIMITER)) {
    error
        ("intersect target and clippers can't both be absolute names.\n");
    return -1;
  }

  if (obj->from_name[0] == DELIMITER) { // absolute naming
    clippers = obj->from_name + 1;
  }
  else {
    append_name (obj->from_name);
    clippers = strdup (GroupDepthName);
    if (clippers == NULL) {
      perror ("generate_intersect (strdup)");
      return -1;
    }
    remove_name (obj->from_name);
  }

  if (obj->to_name[0] == DELIMITER) {   // absolute naming
    targets = obj->to_name + 1;
  }
  else {
    append_name (obj->to_name);
    targets = GroupDepthName;
  }

  firstrun = 2;
  firstmatch = 0;
  firstmatch_tmp = -1;
  lastmatch = GroupDefs[WorldDef].objs_count;
  lastmatch_tmp = 0;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    int clip_match = 0;

    clipper = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (clipper) ||
        !match_object_name (clippers, clipper->name)) {
      continue;
    }

    if (match_object_name (targets, clipper->name)) {
      clip_match = 1;
    }

    if (firstrun == 2) {
      firstrun = 1;
    }
    else if (firstrun == 1) {
      if (firstmatch_tmp != -1)
        firstmatch = firstmatch_tmp;
      else
        firstmatch = 0;
      lastmatch = lastmatch_tmp;
      firstrun = 0;

      DEBUG3 ("generate_intersect(firstrun) first=%i last=%i\n",
              firstmatch, lastmatch);
    }

    for (j = firstmatch; j < lastmatch; j++) {
      target = &GroupDefs[WorldDef].objs[j];
      if (!solid_core_object (target) ||
          !match_object_name (targets, target->name) ||
          (clip_match && match_object_name (clippers, target->name))) {
        continue;
      }

      if (firstrun == 1) {
        if (firstmatch_tmp == -1)
          firstmatch_tmp = j;

        lastmatch_tmp = j + 1;
      }

      if (intersect_objs (target, clipper)) {
        DEBUG5 ("Clipper: deleteing %s\n", target->name);
        count++;
        target->deleted = 1;
        target->srcfile = current_file ();
        target->srcline = current_line_number ();
      }
    }
  }

  DEBUG2 ("CLIPPED %i objects (%s against %s)\n", count, targets,
          clippers);

  if (obj->to_name[0] != DELIMITER) {
    remove_name (obj->to_name);
  }

  if (clippers != (obj->from_name + 1)) {
    free (clippers);
  }

  return 0;
}

/******************************************************************************/

static int
generate_hole (object * obj)
{
  int i;
  object *box;
  int current_count;
  hole_def hole, tmphole;
  double holerot, rotadj;
  char matchName[LINE_LENGTH] = "";
  vec3 oldPosition;
  vec3 oldSize;
  double oldRotation;
  int oldDriveThrough;
  int oldShootThrough;
  char oldGroupDepthName[LINE_LENGTH] = "";

  if (obj->type == tophole_key) {
    strcpy (matchName, "*");
  }
  else {
    append_name ("*");          // match all objects in the local group
    strcpy (matchName, GroupDepthName);
    remove_name ("*");
  }
  DEBUG3 ("Generating hole: %s, (matchName = %s)\n", obj->name, matchName);

  strcpy (oldGroupDepthName, GroupDepthName);

  COPY_VECTOR (oldPosition, Position);
  COPY_VECTOR (oldSize, Size);
  oldRotation = Rotation;
  oldDriveThrough = DriveThrough;
  oldShootThrough = ShootThrough;


  calc_position (Position, Size, Rotation, Tilt, obj->pos, hole.pos);
  calc_rotation (Size, obj->rotation, Rotation, &holerot);

  hole.size[0] = fabs (obj->size[0] * Size[0]);
  hole.size[1] = fabs (obj->size[1] * Size[1]);
  hole.size[2] = obj->size[2] * Size[2];
  if (hole.size[2] < 0) {
    hole.pos[2] = hole.pos[2] + hole.size[2];
    hole.size[2] = -(hole.size[2]);
  }


  /*
   * Save the top level state, the objects being operated
   * upon have already been generated, and shouldn't be
   * adjusted by the local group state.
   */
  COPY_VECTOR (Position, Zeroes);
  COPY_VECTOR (Size, Ones);
  Rotation = 0.0;
  GroupDepthName[0] = '\0';
  DriveThrough = 0;
  ShootThrough = 0;


  /*
   * If holes are added, then the total number of objects increases.
   * Using GroupDefs[WorldDef].objs_count then becomes unsafe.
   */
  current_count = GroupDefs[WorldDef].objs_count;

  for (i = 0; i < current_count; i++) {
    object holy_box;

    box = &GroupDefs[WorldDef].objs[i];

    if ((box->type != box_key) || (box->deleted)) {
      continue;
    }

    if (!IS_ZERO (fmod (holerot - box->rotation, 90.0))) {
      continue;
    }

    if (!match_object_name (matchName, box->name)) {
      continue;
    }

    if (!box_to_box (box->pos, box->size, box->rotation,
                     hole.pos, hole.size, holerot)) {
      continue;
    }


    DEBUG3 ("  Adding hole to box (%s)\n", box->name);

    center_square (box->pos, box->rotation, hole.pos, holerot,
                   tmphole.pos, &rotadj);
    tmphole.pos[2] = hole.pos[2] - box->pos[2];
    tmphole.size[2] = hole.size[2];

    if (!IS_ZERO (fmod (rotadj, 180.0))) {
      tmphole.size[1] = hole.size[0];
      tmphole.size[0] = hole.size[1];
    }
    else {
      tmphole.size[0] = hole.size[0];
      tmphole.size[1] = hole.size[1];
    }

    // avoid touching the original
    memcpy (&holy_box, box, sizeof (object));
    holy_box.holes = &tmphole;
    holy_box.holes_count = 1;

    box->deleted = 1;           // delete the old

    generate_boxes (&holy_box);
  }                             // for loop

  COPY_VECTOR (Position, oldPosition);
  COPY_VECTOR (Size, oldSize);
  Rotation = oldRotation;
  DriveThrough = oldDriveThrough;
  ShootThrough = oldShootThrough;

  strcpy (GroupDepthName, oldGroupDepthName);

  return 0;
}

/******************************************************************************/

static int
match_object_name (const char *string, const char *objname)
{
  if (string == NULL)
    return 0;

  if (objname == NULL)
    return 0;

  if ((string[0] == MATCH_MULTI) && (string[1] == '\0')) {
    DEBUG9 ("match_object_name --> %s || %s (instant match)\n",
            string, objname);
    return 1;
  }


  while (*string != '\0') {
    DEBUG9 ("match_object_name --> %s || %s\n", string, objname);

    if (*string == MATCH_MULTI) {
      string++;
      switch (match_multi (&string, &objname)) {
        case +1:
          return 1;
        case -1:
          return 0;
      }
    }
    else if (*objname == '\0') {
      return 0;
    }
    else if ((*string == MATCH_SINGLE) || (*string == *objname)) {
      string++;
      objname++;
    }
    else {
      return 0;
    }
  }

  if (*objname == '\0') {
    return 1;
  }
  else {
    return 0;
  }
}

/******************************************************************************/

static int
match_multi (const char **string, const char **objname)
{
  const char *str;
  const char *obj;
  const char *strtop;
  const char *objtop;

  str = *string;
  obj = *objname;

  while ((*str != '\0') && (*str == MATCH_MULTI))
    str++;                      /* get rid of multiple '*'s */

  if (*str == '\0') {           /* '*' was last, auto-match */
    return +1;
  }

  strtop = str;
  objtop = obj;


  while (*str != '\0') {
    if (*str == MATCH_MULTI) {
      *string = str;
      *objname = obj;
      return 0;                 /* matched this segment */
    }
    else if (*obj == '\0') {
      return -1;                /* can't match */
    }
    else {
      if ((*str == MATCH_SINGLE) || (*str == *obj)) {
        str++;
        obj++;
        if ((*str == '\0') && (*obj != '\0')) { /* advanced check */
          obj++;
          objtop++;
          obj = objtop;
          str = strtop;
        }
      }
      else {
        obj++;
        objtop++;
        obj = objtop;
        str = strtop;
      }
    }
  }

  *string = str;
  *objname = obj;
  return +1;                    /* full match */
}

/******************************************************************************/

static void
calc_position (vec3 gpos, vec3 gsize, double grot, double tilt,
               vec3 ipos, vec3 opos)
{
  double cos_val, tcos_val;
  double sin_val, tsin_val;
  double stmp[3], ttmp[3], rtmp[3];

  // scale, tilt, rotate, scale, and then translate
  cos_val = cos (M_PI * grot / 180.0);
  sin_val = sin (M_PI * grot / 180.0);
  tcos_val = cos (M_PI * tilt / 180.0);
  tsin_val = sin (M_PI * tilt / 180.0);

  // scale  
  stmp[0] = ipos[0] * gsize[0];
  stmp[1] = ipos[1] * gsize[1];
  stmp[2] = ipos[2] * gsize[2];

  // rotate
  rtmp[0] = (cos_val * stmp[0]) - (sin_val * stmp[1]);
  rtmp[1] = (cos_val * stmp[1]) + (sin_val * stmp[0]);
  rtmp[2] = stmp[2];
  
  // tilt  
  ttmp[0] = rtmp[0];
  ttmp[1] = (tcos_val * rtmp[1]) - (tsin_val * rtmp[2]);
  ttmp[2] = (tcos_val * rtmp[2]) + (tsin_val * rtmp[1]);

  // translate
  opos[0] = gpos[0] + ttmp[0];
  opos[1] = gpos[1] + ttmp[1];
  opos[2] = gpos[2] + ttmp[2];

  return;
}

/******************************************************************************/

static void
calc_rotation (const vec3 size, const double objrot,
               const double staterot, double *outrot)
{
  if (size[0] * size[1] > 0.0) {
    *outrot = staterot + objrot;
  }
  else {
    *outrot = staterot - objrot;
  }

  return;
}

/******************************************************************************/

static int
matrix_pop (void)
{
  TransformMatrix *ptr;
  
  if (MatrixCount <= 0) {
    error ("Matrix stack underflowed\n");
    return -1;
  }
  
  MatrixCount--;
  memcpy (TheMatrix, MatrixStack[MatrixCount], sizeof (TransformMatrix));

  ptr = (TransformMatrix *) realloc (MatrixStack,
                                     MatrixCount * sizeof (TransformMatrix));
  if (ptr == NULL) {
    perror ("matrix_pop(realloc)");
    return -1;
  }
  else {
    MatrixStack = ptr;
  }
    
  return 0;
}

static int
matrix_push (void)
{
  TransformMatrix *ptr;
  MatrixCount++;
  ptr = (TransformMatrix *) realloc (MatrixStack,
                                     MatrixCount * sizeof (TransformMatrix));
  if (ptr == NULL) {
    perror ("matrix_push(realloc)");
    MatrixCount--;
    return -1;
  }
  else {
    MatrixStack = ptr;
  }
  
  memcpy (MatrixStack[MatrixCount - 1], TheMatrix,
          sizeof (TransformMatrix));
          
  return 0;
}

static void
matrix_ident (void)
{
  static TransformMatrix IdentityMatrix = {
    {1.0, 0.0, 0.0, 0.0},
    {0.0, 1.0, 0.0, 0.0},
    {0.0, 0.0, 1.0, 0.0},
    {0.0, 0.0, 0.0, 1.0}
  };
  memcpy (TheMatrix, IdentityMatrix, sizeof (TransformMatrix));
  return;
}

static void
matrix_shift (vec3 pos)
{
  TheMatrix[0][3] = TheMatrix[0][3] + pos[0];
  TheMatrix[1][3] = TheMatrix[1][3] + pos[1];
  TheMatrix[2][3] = TheMatrix[2][3] + pos[2];
  return;
}

static void
matrix_scale (vec3 size)
{
  TheMatrix[0][0] = TheMatrix[0][0] * size[0];
  TheMatrix[1][1] = TheMatrix[1][1] * size[1];
  TheMatrix[2][2] = TheMatrix[2][2] * size[2];
  return;
}

static void
matrix_rotate (double angle, vec3 normal)
{
  return;
}

static void
matrix_mult (vec3 output, const vec3 input)
{
  float tmp[3];
  
  tmp[0] = (input[0] * TheMatrix[0][0]) +
           (input[1] * TheMatrix[0][1]) +
           (input[2] * TheMatrix[0][2]) + TheMatrix[0][3];
  tmp[1] = (input[0] * TheMatrix[1][0]) +
           (input[1] * TheMatrix[1][1]) +
           (input[2] * TheMatrix[1][2]) + TheMatrix[1][3];
  tmp[2] = (input[0] * TheMatrix[2][0]) +
           (input[1] * TheMatrix[2][1]) +
           (input[2] * TheMatrix[2][2]) + TheMatrix[2][3];
              
  memcpy (output, tmp, sizeof(float[3]));
  
  return;
}

/******************************************************************************/

static inline int
core_object (object * obj)
{
  switch (obj->type) {
    case box_key:
    case pyr_key:
    case tetra_key:
    case base_key:
    case tele_key:
    case link_key:
    case weapon_key:
    case zone_key:
      return 1;
    default:
      return 0;
  }
}

/******************************************************************************/

static inline int
solid_core_object (object * obj)
{
  int type = obj->type;
  return (((type == box_key) || (type == pyr_key) || (type == tele_key) ||
           (type == base_key) || (type == tetra_key)) && (obj->deleted == 0));
}

/******************************************************************************/

static inline int
composite_object (object * obj)
{
  int type = obj->type;
  return ((type == arc_key) || (type == group_key) || (type == paths_key)
          || (type == steps_key) || (type == text_key)
          || (type == image_key));
}

/******************************************************************************/

static int
last_object_solid (void)
{
  int i;
  object *obj;

  /*
   * Check the last object type added to the 
   * ShadowDef objects.
   */
  i = GroupDefs[ShadowDef].objs_count - 1;
  obj = &GroupDefs[ShadowDef].objs[i];

  if (!solid_core_object (obj) && !composite_object (obj)) {
    error ("improper object type (%s)\n", keyword_string (obj->type));
    return -1;
  }

  return 0;
}

/******************************************************************************/

static int
intersect_objs (object * obj1, object * obj2)
{
  object mod1, mod2;
  int type1, type2;
  int result = 0;

  DEBUG9 ("intersect: testing [%s VS. %s]\n", obj1->name, obj2->name);

  if (obj1->deleted || obj2->deleted) {
    return 0;
  }

  type1 = obj1->type;
  type2 = obj2->type;

  if (type1 == tele_key) {      /* modify teleporter shape */
    COPY_VECTOR (mod1.pos, obj1->pos);
    mod1.rotation = obj1->rotation;
    mod1.size[0] = obj1->border * 0.5;
    mod1.size[1] = obj1->size[1] + (obj1->border * 2.0);
    if (obj1->size[2] < 0.0)
      mod1.size[2] = obj1->size[2] - obj1->border;
    else
      mod1.size[2] = obj1->size[2] + obj1->border;
    obj1 = &mod1;
  }

  if (type2 == tele_key) {      /* modify teleporter shape */
    COPY_VECTOR (mod2.pos, obj2->pos);
    mod2.rotation = obj2->rotation;
    mod2.size[0] = obj2->border * 0.5;
    mod2.size[1] = obj2->size[1] + (obj2->border * 2.0);
    if (obj2->size[2] < 0.0)
      mod2.size[2] = obj2->size[2] - obj2->border;
    else
      mod2.size[2] = obj2->size[2] + obj2->border;
    obj2 = &mod2;
  }

  if ((type1 == base_key) || (type1 == tele_key))
    type1 = box_key;

  if ((type2 == base_key) || (type2 == tele_key))
    type2 = box_key;


  if (type1 == box_key) {
    if (type2 == box_key) {
      result = box_to_box (obj1->pos, obj1->size, obj1->rotation,
                           obj2->pos, obj2->size, obj2->rotation);
    }
    else if (type2 == pyr_key) {
      result = box_to_pyr (obj1->pos, obj1->size, obj1->rotation,
                           obj2->pos, obj2->size, obj2->rotation);
    }
  }

  else if (type1 == pyr_key) {
    if (type2 == pyr_key) {
      result = pyr_to_pyr (obj1->pos, obj1->size, obj1->rotation,
                           obj2->pos, obj2->size, obj2->rotation);
    }
    else if (type2 == box_key) {
      result = box_to_pyr (obj2->pos, obj2->size, obj2->rotation,
                           obj1->pos, obj1->size, obj1->rotation);
    }
  }

  return result;
}

/******************************************************************************/

static double
exists_func (parameter * params)
{
  int i;
  object *obj;
  const char *name;

  name = params[0].string;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    obj = &GroupDefs[WorldDef].objs[i];
    if (match_object_name (name, obj->name) && !obj->deleted) {
      return +1.0;
    }
  }

  return -1.0;
}

/******************************************************************************/

static double
clip_last (parameter * params)
{
  int i, j;
  object *obj1, *obj2;

  params = params;              /* lose the warnings */

  if (LastObjectNumber < 0)
    return -1.0;

  if (last_object_solid () == -1)
    return -1.0;

  for (i = LastObjectNumber; i < GroupDefs[WorldDef].objs_count; i++) {
    obj1 = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (obj1)) {
      continue;
    }

    for (j = 0; j < LastObjectNumber; j++) {
      obj2 = &GroupDefs[WorldDef].objs[j];
      if (!solid_core_object (obj2)) {
        continue;
      }

      if (intersect_objs (obj1, obj2)) {
        nuke_objects (LastObjectNumber);
        nuke_shadow ();
        LastObjectNumber = -1;
        return +1.0;
      }
    }
  }

  return -1.0;
}

/******************************************************************************/

static double
clip_against (parameter * params)
{
  int i, j;
  int firstmatch, firstmatch_tmp;
  int lastmatch, lastmatch_tmp;
  object *obj1, *obj2;
  const char *clippers;


  if (LastObjectNumber < 0)
    return -1.0;

  if (last_object_solid () == -1)
    return -1.0;

  clippers = params[0].string;


  firstmatch = 0;
  lastmatch = LastObjectNumber;
  firstmatch_tmp = -1;
  lastmatch_tmp = 0;

  for (i = LastObjectNumber; i < GroupDefs[WorldDef].objs_count; i++) {

    if (i == (LastObjectNumber + 1)) {
      if (firstmatch_tmp != -1)
        firstmatch = firstmatch_tmp;
      else
        firstmatch = 0;
      lastmatch = lastmatch_tmp;
    }

    obj1 = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (obj1)) {
      continue;
    }

    for (j = firstmatch; j < lastmatch; j++) {

      if (!match_object_name (clippers, GroupDefs[WorldDef].objs[j].name))
        continue;

      if (i == LastObjectNumber) {
        if (firstmatch_tmp == -1)
          firstmatch_tmp = j;

        lastmatch_tmp = j + 1;
      }

      obj2 = &GroupDefs[WorldDef].objs[j];
      if (!solid_core_object (obj2)) {
        continue;
      }

      if (intersect_objs (obj1, obj2)) {
        nuke_objects (LastObjectNumber);
        nuke_shadow ();
        LastObjectNumber = -1;
        return +1.0;
      }
    }
  }

  return -1.0;
}

/******************************************************************************/

static double
shiftinc_last (parameter * params)
{
  double dx = params[0].number;
  double dy = params[1].number;
  double dz = params[2].number;
  object *obj1, *obj2, *tmpobj;
  int i, j, k;

  if (LastObjectNumber < 0)
    return -1.0;

  if (last_object_solid () == -1)
    return -1.0;

  if (IS_ZERO (dx) && IS_ZERO (dy) && IS_ZERO (dz)) {
    error ("ShiftInc() called with all 0's.\n");
    return -1.0;
  }

  for (i = LastObjectNumber; i < GroupDefs[WorldDef].objs_count; i++) {
    obj1 = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (obj1)) {
      continue;
    }

    for (j = 0; j < LastObjectNumber; j++) {
      int flag = 0;

      obj2 = &GroupDefs[WorldDef].objs[j];
      if (!solid_core_object (obj2)) {
        continue;
      }

      while (intersect_objs (obj1, obj2)) {
        flag = 1;
        for (k = LastObjectNumber; k < GroupDefs[WorldDef].objs_count; k++) {
          tmpobj = &GroupDefs[WorldDef].objs[k];
          tmpobj->pos[0] = tmpobj->pos[0] + dx;
          tmpobj->pos[1] = tmpobj->pos[1] + dy;
          tmpobj->pos[2] = tmpobj->pos[2] + dz;
        }
      }

      if (flag) {               /* restart the whole process */
        i = LastObjectNumber - 1;   /* the end of the FOR does i++ */
        break;
      }
    }
  }

  return +1.0;
}

/******************************************************************************/

static double
shiftinc_against (parameter * params)
{
  double dx = params[0].number;
  double dy = params[1].number;
  double dz = params[2].number;
  const char *obstacles = params[3].string;
  object *obj1, *obj2, *tmpobj;
  int i, j, k;

  if (LastObjectNumber < 0)
    return -1.0;

  if (last_object_solid () == -1)
    return -1.0;

  if (IS_ZERO (dx) && IS_ZERO (dy) && IS_ZERO (dz)) {
    error ("ShiftIncAgainst() called with all 0's.\n");
    return -1.0;
  }

  for (i = LastObjectNumber; i < GroupDefs[WorldDef].objs_count; i++) {
    obj1 = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (obj1)) {
      continue;
    }

    for (j = 0; j < LastObjectNumber; j++) {
      int flag = 0;

      obj2 = &GroupDefs[WorldDef].objs[j];
      if (!solid_core_object (obj2) ||
          !match_object_name (obstacles, obj2->name)) {
        continue;
      }

      while (intersect_objs (obj1, obj2)) {
        flag = 1;
        for (k = LastObjectNumber; k < GroupDefs[WorldDef].objs_count; k++) {
          tmpobj = &GroupDefs[WorldDef].objs[k];
          tmpobj->pos[0] = tmpobj->pos[0] + dx;
          tmpobj->pos[1] = tmpobj->pos[1] + dy;
          tmpobj->pos[2] = tmpobj->pos[2] + dz;
        }

      }

      if (flag) {               /* restart the whole process */
        i = LastObjectNumber - 1;   /* the end of the FOR loop does i++ */
        break;
      }
    }
  }

  return +1.0;
}

/******************************************************************************/

static double
ztop (parameter * params)
{
  params[5].string = "*";

  return ztop_against (params);
}

/******************************************************************************/

static double
ztop_against (parameter * params)
{
  vec3 pos;
  vec3 size;
  double rot;
  const char *obstacles;
  object *obj;
  object *maxobj = NULL;
  int i;
  double current_ztop;
  double ztop_val = 0.0;        /* ground is the lowest check */

  pos[0] = params[0].number;
  pos[1] = params[1].number;
  size[0] = params[2].number;
  size[1] = params[3].number;
  rot = params[4].number;
  obstacles = params[5].string;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    obj = &GroupDefs[WorldDef].objs[i];
    if (!solid_core_object (obj) ||
        !match_object_name (obstacles, obj->name)) {
      continue;
    }

    if (square_to_square (pos, size, rot,
                          obj->pos, obj->size, obj->rotation)) {

      if (obj->size[2] < 0.0) {
        current_ztop = obj->pos[2];
      }
      else if (obj->type == tele_key) {
        current_ztop = obj->pos[2] + obj->size[2] + obj->border;
      }
      else {
        current_ztop = obj->pos[2] + obj->size[2];
      }

      if (current_ztop > ztop_val) {
        ztop_val = current_ztop;
        maxobj = obj;
      }
    }
  }

  if (maxobj == NULL) {
    TopType = MAXZ_TYPE_GROUND;
    TopSupport = MAXZ_SUPPORT_FULL;
    return 0.0;
  }
  else {
    switch (maxobj->type) {
      case box_key:
        TopType = MAXZ_TYPE_BOX;
        break;
      case pyr_key:
        TopType = MAXZ_TYPE_PYR;
        break;
      case base_key:
        TopType = MAXZ_TYPE_BASE;
        break;
      case tele_key:
        TopType = MAXZ_TYPE_TELE;
        break;
      default:{
        error ("ztop() : internal error, unexpected type (%s)\n",
               keyword_string (maxobj->type));
      }
    }

    if ((maxobj->type == pyr_key) && (maxobj->size[2] > 0.0)) {
      TopSupport = MAXZ_SUPPORT_SLOPE;
    }
    else if (square_in_square (pos, size, rot, maxobj->pos,
                               maxobj->size, maxobj->rotation)) {
      TopSupport = MAXZ_SUPPORT_FULL;
    }
    else if (point_to_square (pos, maxobj->pos,
                              maxobj->size, maxobj->rotation)) {
      TopSupport = MAXZ_SUPPORT_CENTER;
    }
    else {
      TopSupport = MAXZ_SUPPORT_PARTIAL;
    }
  }

  return ztop_val;
}

/******************************************************************************/

static double
ztop_type (parameter * params)
{
  params = params;              /* get rid of warnings */
  return TopType;
}

/******************************************************************************/

static double
ztop_support (parameter * params)
{
  params = params;              /* get rid of warnings */
  return TopSupport;
}

/******************************************************************************/
/******************************************************************************/

static int
export_world (int grpdef)
{
  int i;
  object *obj;

  if (ConfigOptions != NULL) {
    if (make_options (ConfigOptions) != 0) {
      error ("error generating config options object.\n");
      return -1;
    }
  }

  if (FoundWorld) {
    if (make_world (WorldDimension, FlagHeight) != 0) {
      error ("error generating world object.\n");
      return -1;
    }
  }

  for (i = 0; i < GroupDefs[grpdef].objs_count; i++) {

    obj = &GroupDefs[grpdef].objs[i];

    if (obj->deleted) {
      continue;
    }

    export_adjust (obj);        /* flipz, appearance, etc... */

    switch (obj->type) {
      case box_key:{
        if (make_box (obj->name, obj->pos, obj->size, obj->rotation,
                      obj->drivethrough, obj->shootthrough) == -1)
          return -1;
        break;
      }
      case pyr_key:{
        if (make_pyr (obj->name, obj->pos, obj->size, obj->rotation,
                      obj->drivethrough, obj->shootthrough,
                      obj->flipz) == -1)
          return -1;
        break;
      }
      case tetra_key:{
        if (make_tetra (obj->name, (const vec3 *)obj->vertices, obj->visible,
                        obj->colored, (const unsigned char (*)[4]) obj->colors,
                        obj->drivethrough, obj->shootthrough) == -1)
          return -1;
        break;
      }
      case base_key:{
        if (make_base (obj->name, obj->pos, obj->size, obj->rotation,
                       obj->color,
                       obj->drivethrough, obj->shootthrough) == -1)
          return -1;
        break;
      }
      case tele_key:{
        if (make_tele (obj->name, obj->pos, obj->size, obj->rotation,
                       obj->border,
                       obj->drivethrough, obj->shootthrough) == -1)
          return -1;
        break;
      }
      case weapon_key:{
        if (make_weapon (obj->name, obj->pos, obj->size, obj->rotation,
                         obj->drivethrough, obj->shootthrough,
                         obj->initdelay, obj->delays, obj->delay_count,
                         obj->text) == -1)
          return -1;
        break;
      }
      case link_key:{
        if (make_link (obj->name, obj->from, obj->to) == -1)
          return -1;
        break;
      }
      case zone_key:{
        if (make_zone (obj->name, obj->pos, obj->size, obj->rotation,
                       obj->text, obj->teams, obj->safety) == -1)
          return -1;
        break;
      }
      default:{
        error ("export_world(): internal error (%s)\n",
               keyword_string (obj->type));
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
export_adjust (object * obj)
{
  // at least clean-up the rotation
  obj->rotation = fmod (obj->rotation, 360.0);  // -360 to +360
  if (obj->rotation > 180.0) {
    obj->rotation = obj->rotation - 360.0;
  }
  else if (obj->rotation < -180.0) {
    obj->rotation = obj->rotation + 360.0;
  }

  if (!solid_core_object (obj)) {
    return -1;
  }

  obj->size[0] = fabs (obj->size[0]);
  obj->size[1] = fabs (obj->size[1]);

  obj->flipz = 0;               // only size[2] matters right now

  if (obj->size[2] < 0.0) {

    if (OutSizeStyle == BZMAPPER_FLIPZ) {   // sizes are all positive
      obj->size[2] = -obj->size[2];
      obj->pos[2] = obj->pos[2] - obj->size[2];

      if (obj->type == pyr_key) {
        obj->flipz = 1;
      }
      else if (obj->type == tele_key) {
        obj->pos[2] = obj->pos[2] - obj->border;
      }
    }
    else {                      // make it look pretty for bzedit32
      if (obj->type == pyr_key) {
        obj->size[1] = -obj->size[1];   // looks solid
      }
      else if (obj->type != tele_key) { // leave teles alone
        obj->size[2] = -obj->size[2];
        obj->pos[2] = obj->pos[2] - obj->size[2];
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
remove_pyr_overlaps (void)
{
  int i, j;
  double factor;
  object *pyr, *box;

  /*
   * Test pyramids for overlaps 
   */

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {

    pyr = &GroupDefs[WorldDef].objs[i];
    if ((pyr->type != pyr_key) || (pyr->deleted)) {
      continue;
    }

    for (j = 0; j < GroupDefs[WorldDef].objs_count; j++) {

      box = &GroupDefs[WorldDef].objs[j];
      if ((box->type != box_key) || (box->deleted)) {
        continue;
      }

      if (check_pyr_overlap (pyr->pos, pyr->size, pyr->rotation,
                             box->pos, box->size, box->rotation,
                             &factor)) {

        if (factor < 0.0) {
          printf ("Found an enclosed pyramid: %s\n", pyr->name);
        }
        else {
          printf ("Found a pyramid overlap: %s vs. %s (%i, %i) [%f]\n",
                  pyr->name, box->name, i, j, factor);
          pyr->pos[2] = pyr->pos[2] + ((1.0 - factor) * pyr->size[2]);
          pyr->size[2] = pyr->size[2] * factor;
          pyr->size[1] = pyr->size[1] * factor;
          pyr->size[0] = pyr->size[0] * factor;
        }
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
remove_box_overlaps (void)
{
  int i, j, target;
  hole_def hole;
  object *box1, *box2, *tmpobj;

  /*
   *  safety
   */
  COPY_VECTOR (Position, Zeroes);
  COPY_VECTOR (Size, Ones);
  Rotation = 0.0;

  /*
   * Test boxes for overlaps 
   */

  for (i = 0; i < (GroupDefs[WorldDef].objs_count - 1); i++) {
    object holy_box;

    box1 = &GroupDefs[WorldDef].objs[i];
    if ((box1->type != box_key) || (box1->deleted)) {
      continue;
    }

    for (j = (i + 1); j < GroupDefs[WorldDef].objs_count; j++) {

      box2 = &GroupDefs[WorldDef].objs[j];
      if ((box2->type != box_key) || (box2->deleted)) {
        continue;
      }

      if (check_box_overlap (box1->pos, box1->size, box1->rotation,
                             box2->pos, box2->size, box2->rotation)) {
        double cos_val, sin_val;
        double tmpx, tmpy;

        if (fabs (box1->size[0] * box1->size[1]) >
            fabs (box2->size[0] * box2->size[1])) {
          target = j;
          tmpobj = box1;
          box1 = box2;
          box2 = tmpobj;
        }
        else {
          target = i;
        }

        printf ("Found a box overlap: %s vs. %s (%i, %i)\n",
                box1->name, box2->name, i, j);
        fflush (stdout);        // dbg
        if (get_debug () >= 4) {
          printf ("  pos1 (%f, %f, %f) size1 (%f, %f, %f) rot1 (%f)\n",
                  box1->pos[0], box1->pos[1], box1->pos[2],
                  box1->size[0], box1->size[1], box1->size[2],
                  box1->rotation);
          printf ("  pos2 (%f, %f, %f) size2 (%f, %f, %f) rot2 (%f)\n",
                  box2->pos[0], box2->pos[1], box2->pos[2],
                  box2->size[0], box2->size[1], box2->size[2],
                  box2->rotation);
          printf ("box1 = %s, target = %s\n", box1->name,
                  GroupDefs[WorldDef].objs[target].name);
        }

        tmpx = box2->pos[0] - box1->pos[0];
        tmpy = box2->pos[1] - box1->pos[1];
        cos_val = cos (-(box1->rotation) * (M_PI / 180.0));
        sin_val = sin (-(box1->rotation) * (M_PI / 180.0));
        hole.pos[0] = (cos_val * tmpx) - (sin_val * tmpy);
        hole.pos[1] = (cos_val * tmpy) + (sin_val * tmpx);
        hole.pos[2] = box2->pos[2] - box1->pos[2];

        if (box2->size[2] < 0.0) {
          hole.pos[2] = hole.pos[2] + box2->size[2];
        }
        hole.size[2] = fabs (box2->size[2]);

        if (IS_ZERO (fmod (box1->rotation - box2->rotation, 180.0))) {
          hole.size[0] = fabs (box2->size[0]);
          hole.size[1] = fabs (box2->size[1]);
        }
        else {
          hole.size[1] = fabs (box2->size[0]);
          hole.size[0] = fabs (box2->size[1]);
        }


        GroupDepthName[0] = '\0';

        // avoid touching the original
        memcpy (&holy_box, box1, sizeof (object));
        holy_box.holes = &hole;
        holy_box.holes_count = 1;

        box1->deleted = 1;      // delete the old

        generate_boxes (&holy_box);

        if (target == j) {
          i = i - 1;
        }
        break;
      }
    }
  }

  return 0;
}

/******************************************************************************/

#define MERGE_X (1 << 0)
#define MERGE_Y (1 << 1)
#define MERGE_Z (1 << 2)

static int
merge_boxes (void)
{
  positize_boxes ();

  while (merge_boxes2 (MERGE_Y) > 0);
  while (merge_boxes2 (MERGE_X) > 0);
  while (merge_boxes2 (MERGE_Z) > 0);
  while (merge_boxes2 (MERGE_X | MERGE_Y | MERGE_Z) > 0);

  return 0;
}

/******************************************************************************/

static int
merge_boxes2 (int direction)
{
  int i, j;
  int modcount = 0;             // number of modifications
  object *box1, *box2;
  vec3 adjpos;
  double adjrot;
  double sx2, sy2;
  double max, min, mid;

  for (i = 0; i < (GroupDefs[WorldDef].objs_count - 1); i++) {
    double sin_val, cos_val;

    box1 = &GroupDefs[WorldDef].objs[i];

    if ((box1->type != box_key) || (box1->deleted)) {
      continue;
    }

    sin_val = sin (M_PI * box1->rotation / 180.0);
    cos_val = cos (M_PI * box1->rotation / 180.0);

    for (j = (i + 1); j < GroupDefs[WorldDef].objs_count; j++) {
      int xalign = 0;
      int yalign = 0;
      int zalign = 0;

      box2 = &GroupDefs[WorldDef].objs[j];

      if ((box2->type != box_key) || (box2->deleted)) {
        continue;
      }

      if ((box1->drivethrough != box2->drivethrough) ||
          (box1->shootthrough != box2->shootthrough)) {
        continue;
      }

#ifdef MERGE_CONTAINED
      if ((box1->pos[2] <= box2->pos[2]) &&
          ((box1->pos[2] + box1->size[2]) >=
           (box2->pos[2] + box2->size[2])) &&
          square_in_square (box2->pos, box2->size, box2->rotation,
                            box1->pos, box1->size, box1->rotation)) {
        box2->deleted = 1;
        printf ("Merging boxes: %s was inside %s\n",
                box2->name, box1->name);
        modcount++;
        continue;
      }

      if ((box2->pos[2] <= box1->pos[2]) &&
          ((box2->pos[2] + box2->size[2]) >=
           (box1->pos[2] + box1->size[2])) &&
          square_in_square (box1->pos, box1->size, box1->rotation,
                            box2->pos, box2->size, box2->rotation)) {
        box1->deleted = 1;
        printf ("Merging boxes: %s was inside %s\n",
                box1->name, box2->name);
        modcount++;
        continue;
      }
#endif                          // MERGE_CONTAINED

      center_square (box1->pos, box1->rotation, box2->pos, box2->rotation,
                     adjpos, &adjrot);

      if (!IS_ZERO (fmod (adjrot, 90.0))) {
        continue;
      }

      if (!IS_ZERO (fmod (adjrot, 180.0))) {
        sx2 = box2->size[1];
        sy2 = box2->size[0];
      }
      else {
        sx2 = box2->size[0];
        sy2 = box2->size[1];
      }

      if (IS_ZERO (adjpos[0]) && IS_EQUAL (box1->size[0], sx2)) {
        xalign = 1;
      }
      if (IS_ZERO (adjpos[1]) && IS_EQUAL (box1->size[1], sy2)) {
        yalign = 1;
      }
      if (IS_EQUAL (box1->pos[2], box2->pos[2]) &&
          IS_EQUAL (box1->size[2], box2->size[2])) {
        zalign = 1;
      }


      if ((xalign + yalign + zalign) < 2) {
        continue;
      }

      // merging along the Y axis

      if ((direction & MERGE_Y) && (zalign && xalign) &&
          ((box1->size[1] + sy2) > fabs (adjpos[1]) - ACCURACY)) {

        if ((adjpos[1] + sy2) > box1->size[1]) {
          max = adjpos[1] + sy2;
        }
        else {
          max = box1->size[1];
        }
        if ((adjpos[1] - sy2) < -box1->size[1]) {
          min = adjpos[1] - sy2;
        }
        else {
          min = -box1->size[1];
        }

        box1->size[1] = (max - min) / 2.0;
        mid = (max + min) / 2.0;
        box1->pos[0] = box1->pos[0] - (mid * sin_val);
        box1->pos[1] = box1->pos[1] + (mid * cos_val);

        printf ("Merging boxes: %s Y-swallowed %s\n",
                box1->name, box2->name);

        box2->deleted = 1;
        modcount++;
        continue;
      }

      // merging along the X axis

      if ((direction & MERGE_X) && (zalign && yalign) &&
          ((box1->size[0] + sx2) >= fabs (adjpos[0]) - ACCURACY)) {

        if ((adjpos[0] + sx2) > box1->size[0]) {
          max = adjpos[0] + sx2;
        }
        else {
          max = box1->size[0];
        }
        if ((adjpos[0] - sx2) < -box1->size[0]) {
          min = adjpos[0] - sx2;
        }
        else {
          min = -box1->size[0];
        }

        box1->size[0] = (max - min) / 2.0;
        mid = (max + min) / 2.0;
        box1->pos[0] = box1->pos[0] + (mid * cos_val);
        box1->pos[1] = box1->pos[1] + (mid * sin_val);

        printf ("Merging boxes: %s X-swallowed %s\n",
                box1->name, box2->name);

        box2->deleted = 1;
        modcount++;
        continue;
      }

      // merging along the Z axis

      min = box1->pos[2] + (box1->size[2] / 2.0);
      max = box2->pos[2] + (box2->size[2] / 2.0);
      mid = (box1->size[2] + box2->size[2]) / 2.0;

      if ((direction & MERGE_Z) && (xalign && yalign) &&
          (mid >= fabs (min - max) - ACCURACY)) {

        if ((box2->pos[2] + box2->size[2]) >
            (box1->pos[2] + box1->size[2])) {
          max = box2->pos[2] + box2->size[2];
        }
        else {
          max = box1->pos[2] + box1->size[2];
        }
        if (box2->pos[2] < box1->pos[2]) {
          min = box2->pos[2];
        }
        else {
          min = box1->pos[2];
        }

        box1->size[2] = max - min;
        box1->pos[2] = min;

        printf ("Merging boxes: %s Z-swallowed %s\n",
                box1->name, box2->name);

        box2->deleted = 1;
        modcount++;
        continue;
      }
    }
  }

  return modcount;
}

/******************************************************************************/

static int
positize_boxes (void)
{
  int i;
  object *box;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    box = &GroupDefs[WorldDef].objs[i];

    if (box->type != box_key) {
      continue;
    }

    if (box->size[2] < 0.0) {
      box->pos[2] = box->pos[2] + box->size[2];
      box->size[2] = -(box->size[2]);
    }
  }

  return 0;
}

/******************************************************************************/

static int
resolve_links (void)
{
  int i;
  object *link;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    link = &GroupDefs[WorldDef].objs[i];
    if (link->type == link_key) {

      if (link->to != RANDOM_LINK) {
        if (resolve_link (link->to_name, &link->to) == -1) {
          error ("Couldn't resolve _to_ link: %s\n", link->to_name);
          print_tele_names ();
          return -1;
        }
        else {
          link->to = link->to + link->to_gate;
        }
      }

      if (resolve_link (link->from_name, &link->from) == -1) {
        error ("Couldn't resolve _from_ link: %s\n", link->from_name);
        print_tele_names ();
        return -1;
      }
      else {
        link->from = link->from + link->from_gate;
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
resolve_link (char *name, int *num)
{
  int i;
  object *tele;
  int count = 0;

  if (*num >= 0) {
    return 0;
  }

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    tele = &GroupDefs[WorldDef].objs[i];
    if (tele->type == tele_key) {
      if (strcmp (name, tele->name) == 0) {
        *num = 2 * count;
      }
      count++;
    }
  }

  if (*num == -1)
    return -1;
  else
    return 0;
}

/******************************************************************************/

static int
pre_order_link (void)
{
  int i;
  object *link;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    link = &GroupDefs[WorldDef].objs[i];

    if ((link->type == link_key) && !link->deleted) {
      if (link->from >= 0) {
        if (link_pos_to_id (&link->from) < 0) {
          printf ("broken absolute link: from %i (%s: line %i)\n",
                  link->from, top_filename (link->srcfile), link->srcline);
          link->deleted = 1;
        }
      }
      if ((link->to != RANDOM_LINK) && (link->to >= 0)) {
        if (link_pos_to_id (&link->to) < 0) {
          printf ("broken absolute link: to %i (%s: line %i)\n",
                  link->to, top_filename (link->srcfile), link->srcline);
          link->deleted = 1;
        }
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
post_order_link (void)
{
  int i;
  object *link;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    link = &GroupDefs[WorldDef].objs[i];

    if ((link->type == link_key) && !link->deleted) {
      if (link->from >= 0) {
        if (link_id_to_pos (&link->from) < 0) {
          return -1;
        }
      }
      if ((link->to != RANDOM_LINK) && (link->to >= 0)) {
        if (link_id_to_pos (&link->to) < 0) {
          return -1;
        }
      }
    }
  }

  return 0;
}

/******************************************************************************/

static int
link_pos_to_id (int *linknum)
{
  int i;
  object *tele = NULL;
  int tele_pos;
  int tele_gate;
  int tele_count;

  tele_pos = *linknum / 2;
  tele_gate = *linknum % 2;
  tele_count = -1;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    tele = &GroupDefs[WorldDef].objs[i];
    if ((tele->type == tele_key) && !tele->deleted) {
      tele_count++;
      if (tele_count == tele_pos) {
        break;
      }
    }
  }

  if (tele_count != tele_pos) {
    return -1;
  }

  *linknum = (tele->id * 2) + tele_gate;

  return 0;
}

/******************************************************************************/

static int
link_id_to_pos (int *linknum)
{
  int i;
  object *tele = NULL;
  int tele_id;
  int tele_gate;
  int tele_count;

  tele_id = *linknum / 2;
  tele_gate = *linknum % 2;
  tele_count = -1;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    tele = &GroupDefs[WorldDef].objs[i];
    if ((tele->type == tele_key) && !tele->deleted) {
      tele_count++;
      if (tele_id == tele->id) {
        break;
      }
    }
  }

  if (tele_id != tele->id) {
    error ("link_id_to_pos: internal error (link_id %i)\n", *linknum);
    return -1;
  }

  *linknum = (tele_count * 2) + tele_gate;

  return 0;
}

/******************************************************************************/

static int
add_group_def (const char *name)
{
  group_def *ptr;

  GroupDefsCount++;
  ptr =
      (group_def *) realloc (GroupDefs,
                             sizeof (group_def) * GroupDefsCount);

  if (ptr == NULL) {
    perror ("add_group_def(realloc)");
    GroupDefsCount--;
    return -1;
  }
  else {
    GroupDefs = ptr;
  }

  GroupDefs[GroupDefsCount - 1].objs = NULL;
  GroupDefs[GroupDefsCount - 1].objs_count = 0;
  GroupDefs[GroupDefsCount - 1].name = strdup (name);

  GroupDefs[GroupDefsCount - 1].group_count = 0;
  GroupDefs[GroupDefsCount - 1].box_count = 0;
  GroupDefs[GroupDefsCount - 1].pyr_count = 0;
  GroupDefs[GroupDefsCount - 1].tetra_count = 0;
  GroupDefs[GroupDefsCount - 1].base_count = 0;
  GroupDefs[GroupDefsCount - 1].tele_count = 0;
  GroupDefs[GroupDefsCount - 1].link_count = 0;
  GroupDefs[GroupDefsCount - 1].weapon_count = 0;
  GroupDefs[GroupDefsCount - 1].zone_count = 0;
  GroupDefs[GroupDefsCount - 1].arc_count = 0;
  GroupDefs[GroupDefsCount - 1].text_count = 0;
  GroupDefs[GroupDefsCount - 1].paths_count = 0;
  GroupDefs[GroupDefsCount - 1].steps_count = 0;
  GroupDefs[GroupDefsCount - 1].hole_count = 0;

  if (GroupDefs[GroupDefsCount - 1].name == NULL) {
    perror ("strdup");
    return -1;
  }

  if (GroupDefs[GroupDefsCount - 1].name == NULL) {
    perror ("add_group_def(strdup)\n");
    return -1;
  }

  return (GroupDefsCount - 1);
}

/******************************************************************************/

static int
find_group_def (const char *name)
{
  int i;

  for (i = 0; i < GroupDefsCount; i++) {
    if (strcmp (name, GroupDefs[i].name) == 0)
      return i;
  }

  return -1;
}

/******************************************************************************/

static int
add_object (int grpdef, object * obj)
{
  object *ptr;

  obj->id = GroupDefs[grpdef].objs_count;

  GroupDefs[grpdef].objs_count++;
  ptr = (object *) realloc (GroupDefs[grpdef].objs,
                            sizeof (object) *
                            GroupDefs[grpdef].objs_count);

  if (ptr == NULL) {
    perror ("add_object(realloc)");
    GroupDefs[grpdef].objs_count--;
    return -1;
  }
  else {
    GroupDefs[grpdef].objs = ptr;
  }

  memcpy (GroupDefs[grpdef].objs + (GroupDefs[grpdef].objs_count - 1),
          obj, sizeof (object));


  switch (obj->type) {
    case group_key:
      GroupDefs[grpdef].group_count++;
      break;
    case box_key:
      GroupDefs[grpdef].box_count++;
      break;
    case pyr_key:
      GroupDefs[grpdef].pyr_count++;
      break;
    case tetra_key:
      GroupDefs[grpdef].tetra_count++;
      break;
    case base_key:
      GroupDefs[grpdef].base_count++;
      break;
    case tele_key:
      GroupDefs[grpdef].tele_count++;
      break;
    case link_key:
      GroupDefs[grpdef].link_count++;
      break;
    case weapon_key:
      GroupDefs[grpdef].weapon_count++;
      break;
    case zone_key:
      GroupDefs[grpdef].zone_count++;
      break;
    case arc_key:
      GroupDefs[grpdef].arc_count++;
      break;
    case text_key:
      GroupDefs[grpdef].text_count++;
      break;
    case image_key:
      GroupDefs[grpdef].image_count++;
      break;
    case paths_key:
      GroupDefs[grpdef].paths_count++;
      break;
    case steps_key:
      GroupDefs[grpdef].steps_count++;
      break;
    case hole_key:
    case tophole_key:
      GroupDefs[grpdef].hole_count++;
      break;
  }

  DEBUG9 ("ADD_OBJECT: Added %s to %s\n",
          obj->name, GroupDefs[grpdef].name);

  return 0;
}

/******************************************************************************/

static void
free_object (object * obj)
{
  free (obj->name);
  free (obj->text);
  free (obj->to_name);
  free (obj->from_name);
  free (obj->delays);
  free (obj->holes);
  free (obj->points);

  return;
}

/******************************************************************************/

static void
uncount_object (int type, int grpdef)
{
  switch (type) {
    case group_key:
      GroupDefs[grpdef].group_count--;
      break;
    case box_key:
      GroupDefs[grpdef].box_count--;
      break;
    case pyr_key:
      GroupDefs[grpdef].pyr_count--;
      break;
    case tetra_key:
      GroupDefs[grpdef].tetra_count--;
      break;
    case base_key:
      GroupDefs[grpdef].base_count--;
      break;
    case tele_key:
      GroupDefs[grpdef].tele_count--;
      break;
    case link_key:
      GroupDefs[grpdef].link_count--;
      break;
    case weapon_key:
      GroupDefs[grpdef].weapon_count--;
      break;
    case zone_key:
      GroupDefs[grpdef].zone_count--;
      break;
    case arc_key:
      GroupDefs[grpdef].arc_count--;
      break;
    case text_key:
      GroupDefs[grpdef].text_count--;
      break;
    case image_key:
      GroupDefs[grpdef].image_count--;
      break;
    case paths_key:
      GroupDefs[grpdef].paths_count--;
      break;
    case steps_key:
      GroupDefs[grpdef].steps_count--;
      break;
    case hole_key:
    case tophole_key:
      GroupDefs[grpdef].hole_count--;
      break;
  }

  return;
}

/******************************************************************************/

static void
nuke_objects (int last_id)
{
  int i;
  object *obj;

  for (i = last_id; i < GroupDefs[WorldDef].objs_count; i++) {
    obj = &GroupDefs[WorldDef].objs[i];
    uncount_object (obj->type, WorldDef);
    free_object (obj);
  }
  GroupDefs[WorldDef].objs_count = last_id;

  return;
}

/******************************************************************************/

static void
nuke_shadow (void)
{
  int i;
  object *obj;

  i = GroupDefs[ShadowDef].objs_count - 1;
  obj = &GroupDefs[ShadowDef].objs[i];

  uncount_object (obj->type, ShadowDef);
  free_object (obj);
  GroupDefs[ShadowDef].objs_count = i;

  return;
}

/******************************************************************************/

static int
add_hole (object * obj, hole_def * hole)
{
  hole_def *ptr;

  obj->holes_count++;
  ptr = (hole_def *) realloc (obj->holes,
                              sizeof (hole_def) * obj->holes_count);

  if (ptr == NULL) {
    perror ("add_hole(realloc)");
    obj->holes_count--;
    return -1;
  }
  else {
    obj->holes = ptr;
  }

  memcpy (obj->holes + (obj->holes_count - 1), hole, sizeof (hole_def));

  return 0;
}

/******************************************************************************/

static int
add_point (object * obj, point_def * point)
{
  point_def *ptr;

  obj->points_count++;
  ptr = (point_def *) realloc (obj->points,
                               sizeof (point_def) * obj->points_count);

  if (ptr == NULL) {
    perror ("add_point(realloc)");
    obj->points_count--;
    return -1;
  }
  else {
    obj->points = ptr;
  }

  memcpy (obj->points + (obj->points_count - 1), point,
          sizeof (point_def));

  return 0;
}

/******************************************************************************/

static int
append_name (const char *name)
{
  if (name != NULL) {
    if ((strlen (name) + strlen (GroupDepthName) + 1) > LINE_LENGTH) {
      error ("append_name(): internal error, "
             "GroupDepthName length exceeded\n");
      return -1;
    }

    strcat (GroupDepthName, name);
  }

  return 0;
}

/******************************************************************************/

static int
remove_name (const char *name)
{
  int len1, len2;

  if (name != NULL) {
    len1 = strlen (GroupDepthName);
    len2 = strlen (name);

    GroupDepthName[len1 - len2] = '\0';
  }

  return 0;
}

/******************************************************************************/

static int
append_string (char **head, const char *tail)
{
  if (*head == NULL) {
    *head = strdup (tail);
    if (*head == NULL) {
      perror ("strdup");
      return -1;
    }
  }
  else {
    *head = (char *) realloc (*head, strlen (*head) + strlen (tail) + 1);
    if (*head == NULL) {
      perror ("realloc");
      return -1;
    }
    strcat (*head, tail);
  }

  return 0;
}

/******************************************************************************/

static void
free_all (void)
{
  int i, j;

  /*
   * free the group/object data
   */

  for (i = 0; i < GroupDefsCount; i++) {
    for (j = 0; j < GroupDefs[i].objs_count; j++) {
      free_object (&GroupDefs[i].objs[j]);
    }
    free (GroupDefs[i].objs);
    free (GroupDefs[i].name);
  }

  free (GroupDefs);
  GroupDefs = NULL;
  GroupDefsCount = 0;

  free (ConfigOptions);
  ConfigOptions = NULL;

  /*
   * Free all of the other modules
   */

  free_boxfont ();
  free_variables ();
  free_functions ();
  free_line_buffer ();
  free_flow_parser ();


  return;
}

/******************************************************************************/

extern int
stats_boxes (void)
{
  return StatsBoxes;
}

extern int
stats_pyramids (void)
{
  return StatsPyrs;
}

extern int
stats_tetras (void)
{
  return StatsTetras;
}

extern int
stats_bases (void)
{
  return StatsBases;
}

extern int
stats_teleporters (void)
{
  return StatsTeles;
}

extern int
stats_links (void)
{
  return StatsLinks;
}

extern int
stats_weapons (void)
{
  return StatsWeapons;
}

extern int
stats_zones (void)
{
  return StatsZones;
}


/******************************************************************************/
/******************************************************************************/

static int
calc_statistics (void)
{
  int i;

  StatsBoxes = 0;
  StatsPyrs = 0;
  StatsTetras = 0;
  StatsBases = 0;
  StatsTeles = 0;
  StatsLinks = 0;
  StatsWeapons = 0;
  StatsZones = 0;

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    if (GroupDefs[WorldDef].objs[i].deleted) {
      continue;
    }
    switch (GroupDefs[WorldDef].objs[i].type) {
      case box_key:
        StatsBoxes++;
        break;
      case pyr_key:
        StatsPyrs++;
        break;
      case tetra_key:
        StatsTetras++;
        break;
      case base_key:
        StatsBases++;
        break;
      case tele_key:
        StatsTeles++;
        break;
      case link_key:
        StatsLinks++;
        break;
      case weapon_key:
        StatsWeapons++;
        break;
      case zone_key:
        StatsZones++;
        break;
    }
  }

  return 0;
}

/******************************************************************************/

static int
world_order (const void *a, const void *b)
{
  const object *obja = (const object *) a;
  const object *objb = (const object *) b;

  if (obja->type > objb->type) {
    return +1;
  }
  else if (obja->type < objb->type) {
    return -1;
  }
  else {
    if ((obja->pos[2] + (obja->size[2] / 2.0)) <
        (objb->pos[2] + (objb->size[2] / 2.0))) {
      if (OrderZ == BZMAPPER_ASCENDING) {
        return +1;
      }
      else {
        return -1;
      }
    }
    else if ((obja->pos[2] + (obja->size[2] / 2.0)) >
             (objb->pos[2] + (objb->size[2] / 2.0))) {
      if (OrderZ == BZMAPPER_ASCENDING) {
        return -1;
      }
      else {
        return +1;
      }
    }
    else {
      if (obja->id > objb->id) {
        return +1;
      }
      else {
        return -1;
      }
    }
  }
}

/******************************************************************************/

static int
setup_static_vars (void)
{
  GroupDepthName[0] = '\0';
  LastObjectNumber = -1;
  TopType = MAXZ_INVALID;
  
  matrix_ident (); // load the identity matrix

  Position[0] = 0.0;
  Position[1] = 0.0;
  Position[2] = 0.0;
  Size[0] = 1.0;
  Size[1] = 1.0;
  Size[2] = 1.0;
  Rotation = 0.0;
  Tilt = 0.0;
  DriveThrough = 0;
  ShootThrough = 0;

  FoundWorld = 0;
  WorldDimension = WorldSize / 2.0;
  FlagHeight = FlagAltitude;

  StatsBoxes = -1;
  StatsPyrs = -1;
  StatsTetras = -1;
  StatsBases = -1;
  StatsTeles = -1;
  StatsLinks = -1;
  StatsWeapons = -1;
  StatsZones = -1;

  return 1;
}

/******************************************************************************/

static int
setup_expr_vars (void)
{
  int i;

  set_constant ("_BZMAPPER_VERSION_", BZMAPPER_VERSION_FLOAT);
  set_constant ("_DRIVETHROUGH_", 1.0);
  set_constant ("_SHOOTTHROUGH_", 1.0);
  set_constant ("_PASSABLE_", 1.0);
  set_constant ("_WEAPON_", 1.0);
  set_constant ("_ZONE_", 1.0);
  set_constant ("_OPTIONS_", 1.0);
  set_constant ("_TETRA_", 1.0);

  if (InSizeStyle == BZMAPPER_FLIPZ) {
    set_constant ("_FLIPZ_", 1.0);
  }
  else {
    set_constant ("_NEGZ_", 1.0);
  }

  for (i = 0; i < TEAM_NAMES; i++) {
    set_constant (TeamNames[i], (double) i);
  }

  set_constant ("PI", M_PI);

  set_constant ("BOX", MAXZ_TYPE_BOX);
  set_constant ("PYR", MAXZ_TYPE_PYR);
  set_constant ("BASE", MAXZ_TYPE_BASE);
  set_constant ("TELE", MAXZ_TYPE_TELE);
  set_constant ("GROUND", MAXZ_TYPE_GROUND);

  set_constant ("FULL", MAXZ_SUPPORT_FULL);
  set_constant ("CENTER", MAXZ_SUPPORT_CENTER);
  set_constant ("PARTIAL", MAXZ_SUPPORT_PARTIAL);
  set_constant ("SLOPE", MAXZ_SUPPORT_SLOPE);

  set_constant ("INVALID", MAXZ_INVALID);

  return 1;
}
/******************************************************************************/

static int
setup_expr_funcs (void)
{
  add_function ("Exists", "s", exists_func);
  add_function ("Clip", NULL, clip_last);
  add_function ("ClipAgainst", "s", clip_against);
  add_function ("ShiftInc", "fff", shiftinc_last);
  add_function ("ShiftIncAgainst", "fffs", shiftinc_against);
  add_function ("Top", "fffff", ztop);
  add_function ("TopAgainst", "fffffs", ztop_against);
  add_function ("TopType", NULL, ztop_type);
  add_function ("TopSupport", NULL, ztop_support);

  return 1;
}

/******************************************************************************/



/******************************************************************************/
/******************************************************************************/
/***                                                                        ***/
/***              T H E   M A I N   E N T R Y   P O I N T                   ***/
/***                                                                        ***/
/******************************************************************************/
/******************************************************************************/

extern int
load_map (const char *filename)
{
  int tmp;
  time_t t;

  DEBUG9 ("SIZEOF (object) = %i\n", (int) sizeof (object));

  t = time (NULL);
  my_srand ((unsigned int) t);

  clear_error ();               /* for debugging */

  init_boxfont (NULL);
  init_line_buffer ();
  init_flow_parser ();

  setup_static_vars ();
  setup_expr_vars ();
  setup_expr_funcs ();


  if (filename == NULL) {
    print_variables ();
    print_functions ();
    return -1;
  }


  WorldDef = add_group_def (WORLD_NAME);
  if (WorldDef == -1) {
    free_all ();
    return -1;
  }

  ShadowDef = add_group_def (SHADOW_NAME);
  if (ShadowDef == -1) {
    free_all ();
    return -1;
  }


  if (include_file (filename) == LINEBUF_ERROR) {
    free_all ();
    return -1;
  }

  /*************************
   * --------------------- *
   * The Main Parsing Loop *
   * --------------------- *
   *************************/

  do {
    tmp = parse_command (WorldDef);
  }
  while (tmp == PARSE_OK);



  if (tmp != PARSE_EOF) {
    if (tmp == PARSE_END) {
      error ("extra end\n");
    }
    free_all ();
    return -1;
  }

  DEBUG1 ("\n(%s) finished parsing.\n\n", filename);


  // Merging can cause overlaps, so do it first

  if (MergeBoxes) {
    merge_boxes ();
  }

  if (Overlaps) {
    remove_pyr_overlaps ();
    remove_box_overlaps ();
  }


  if (OrderZ != 0) {
    if (pre_order_link () < 0) {
      free_all ();
      return -1;
    }
    qsort (GroupDefs[WorldDef].objs,
           (unsigned int) GroupDefs[WorldDef].objs_count,
           sizeof (object), world_order);
    if (post_order_link () < 0) {
      free_all ();
      return -1;
    }
  }


  /*
   * This has to be done _after_ any ordering is done.
   * If not, the teleporter numbers will be incorrect.
   */
  if (resolve_links () != 0) {
    free_all ();
    return -1;
  }

  if (export_world (WorldDef) != 0) {
    error ("error exporting objects.\n");
    free_all ();
    return -1;
  }


  if (get_debug_objects ()) {
    print_objects ();
  }

  if (get_debug_teleporters ()) {
    print_tele_names ();
  }


  calc_statistics ();

  free_all ();


  return 0;
}


/******************************************************************************/
/******************************************************************************/

void
set_ignore_names (int value)
{
  IgnoreName = value;
  return;
}

/******************************************************************************/

void
set_input_size_style (int value)
{
  InSizeStyle = value;
  return;
}

/******************************************************************************/

void
set_output_size_style (int value)
{
  OutSizeStyle = value;
  return;
}

/******************************************************************************/

void
set_z_ordering (int order)
{
  if (order < 0)
    OrderZ = -1;
  else if (order > 0)
    OrderZ = +1;
  else
    OrderZ = 0;

  return;
}

/******************************************************************************/

void
set_overlaps (void)
{
  Overlaps = 1;
  return;
}

/******************************************************************************/

void
set_merge_boxes (void)
{
  MergeBoxes = 1;
  return;
}

/******************************************************************************/

void
set_convert_old (void)
{
  ConvertOld = 1;
  return;
}

/******************************************************************************/
/******************************************************************************/

static void
print_dbgobj (object * obj)
{
  int i;
  int type;
  double *v;

  type = obj->type;

  printf (INDENT3 "type:        %s\n", keyword_string (obj->type));
  printf (INDENT3 "name:        %s\n", obj->name);
  printf (INDENT3 "id:          %i\n", obj->id);
  printf (INDENT3 "srcfile:     %s\n", top_filename (obj->srcfile));
  printf (INDENT3 "srcline:     %i\n", obj->srcline);
  printf (INDENT3 "pointer:     %p\n", obj);

  if ((type == image_key) || (type == text_key))
    printf (INDENT3 "text:        %s\n", obj->text);

  if (type == group_key)
    printf (INDENT3 "grpdef:      %i\n", obj->grpdef);

  if ((type != link_key) && (type != delete_key) && (type != undelete_key)) {
    v = obj->pos;
    printf (INDENT3 "position:    %11f, %11f, %11f\n", v[0], v[1], v[2]);
    v = obj->size;
    printf (INDENT3 "size:        %11f, %11f, %11f\n", v[0], v[1], v[2]);
    printf (INDENT3 "rotation:    %f\n", obj->rotation);
  }

  if (type == tele_key)
    printf (INDENT3 "border:      %f\n", obj->border);

  if (type == base_key)
    printf (INDENT3 "color:       %i\n", obj->color);

  if ((type == arc_key) || (type == paths_key) || (type == steps_key))
    printf (INDENT3 "segments:    %i\n", obj->segments);

  if ((type == paths_key) || (type == steps_key))
    printf (INDENT3 "closedloop:  %i\n", obj->closedloop);

  if (type == link_key) {
    printf (INDENT3 "to_name:     %s\n", obj->to_name);
    printf (INDENT3 "from_name:   %s\n", obj->from_name);
    printf (INDENT3 "to:          %i\n", obj->to);
    printf (INDENT3 "from:        %i\n", obj->from);
    printf (INDENT3 "to_gate:     %i\n", obj->to_gate);
    printf (INDENT3 "from_gate:   %i\n", obj->from_gate);
  }

  for (i = 0; i < obj->holes_count; i++) {
    print_hole (&obj->holes[i], i);
  }

  for (i = 0; i < obj->points_count; i++) {
    print_point (&obj->points[i], i);
  }

  if (type == weapon_key) {
    printf (INDENT3 "wtype:       ");
    if (obj->text != NULL)
      printf ("%s\n", obj->text);
    else
      printf ("\n");

    printf (INDENT3 "initdelay:   %f\n", obj->initdelay);
    printf (INDENT3 "delays:     ");
    for (i = 0; i < obj->delay_count; i++) {
      printf (" %f", obj->delays[i]);
    }
    printf ("\n");
  }

  return;
}

/******************************************************************************/

static void
print_hole (hole_def * hole, int count)
{
  double *v;

  printf (INDENT3 "hole %i:\n", count);
  v = hole->pos;
  printf (INDENT4 "position:    %11f, %11f, %11f\n", v[0], v[1], v[2]);
  v = hole->size;
  printf (INDENT4 "size:        %11f, %11f, %11f\n", v[0], v[1], v[2]);

  return;
}

/******************************************************************************/

static void
print_point (point_def * point, int count)
{
  printf (INDENT3 "point %i:\n", count);
  printf (INDENT4 "x: %f\n", point->x);
  printf (INDENT4 "y: %f\n", point->y);

  return;
}

/******************************************************************************/

static void
print_objects (void)
{
  int i, j;

  printf ("\n");

  for (i = 0; i < GroupDefsCount; i++) {

    printf ("GroupDef (%i)\n", i);

    printf (INDENT "name:    %s\n", GroupDefs[i].name);
    printf (INDENT "objects: %i\n", GroupDefs[i].objs_count);

    /*
     * print the objects
     */
    for (j = 0; j < GroupDefs[i].objs_count; j++) {
      object *obj;

      obj = &GroupDefs[i].objs[j];

      printf (INDENT2 "object (%i)", j);
      if (obj->deleted)
        printf (" *DELETED*");
      printf ("\n");
      print_dbgobj (obj);
    }
  }

  printf ("\n");

  return;
}

/******************************************************************************/

static void
print_object_names (void)
{
  int i;

  printf ("\n");
  printf ("Object Names:\n");

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    printf (INDENT "%s\n", GroupDefs[WorldDef].objs[i].name);
  }

  printf ("\n");

  return;
}

/******************************************************************************/

static void
print_tele_names (void)
{
  int i;
  int count = 0;
  object *obj;

  printf ("\nTELEPORTER NAMES\n");

  for (i = 0; i < GroupDefs[WorldDef].objs_count; i++) {
    obj = &GroupDefs[WorldDef].objs[i];
    if (obj->type == tele_key) {
      printf (INDENT "Tele(%i): %s\n", count, obj->name);
      count++;
    }
  }

  return;
}

/******************************************************************************/
/******************************************************************************/
