/*
 *  E S C A N . C
 *
 *  Author:     Dave R (trepan)
 *  Date:       Feb 2003
 *
 *  A mathematical expression parser for the bzmapper program.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>

#include "escan.h"
#include "utils.h"


/*
 * #define ALLOW_FUNCTION_SHADOWS
 *
 * Lets a variable have the same name as a function.
 */


#ifdef _WIN32
#  include <float.h>
#  define FINITE(x) _finite(x)
#else
#  define FINITE(x) finite(x)
#endif

#define ESCAPE_CHAR (0x1B)


/*
 * Typedefs
 */

typedef double (*opfunc) (double x, double y);

typedef struct {
  unsigned char type;
  double value;
  int func;
  char *string;
} entry;

typedef struct {
  char *name;
  int id;
  unsigned int name_len;
  double value;
  int constant;
} variable;

typedef struct {
  char *name;
  unsigned int name_len;
  char *argtypes;
  int argcount;
  parsefunc func;
} function;


/*
 * Function Prototypes
 */

static int parse_subexprs (const char *c, char *buffer, int buflen);
static int parse_tokens (const char *c);
static int parse_value (const char *c, double *value);
static int eval (double *value);
static double eval_value (void);
static double eval_expr (double value1, int pull);
static int order (int op);

static int add_entry (entry * e);
static void free_stack (void);

static int set_var (const char *name, double value, int constant);
static int add_variable (const char *name, double value, int constant);
static int add_functions (void);
static int compare_var_name (const void *str, const void *var);
static int find_variable (const char *c);
static int find_token (const char *c);
static int find_function (const char *c);

static void print_stack (void);

static const char *parse_quote (const char *c, char *quote,
                                int forcequote, int noescapes);
static const char *parse_label (const char *c, char *label);
static const char *eat_white_space (const char *cing);
static int parse_hex_code (const char *c, const char **end);
static int parse_octal_code (const char *c, const char **end);
static int non_white_space (const char *c);

static double add_func (double x, double y);
static double sub_func (double x, double y);
static double mul_func (double x, double y);
static double div_func (double x, double y);
static double pow_func (double x, double y);
static double mod_func (double x, double);
static double or_func (double x, double y);
static double and_func (double x, double y);
static double gte_func (double x, double y);
static double lte_func (double x, double y);
static double gt_func (double x, double y);
static double lt_func (double x, double y);
static double equ_func (double x, double y);
static double nequ_func (double x, double y);
static double sin_func (parameter * params);
static double cos_func (parameter * params);
static double tan_func (parameter * params);
static double asin_func (parameter * params);
static double acos_func (parameter * params);
static double atan_func (parameter * params);
static double atan2_func (parameter * params);
static double sqrt_func (parameter * params);
static double rand_func (parameter * params);
static double abs_func (parameter * params);
static double floor_func (parameter * params);
static double defined_func (parameter * params);


enum keys_type {
  start_tok,                    /* parentheses */
  close_tok,

  separator_tok,                /* for multiple func args */

  add_tok,                      /* basic operators */
  sub_tok,
  mul_tok,
  div_tok,
  mod_tok,
  pow_tok,

  or_tok,                       /* logical operators */
  and_tok,

  gte_tok,                      /* comparison operators */
  lte_tok,
  gt_tok,
  lt_tok,
  equ_tok,
  nequ_tok,
  inv_tok,

  func_tok,                     /* functions */
  value_tok,                    /* numbers */
  string_tok,                   /* strings */

  end_tok                       /* end of tokens */
};


/*
 * String Constants
 */

static const char *token_strings[] = {
  "(",
  ")",

  ",",

  "+",
  "-",
  "*",
  "/",
  "%",
  "^",

  "||",
  "&&",

  ">=",                         /* must be before > and < */
  "<=",
  ">",
  "<",
  "==",
  "!=",                         /* must be before ! */
  "!",

  "F",                          /* for printing token types */
  "V",
  "S",

  "E"
};
#define TOKENS ((int)(sizeof(token_strings)/sizeof(*token_strings)) - 4)


/*
 * Arrays of functions
 */

static opfunc opfuncs[] = {
  add_func,
  sub_func,
  mul_func,
  div_func,
  mod_func,
  pow_func,
  or_func,
  and_func,
  gte_func,
  lte_func,
  gt_func,
  lt_func,
  equ_func,
  nequ_func,
};


static char sin_name[] = "sin";
static char cos_name[] = "cos";
static char tan_name[] = "tan";
static char asin_name[] = "asin";
static char acos_name[] = "acos";
static char atan_name[] = "atan";
static char atan2_name[] = "atan2";
static char sqrt_name[] = "sqrt";
static char rand_name[] = "rand";
static char abs_name[] = "abs";
static char floor_name[] = "floor";
static char defined_name[] = "defined";

static char DoubleString[] = "f";
static char TwoDoubleString[] = "ff";
static char StringString[] = "s";

static function DefFuncs[] = {
  {sin_name, 3, DoubleString, 1, sin_func},
  {cos_name, 3, DoubleString, 1, cos_func},
  {tan_name, 3, DoubleString, 1, tan_func},
  {asin_name, 4, DoubleString, 1, asin_func},
  {acos_name, 4, DoubleString, 1, acos_func},
  {atan_name, 4, DoubleString, 1, atan_func},
  {atan2_name, 5, TwoDoubleString, 2, atan2_func},
  {sqrt_name, 4, DoubleString, 1, sqrt_func},
  {rand_name, 4, DoubleString, 1, rand_func},
  {abs_name, 3, DoubleString, 1, abs_func},
  {floor_name, 5, DoubleString, 1, floor_func},
  {defined_name, 7, StringString, 1, defined_func}
};
#define FUNCTIONS ((int)(sizeof(DefFuncs)/sizeof(*DefFuncs)))


/*
 * Static Variables
 */

static entry *Stack = NULL;
static int StackCount = 0;

#define TYPE(x)         Stack[x].type
#define STACK_VALUE(x)  Stack[x].value
#define STACK_STRING(x) Stack[x].string

#define FUNC(x,y)        Funcs[Stack[x].func].func(y)
#define FUNC_ARGCOUNT(x) Funcs[Stack[x].func].argcount
#define FUNC_ARGTYPES(x) Funcs[Stack[x].func].argtypes

static variable *Vars = NULL;
static int VarsCount = 0;

static function *Funcs = DefFuncs;
static int FuncsCount = FUNCTIONS;

static int Pos = 0;
static int Error = 0;


/******************************************************************************/
/******************************************************************************/

static int
eval (double *value)
{
  Pos = 0;
  Error = 0;

  *value = eval_value ();

  if (Error != 0) {
    return -1;
  }
  else if (TYPE (Pos) == end_tok) {
    return 0;
  }

  *value = eval_expr (*value, -1);

  if (Error != 0) {
    return -1;
  }
  else if (TYPE (Pos) == end_tok) {
    return 0;
  }
  else {
    return -1;
  }
}

/******************************************************************************/

static double
eval_value (void)
{
  double value;

  if (Error != 0)
    return 0.0;


  if (TYPE (Pos) == value_tok) {
    value = STACK_VALUE (Pos);
    Pos++;
    return value;
  }
  else if ((TYPE (Pos) == sub_tok) || (TYPE (Pos) == add_tok) ||
           (TYPE (Pos) == inv_tok)) {
    if (TYPE (Pos + 1) == value_tok) {
      switch (TYPE (Pos)) {
        case sub_tok:
          value = -STACK_VALUE (Pos + 1);
          break;
        case add_tok:
          value = +STACK_VALUE (Pos + 1);
          break;
        default:               /* inv_tok */
          value = STACK_VALUE (Pos + 1);
          if (value > 0) {
            value = -1.0;
          }
          else {
            value = +1.0;
          }
          break;
      }

      Pos = Pos + 2;

      return value;
    }
    else if ((TYPE (Pos + 1) == start_tok) || (TYPE (Pos + 1) == func_tok)) {
      switch (TYPE (Pos)) {
        case sub_tok:
          Pos++;
          value = -eval_value ();
          break;
        case add_tok:
          Pos++;
          value = +eval_value ();
          break;
        default:               /* inv_tok */
          Pos++;
          value = eval_value ();
          if (value > 0) {
            value = -1.0;
          }
          else {
            value = +1.0;
          }
          break;
      }
      return value;
    }
    else {
      Error = -1;
      return 0.0;
    }
  }
  else if (TYPE (Pos) == start_tok) {
    Pos++;
    value = eval_value ();
    if (TYPE (Pos) == close_tok) {
      Pos++;
      return value;
    }
    else {
      value = eval_expr (value, -1);

      if (Error != 0) {
        return 0.0;
      }
      else if (TYPE (Pos) == close_tok) {
        Pos++;
        return value;
      }
      else {
        Error = -1;
        error ("missing ')'\n");
        return 0.0;
      }
    }
  }
  else if (TYPE (Pos) == func_tok) {
    parameter params[ESCAN_MAX_FUNC_PARAMS];
    int i;
    int argcount;
    int funcPos;

    funcPos = Pos;              /* store the function position */

    argcount = FUNC_ARGCOUNT (funcPos);

    Pos++;                      /* skip the function token */

    if (argcount == 0) {
      Pos++;
      if (TYPE (Pos) != close_tok) {
        error ("missing ')'\n");
        return -1;
      }
      Pos++;
      return FUNC (funcPos, NULL);
    }

    for (i = 0; i < argcount; i++) {

      Pos++;                    /* skip start_tok, or separator_tok */

      if (FUNC_ARGTYPES (funcPos)[i] == PARAM_TYPE_DOUBLE) {
        params[i].number = eval_value ();
      }
      else {

        if (TYPE (Pos) != string_tok) {
          error ("parsing error, expected a string for parameter %i\n", i);
          return -1;
        }

        params[i].string = STACK_STRING (Pos);
        Pos++;

        if (((TYPE (Pos) == separator_tok) && (i < (argcount - 1))) ||
            ((TYPE (Pos) == close_tok) && (i == (argcount - 1)))) {
          continue;
        }
        else {
          Error = -1;
          if (i < (argcount - 1))
            error ("missing ','\n");
          else
            error ("missing ')'\n");
          return 0.0;
        }
      }

      if (((TYPE (Pos) == separator_tok) && (i < (argcount - 1))) ||
          ((TYPE (Pos) == close_tok) && (i == (argcount - 1)))) {
        continue;
      }

      params[i].number = eval_expr (params[i].number, -1);

      if (Error != 0) {
        return 0.0;
      }
      else if (((TYPE (Pos) == separator_tok) && (i < (argcount - 1))) ||
               ((TYPE (Pos) == close_tok) && (i == (argcount - 1)))) {
        continue;
      }
      else {
        Error = -1;
        if (i < (argcount - 1))
          error ("missing ','\n");
        else
          error ("missing ')'\n");
        return 0.0;
      }
    }

    Pos++;
    value = FUNC (funcPos, params);

    return value;
  }
  else {
    Error = -1;
    return 0.0;
  }

  return value;
}

/******************************************************************************/

static double
eval_expr (double value1, int pull)
{
  int optype;
  double value2;
  int OldPos;


  if (Error != 0)
    return 0.0;


  while ((TYPE (Pos) != end_tok) && (TYPE (Pos) != close_tok) &&
         (TYPE (Pos) != separator_tok)) {

    optype = TYPE (Pos);
    OldPos = Pos;

    Pos++;
    value2 = eval_value ();

    if ((optype < 2) || (optype >= TOKENS)) {
      Error = -1;
      return 0.0;
    }
    DEBUG6 ("eval_expr: <old#%i: %s P%i>  ", OldPos,
            token_strings[optype], order (optype));
    DEBUG6 ("vs.  <new#%i %s P%i>\n", Pos,
            token_strings[TYPE (Pos)], order (TYPE (Pos)));


    if (order (optype) >= order (TYPE (Pos)))
      value1 = opfuncs[optype - 3] (value1, value2);
    else
      value1 =
          opfuncs[optype - 3] (value1, eval_expr (value2, order (optype)));


    /*
     * Pull back if the calling routine had a lower precedence,
     * give it a change to do its job.
     */

    if (pull >= order (TYPE (Pos))) {
      DEBUG6 ("eval_expr(): pullback\n");
      return value1;
    }
  }

  return value1;
}

/******************************************************************************/

static int
order (int op)
{
  switch (op) {
    case pow_tok:
    case inv_tok:
      return 7;
    case mul_tok:
    case div_tok:
    case mod_tok:
      return 6;
    case add_tok:
    case sub_tok:
      return 5;
    case gte_tok:
    case lte_tok:
    case gt_tok:
    case lt_tok:
      return 4;
    case equ_tok:
    case nequ_tok:
      return 3;
    case and_tok:
      return 2;
    case or_tok:
      return 1;
    case separator_tok:
    case close_tok:
    case end_tok:
      return 0;
    default:{
      error ("internal error: parser/order(%i)\n", op);
      return -1;
    }
  }
}

/******************************************************************************/
/******************************************************************************/


static double
add_func (double x, double y)
{
  DEBUG5 ("add:  %f + %f = %f\n", x, y, x + y);
  return x + y;
}
static double
sub_func (double x, double y)
{
  DEBUG5 ("sub:   %f - %f = %f\n", x, y, x - y);
  return x - y;
}
static double
mul_func (double x, double y)
{
  DEBUG5 ("mul:   %f * %f = %f\n", x, y, x * y);
  return x * y;
}
static double
div_func (double x, double y)
{
  DEBUG5 ("div:   %f / %f = %f\n", x, y, x / y);
  return x / y;
}

static double
pow_func (double x, double y)
{
  DEBUG5 ("pow:   %f ^ %f = %f\n", x, y, pow (x, y));
  return pow (x, y);
}

static double
mod_func (double x, double y)
{
  DEBUG5 ("mod:  %f %% %f = %f\n", x, y, fmod (x, y));
  return fmod (x, y);
}

static double
or_func (double x, double y)
{
  double result;

  if ((x >= 0.0) || (y >= 0.0)) {
    result = +1.0;
  }
  else {
    result = -1.0;
  }

  DEBUG5 ("or:   %f || %f = %f\n", x, y, result);
  return result;
}

static double
and_func (double x, double y)
{
  double result;

  if ((x >= 0.0) && (y >= 0.0))
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("and:   %f && %f = %f\n", x, y, result);
  return result;
}

static double
gte_func (double x, double y)
{
  double result;

  if (x >= y)
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("gte:   %f >= %f = %f\n", x, y, result);
  return result;
}

static double
lte_func (double x, double y)
{
  double result;

  if (x <= y)
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("lte:   %f <= %f = %f\n", x, y, result);
  return result;
}

static double
gt_func (double x, double y)
{
  double result;

  if (x > y)
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("gt:   %f > %f = %f\n", x, y, result);
  return result;
}

static double
lt_func (double x, double y)
{
  double result;

  if (x < y)
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("lt:   %f < %f = %f\n", x, y, result);
  return result;
}

static double
equ_func (double x, double y)
{
  double result;

  if (IS_EQUAL (x, y))
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("equ:   %f == %f = %f\n", x, y, result);
  return result;
}

static double
nequ_func (double x, double y)
{
  double result;

  if (!IS_EQUAL (x, y))
    result = +1.0;
  else
    result = -1.0;

  DEBUG5 ("equ:   %f == %f = %f\n", x, y, result);
  return result;
}

/******************************************************************************/

static double
sin_func (parameter * params)
{
  double result;
  params[0].number = params[0].number * 2.0 * M_PI / 360.0;
  result = sin (params[0].number);
  DEBUG5 ("sin:   sin(%f) = %f\n", params[0].number, result);
  return result;
}

static double
cos_func (parameter * params)
{
  double result;
  params[0].number = params[0].number * 2.0 * M_PI / 360.0;
  result = cos (params[0].number);
  DEBUG5 ("cos:   cos(%f) = %f\n", params[0].number, result);
  return result;
}

static double
tan_func (parameter * params)
{
  double result;
  params[0].number = params[0].number * 2.0 * M_PI / 360.0;
  result = tan (params[0].number);
  DEBUG5 ("tan:   tan(%f) = %f\n", params[0].number, result);
  return result;
}

static double
asin_func (parameter * params)
{
  double result;
  result = asin (params[0].number) * 360.0 / (M_PI * 2.0);
  DEBUG5 ("asin:  asin(%f) = %f\n", params[0].number, result);
  return result;
}

static double
acos_func (parameter * params)
{
  double result;
  result = acos (params[0].number) * 360.0 / (M_PI * 2.0);
  DEBUG5 ("acos:  acos(%f) = %f\n", params[0].number, result);
  return result;
}

static double
atan_func (parameter * params)
{
  double result;
  result = atan (params[0].number) * 360.0 / (M_PI * 2.0);
  DEBUG5 ("atan:  atan(%f) = %f\n", params[0].number, result);
  return result;
}

static double
atan2_func (parameter * params)
{
  double result;
  result =
      atan2 (params[0].number, params[1].number) * 360.0 / (M_PI * 2.0);
  DEBUG5 ("atan2:  atan2(%f,%f) = %f\n",
          params[0].number, params[1].number, result);
  return result;
}

static double
sqrt_func (parameter * params)
{
  double result;
  result = sqrt (params[0].number);
  DEBUG5 ("sqrt:  sqrt(%f) = %f\n", params[0].number, result);
  return result;
}

static double
rand_func (parameter * params)
{
  double result;
  result = params[0].number * (double) my_rand () / (double) MY_RAND_MAX;
  DEBUG5 ("rand:  rand(%f) = %f\n", params[0].number, result);
  return result;
}

static double
abs_func (parameter * params)
{
  double result;
  result = fabs (params[0].number);
  DEBUG5 ("abs:   abs(%f) = %f\n", params[0].number, result);
  return result;
}

static double
floor_func (parameter * params)
{
  double result;
  result = floor (params[0].number);
  DEBUG5 ("floor: floor(%f) = %f\n", params[0].number, result);
  return result;
}

static double
defined_func (parameter * params)
{
  double result;
  if (find_variable (params[0].string) == -1)
    result = -1.0;
  else
    result = +1.0;
  DEBUG5 ("defined: defined(%s) = %f\n", params[0].string, result);
  return result;
}


/******************************************************************************/
/******************************************************************************/

static int
find_token (const char *c)
{
  int i;

  for (i = 0; i < TOKENS; i++) {
    if (strncmp (token_strings[i], c, strlen (token_strings[i])) == 0)
      return i;
  }

  return -1;
}

/******************************************************************************/

static int
find_function (const char *c)
{
  int i;

  for (i = 0; i < FuncsCount; i++) {
    if (strncmp (Funcs[i].name, c, Funcs[i].name_len) == 0) {
      const char *cptr;
      cptr = c + strlen (Funcs[i].name);

      while ((cptr != '\0') && isspace (*cptr))
        cptr++;

      if (strncmp (token_strings[start_tok], cptr,
                   strlen (token_strings[start_tok])) == 0) {
        return i;
      }
    }
  }

  return -1;
}

/******************************************************************************/

static int
compare_var_name (const void *str, const void *var)
{
  const variable *v = (const variable *) var;
  const char *s = (const char *) str;
  int tmp;
  int len;

  len = v->name_len;
  tmp = strncmp (v->name, s, (unsigned) len);

  if (tmp == 0) {
    s = s + len;

    if ((*s == '\0') || isspace (*s) || (find_token (s) != -1))
      return 0;
    else
      return +1;
  }

  return -tmp;
}

/******************************************************************************/

static int
find_variable (const char *c)
{
  variable *var;

  var = (variable *)
      bsearch (c, Vars, (unsigned) VarsCount, sizeof (variable),
               compare_var_name);

  if (var == NULL)
    return -1;
  else
    return (var - Vars);
}

/******************************************************************************/

static int
parse_value (const char *c, double *value)
{
  char *cptr;

  *value = strtod (c, &cptr);

  if (cptr != c) {
    return (cptr - c);
  }
  else {
    return -1;
  }

}

/******************************************************************************/

static int
parse_subexprs (const char *c, char *buffer, int buflen)
{
  int value;
  int length;
  char *buf;
  char valstr[LINE_LENGTH];

  buf = buffer;

  while (*c != '\0') {
    if ((*c == '\\') && (*(c + 1) == '(')) {
      if (escan (c + 1, "bin", &value, &c) != 1)
        return -1;
      sprintf (valstr, "%i", value);
      length = strlen (valstr);
      if (((buf - buffer) + length) >= buflen) {
        error ("parse_subexprs() buffer overflow\n");
        return -1;
      }
      else {
        strcpy (buf, valstr);
        buf = buf + length;
      }
    }
    else {
      *buf = *c;
      c++;
      buf++;
    }
  }
  *buf = 0;

  return 0;
}

/******************************************************************************/

static int
parse_tokens (const char *c)
{
  int tmp;
  entry e;
  char quote[LINE_LENGTH];
  char buffer[LINE_LENGTH];


  if (parse_subexprs (c, buffer, LINE_LENGTH) < 0)
    return -1;

  c = buffer;


  while (*c != '\0') {

    e.string = NULL;

    while ((*c != '\0') && isspace (*c))
      c++;

    if (isdigit (*c) || (*c == '.')) {
      tmp = parse_value (c, &e.value);
      if (tmp > 0) {
        e.type = value_tok;
        if (add_entry (&e) == -1)
          return -1;
        c = c + tmp;
      }
      else {
        error ("bad value: %s\n", c);
        return -1;
      }
    }
    else if (isalpha (*c) || (*c == '_')) {
      if ((tmp = find_function (c)) >= 0) {
        e.type = func_tok;
        e.func = tmp;
        if (add_entry (&e) == -1)
          return -1;

        c = c + Funcs[tmp].name_len;
      }
      else if ((tmp = find_variable (c)) >= 0) {
        e.value = Vars[tmp].value;
        e.type = value_tok;
        if (add_entry (&e) == -1)
          return -1;

        c = c + Vars[tmp].name_len;
      }
      else {
        error ("bad identifier: %s\n", c);
        return -1;
      }
    }
    else if (*c == '"') {
      c = parse_quote (c, quote, 1, 1);
      if (c == NULL)
        return -1;
      e.type = string_tok;
      e.string = strdup (quote);
      if (e.string == NULL) {
        perror ("parse_tokens (strdup)");
        return -1;
      }
      if (add_entry (&e) == -1)
        return -1;
    }
    else if ((tmp = find_token (c)) >= 0) {
      e.type = tmp;
      if (add_entry (&e) == -1)
        return -1;

      c = c + strlen (token_strings[tmp]);
    }
    else {
      error ("unknown identifier: %s\n", c);
      return -1;
    }
  }

  e.type = end_tok;
  add_entry (&e);

  return 0;
}

/******************************************************************************/

static int
add_entry (entry * e)
{
  entry *ptr;

  StackCount++;
  ptr = (entry *) realloc (Stack, sizeof (entry) * StackCount);

  if (ptr == NULL) {
    perror ("add_entry(realloc)");
    StackCount--;
    return -1;
  }
  else {
    Stack = ptr;
  }

  Stack[StackCount - 1].type = e->type;

  if (e->type == func_tok)
    Stack[StackCount - 1].func = e->func;
  else if (e->type == value_tok)
    Stack[StackCount - 1].value = e->value;

  Stack[StackCount - 1].string = e->string;

  return 0;
}

/******************************************************************************/

static void
free_stack (void)
{
  int i;

  for (i = 0; i < StackCount; i++) {
    free (Stack[i].string);
  }
  free (Stack);
  Stack = NULL;
  StackCount = 0;

  return;
}

/******************************************************************************/

static int
add_variable (const char *name, double value, int constant)
{
  int i;
  variable vtmp;
  variable *ptr;

  VarsCount++;
  ptr = (variable *) realloc (Vars, sizeof (variable) * VarsCount);

  if (ptr == NULL) {
    perror ("add_variable(realloc)");
    VarsCount--;
    return -1;
  }
  else {
    Vars = ptr;
  }

  Vars[VarsCount - 1].name = strdup (name);
  Vars[VarsCount - 1].name_len = strlen (name);
  Vars[VarsCount - 1].value = value;
  Vars[VarsCount - 1].constant = constant;
  Vars[VarsCount - 1].id = VarsCount - 1;

  if (Vars[VarsCount - 1].name == NULL) {
    perror ("add_variable(strdup)\n");
    return -1;
  }

  /*
   * Sort the variables by name. bsearch() is
   * used to find variables later on.
   * FIXME - use a binary search here.
   */

  for (i = 0; i < (VarsCount - 1); i++) {
    if (strcmp (Vars[i].name, Vars[VarsCount - 1].name) > 0) {
      memcpy (&vtmp, &Vars[VarsCount - 1], sizeof (variable));
      memmove (&Vars[i + 1], &Vars[i],
               (VarsCount - i - 1) * sizeof (variable));
      memcpy (&Vars[i], &vtmp, sizeof (variable));
      break;
    }
  }

  return 0;
}

/******************************************************************************/

extern int
set_variable (const char *name, double value)
{
  return set_var (name, value, 0);
}


extern int
set_constant (const char *name, double value)
{
  return set_var (name, value, 1);
}


static int
set_var (const char *name, double value, int constant)
{
  int var;

  var = find_variable (name);

  if (var == -1) {
    int i;

    for (i = 1; i < (int) strlen (name); i++) {
      if (!isalnum (name[i]) && !(name[i] == '_')) {
        error ("bad variable name: %s\n", name);
        return -1;
      }
    }

#ifndef ALLOW_FUNCTION_SHADOWS
    for (i = 0; i < FuncsCount; i++) {
      if (strcmp (Funcs[i].name, name) == 0) {
        error ("variable name mirrors function name: %s\n", name);
        return -1;
      }
    }
#endif

    if (add_variable (name, value, constant) != 0)
      return -1;
  }
  else {
    char end;

    /*
     * find_variable() allows tokens at the end of the string
     */
    end = name[Vars[var].name_len];
    if ((end != '\0') && (end != ' ')) {
      error ("bad variable name: %s\n", name);
      return -1;
    }

    if (Vars[var].constant) {
      error ("trying to redefine a constant %s\n", Vars[var].name);
      return -1;
    }
    else if (constant) {
      error ("trying to redefine \"%s\" as constant\n", Vars[var].name);
      return -1;
    }

    Vars[var].value = value;
    DEBUG9 ("redefined %s: %f\n", Vars[var].name, value);
  }

  return 0;
}

/******************************************************************************/

extern int
add_function (const char *name, const char *argtypes, parsefunc func)
{
  int i;
  int argcount;
  const char *c;
  function *ptr;
  const char *nullstr = "";

  if (Funcs == DefFuncs) {
    Funcs = NULL;
    FuncsCount = 0;
    if (add_functions () == -1)
      return -1;
  }

  if (argtypes == NULL)
    argtypes = nullstr;

  argcount = 0;
  c = argtypes;
  while (*c != '\0') {
    if ((*c != PARAM_TYPE_DOUBLE) && (*c != PARAM_TYPE_STRING)) {
      error ("Internal error: bad parameter type (%c)\n", *c);
      return -1;
    }
    c++;
    argcount++;
  }

  if (argcount > ESCAN_MAX_FUNC_PARAMS) {
    error ("Internal error: too many parameters for add_function\n");
    return -1;
  }

  for (i = 1; i < (int) strlen (name); i++) {
    if (!isalnum (name[i]) && !(name[i] == '_')) {
      error ("internal error: bad function name: %s\n", name);
      return -1;
    }
  }

  for (i = 0; i < FuncsCount; i++) {
    if ((strncmp (Funcs[i].name, name, Funcs[i].name_len) == 0) &&
        (name[Funcs[i].name_len] == '\0')) {
      error ("internal error: function already added: %s\n", name);
      return -1;
    }
  }

  FuncsCount++;
  ptr = (function *) realloc (Funcs, sizeof (function) * FuncsCount);

  if (ptr == NULL) {
    perror ("add_function(realloc)");
    FuncsCount--;
    return -1;
  }
  else {
    Funcs = ptr;
  }

  ptr = &Funcs[FuncsCount - 1];

  ptr->name = strdup (name);
  ptr->name_len = strlen (name);
  ptr->argtypes = strdup (argtypes);
  ptr->argcount = argcount;
  ptr->func = func;

  if (ptr->name == NULL) {
    perror ("add_function(strdup)\n");
    return -1;
  }

  if (ptr->argtypes == NULL) {
    perror ("add_function(strdup)\n");
    free (ptr->name);
    return -1;
  }

  if (get_debug () >= 8) {
    for (i = 0; i < FUNCTIONS; i++)
      if (DefFuncs[i].func == func)
        break;
    if (i == FUNCTIONS)
      print_functions ();
  }

  return 0;
}

static int
add_functions (void)
{
  int i;

  for (i = 0; i < FUNCTIONS; i++) {
    if (add_function (DefFuncs[i].name, DefFuncs[i].argtypes,
                      DefFuncs[i].func) != 0)
      return -1;
  }

  return 0;
}

/******************************************************************************/

extern void
free_variables (void)
{
  int i;

  for (i = 0; i < VarsCount; i++) {
    free (Vars[i].name);
  }

  free (Vars);
  Vars = NULL;
  VarsCount = 0;

  return;
}

/******************************************************************************/

extern void
free_functions (void)
{
  int i;

  if (Funcs == DefFuncs)
    return;

  for (i = 0; i < FuncsCount; i++) {
    fflush (stdout);
    free (Funcs[i].name);
    free (Funcs[i].argtypes);
  }

  free (Funcs);
  Funcs = DefFuncs;
  FuncsCount = 0;

  return;
}

/******************************************************************************/
/******************************************************************************/

extern int
evaluate_string (const char *c, double *value)
{
  int retval;

  while ((*c != '\0') && isspace (*c))
    c++;

  if (*c == '\0') {
    error ("empty string\n");
    return -1;
  }

  if (parse_tokens (c) == 0) {

    if (get_debug () >= 8) {
      print_variables ();
      print_stack ();
    }

    retval = eval (value);

    DEBUG4 (INDENT "evaluate_string() -> \"%s\" = %f\n", c, *value);

    if (retval != 0) {
      /*
       * pass through 
       */
    }
    else if (!FINITE (*value)) {
      error ("not a finite value: %f\n", *value);
      retval = -1;
    }
  }
  else {
    retval = -1;
  }

  free_stack ();

  return retval;
}

/******************************************************************************/
/******************************************************************************/

static void
print_stack (void)
{
  int i;

  printf (INDENT2 "function count: %i\n", FuncsCount);
  printf (INDENT2 "tokens:\n");

  for (i = 0; i < StackCount; i++) {
    if (TYPE (i) < TOKENS)
      printf (INDENT3 "token %i:  %s\n", i, token_strings[Stack[i].type]);
    else if (TYPE (i) == func_tok)
      printf (INDENT3 "token %i:  %s\n", i, Funcs[Stack[i].func].name);
    else if (TYPE (i) == value_tok)
      printf (INDENT3 "value %i:  %f\n", i, Stack[i].value);
    else if (TYPE (i) == string_tok)
      printf (INDENT3 "string %i: %s\n", i, Stack[i].string);
    else if (TYPE (i) == end_tok)
      printf (INDENT3 "token %i:  end\n", i);
    else
      printf (INDENT3 "Bad token: %i\n", TYPE (i));
  }

  return;
}

/******************************************************************************/

extern void
print_variables (void)
{
  int i;

  printf (INDENT2 "variables:\n");

  for (i = 0; i < VarsCount; i++) {
    printf (INDENT3 "name: %-23svalue: %f", Vars[i].name, Vars[i].value);
    if (Vars[i].constant)
      printf ("  [const]\n");
    else
      printf ("\n");
  }


  return;
}

/******************************************************************************/

extern void
print_functions (void)
{
  int i;

  printf (INDENT2 "functions:\n");

  for (i = 0; i < FuncsCount; i++) {
    printf (INDENT3 "name: %-23sargs: %2i -> %s\n", Funcs[i].name,
            Funcs[i].argcount, Funcs[i].argtypes);
  }

  return;
}

/******************************************************************************/
/******************************************************************************/

int
escan (const char *c, const char *format, ...)
{
  va_list ap;
  char buffer[LINE_LENGTH];
  int i;
  int count = 0;
  const char *copy = c;

  int whitespace = 0;           /* check for trailing whitespace only */
  int closebrace = 0;           /* finish expressions on the last brace */
  int noescapes = 0;            /* disallow octal/hex codes in quotes */

  int *iptr;
  double dval;
  double *dptr;
  char *cptr;


  va_start (ap, format);

  while (*format != '\0') {

    if ((*format == 'f') || (*format == 'i')) {

      while ((*c != '\0') && isspace (*c))
        c++;

      if (*c == '(') {
        int depth = 0;
        i = 0;
        while (*c != '\0') {

          if (*c == '(')
            depth++;
          else if (*c == ')')
            depth--;

          buffer[i] = *c;
          c++;
          i++;

          if ((depth == 0) && ((*c == '\0') || isspace (*c) || closebrace))
            break;
        }
        buffer[i] = '\0';

        if (depth != 0) {
          error ("missing bracket: %s\n", copy);
          va_end (ap);
          return -1;
        }
      }
      else {
        i = 0;
        while ((*c != '\0') && !isspace (*c)) {
          buffer[i] = *c;
          c++;
          i++;
        }
        buffer[i] = '\0';
      }


      if (evaluate_string (buffer, &dval) != 0) {
        error ("expression parsing error: %s\n", copy);
        va_end (ap);
        return -1;
      }


      if (*format == 'f') {
        dptr = (double *) va_arg (ap, double *);
        *dptr = dval;
      }
      else {
        iptr = (int *) va_arg (ap, int *);
        *iptr = (int) dval;
      }

      count++;
    }
    else if (*format == 'n') {
      const char **cpptr;
      cpptr = (const char **) va_arg (ap, const char **);
      *cpptr = c;
    }
    else if (*format == 's') {
      cptr = (char *) va_arg (ap, char *);
      c = parse_label (c, cptr);

      if (c == NULL) {
        va_end (ap);
        return count;
      }
      count++;
    }
    else if (*format == 'Q') {
      cptr = (char *) va_arg (ap, char *);
      c = parse_quote (c, cptr, 1, noescapes);

      if (c == NULL) {
        va_end (ap);
        return count;
      }
      count++;
    }
    else if (*format == 'q') {
      cptr = (char *) va_arg (ap, char *);
      c = parse_quote (c, cptr, 0, noescapes);

      if (c == NULL) {
        va_end (ap);
        return count;
      }
      count++;
    }
    else if (*format == 'w') {
      whitespace = 1;
    }
    else if (*format == 'b') {
      closebrace = 1;
    }
    else if (*format == 'e') {
      noescapes = 1;
    }
    else {
      error ("internal error: escan(\"%c\")\n", *format);
      va_end (ap);
      return -1;
    }

    format++;                   /* move to the next argument */

  }                             /* end while loop */

  va_end (ap);

  if (whitespace && (non_white_space (c)))
    return -1;

  return count;
}

/******************************************************************************/

static const char *
parse_label (const char *c, char *label)
{
  int value;
  const char *ctmp;

  while ((*c != '\0') && isspace (*c))
    c++;

  ctmp = c;

  while ((*c != '\0') && !isspace (*c)) {
    if ((*c == '\\') && (*(c + 1) == '(')) {
      if (escan (c + 1, "bin", &value, &c) != 1)
        return NULL;
      sprintf (label, "%i", value);
      label = label + strlen (label);
    }
    else {
      *label = *c;
      label++;
      c++;
    }
  }

  *label = '\0';

  if (c == ctmp) {
    return NULL;
  }
  else {
    return c;
  }
}

/******************************************************************************/

static const char *
parse_quote (const char *c, char *quote, int forcequote, int noescapes)
{
  int tmp;
  double dval;
  int i = 0;
  int quoted = 0;

  c = eat_white_space (c);

  if (*c != '"') {
    if (forcequote) {
      error ("expecting \"string\"\n");
      return NULL;
    }
  }
  else {
    quoted = 1;
    c++;
  }

  /*
   * Allow the use of escape characters in the quote
   */
  while (*c != '\0') {

    if (*c == '\\') {

      c++;                      /* lose the escape character */

      switch (*c) {
        case '\0':
          break;
        case '"':
          quote[i] = *c;
          break;
        case '\\':
          quote[i] = '\\';
          break;
        case 'n':
          quote[i] = '\n';
          break;
        case 'r':
          quote[i] = '\r';
          break;
        case 't':
          quote[i] = '\t';
          break;
        case 'e':
          if (noescapes)
            quote[i] = '*';
          else
            quote[i] = ESCAPE_CHAR;
          break;
        case 'v':
          quote[i] = '\v';
          break;
        case 'f':
          quote[i] = '\f';
          break;
        case 'b':
          quote[i] = '\b';
          break;
        case 'x':
          tmp = parse_hex_code (c, &c);
          if (tmp < 0)
            return NULL;
          quote[i] = tmp;
          if (noescapes) {
            quote[i] = '*';
          }
          break;
        case '(':
          if (escan (c, "bfn", &dval, &c) != 1)
            return NULL;
          sprintf_double (quote + i, dval);
          i = i + strlen (quote + i) - 1;
          c--;
          break;
        case 'N':              /* normal */
          strcpy (quote + i, "\033[0m");
          i = i + 3;
          break;
        case 'H':              /* highlight (bright) */
          strcpy (quote + i, "\033[1m");
          i = i + 3;
          break;
        case 'U':              /* underline */
          strcpy (quote + i, "\033[4m");
          i = i + 3;
          break;
        case 'D':              /* dark (black) */
          strcpy (quote + i, "\033[30m");
          i = i + 4;
          break;
        case 'R':              /* red */
          strcpy (quote + i, "\033[31m");
          i = i + 4;
          break;
        case 'G':              /* green */
          strcpy (quote + i, "\033[32m");
          i = i + 4;
          break;
        case 'Y':              /* yellow */
          strcpy (quote + i, "\033[33m");
          i = i + 4;
          break;
        case 'B':              /* blue */
          strcpy (quote + i, "\033[34m");
          i = i + 4;
          break;
        case 'P':              /* purple (magenta) */
          strcpy (quote + i, "\033[35m");
          i = i + 4;
          break;
        case 'C':              /* cyan */
          strcpy (quote + i, "\033[36m");
          i = i + 4;
          break;
        case 'W':              /* white */
          strcpy (quote + i, "\033[37m");
          i = i + 4;
          break;
        default:
          if ((*c >= '0') && (*c <= '7')) {
            tmp = parse_octal_code (c, &c);
            if (tmp < 0)
              return NULL;
            quote[i] = tmp;
            if (noescapes) {
              quote[i] = '*';
            }
          }
          else {
            error ("Unknown escape character (%c|0x04%X)\n", *c, *c);
            return NULL;
          }
      }
    }
    else if ((*c == '"') && quoted) {
      break;
    }
    else if (isspace (*c) && !quoted) {
      break;
    }
    else {
      if (noescapes && !isprint (*c))
        quote[i] = '*';
      else
        quote[i] = *c;
    }

    c++;
    i++;
  }

  quote[i] = '\0';

  if (quoted && (*c != '"')) {
    error ("expecting \"string\"\n");
    return NULL;
  }

  return (c + 1);
}

/******************************************************************************/

static int
parse_hex_code (const char *c, const char **end)
{
  int i, count = 0;
  unsigned char num[2];

  c++;                          /* lose the 'x' */

  for (i = 0; i < 2; i++) {
    if ((*c >= '0') && (*c <= '9')) {
      num[i] = *c - '0';
      count++;
    }
    else if ((*c >= 'A') && (*c <= 'F')) {
      num[i] = *c - 'A' + 0x0A;
      count++;
    }
    else if ((*c >= 'a') && (*c <= 'f')) {
      num[i] = *c - 'a' + 0x0A;
      count++;
    }
    else {
      break;
    }

    c++;
  }

  *end = c - 1;                 /* last user character */

  if (count == 0) {
    error ("incomplete hexadecimal escape code\n");
    return -1;
  }
  else if (count == 1) {
    return num[0];
  }
  else if (count == 2) {
    return ((num[0] * 0x10) + num[1]);
  }

  return -1;                    /* backup */
}

/******************************************************************************/

static int
parse_octal_code (const char *c, const char **end)
{
  int i, count = 0;
  unsigned char num[3];

  for (i = 0; i < 3; i++) {
    if ((*c >= '0') && (*c <= '7')) {
      num[i] = *c - '0';
      count++;
    }
    else {
      break;
    }
    c++;
  }

  *end = c - 1;                 /* last user character */

  if (count == 0) {
    error ("incomplete octal escape code\n");
    return -1;
  }
  else if (count == 1) {
    return num[0];
  }
  else if (count == 2) {
    return ((num[0] * 8) + num[1]);
  }
  else if (count == 3) {
    return ((num[0] * 64) + (num[1] * 8) + num[2]);
  }

  return -1;                    /* backup */
}

/******************************************************************************/

static int
non_white_space (const char *c)
{
  c = eat_white_space (c);
  if (*c != '\0') {
    error ("extra junk\n");
    return 1;
  }

  return 0;
}

/******************************************************************************/

static const char *
eat_white_space (const char *c)
{
  while ((*c != '\0') && isspace (*c))
    c++;

  return c;
}

/******************************************************************************/

#define MAX_DIGITS      (6)
#define MAX_PRECISION   (0.0000005)

extern void
sprintf_double (char *c, double value)
{
  int i;
  double d, m;
  char format[] = "%.?f";       /* '?' in format[3] */

  d = 1.0;

  for (i = 0; i < MAX_DIGITS; i++) {
    m = fabs (fmod (value, d));
    if ((m < MAX_PRECISION) || (fabs (m - d) < MAX_PRECISION))
      break;
    d = d / 10.0;
  }

  format[2] = '0' + i;

  sprintf (c, format, value);

  return;
}

/******************************************************************************/
/******************************************************************************/
