
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "flowctrl.h"
#include "linebuf.h"
#include "keywords.h"
#include "escan.h"
#include "list.h"

#include "utils.h"


typedef struct {
  int mode;
  int line;
  int depth;
  int count;
  int repcount;
  char exp[LINE_LENGTH];
} flow_marker;


#define FLOW_NORMAL		(0)
#define FLOW_IF 		(1)
#define FLOW_ELSE		(2)
#define FLOW_WHILE		(3)
#define FLOW_REPEAT		(4)
#define FLOW_RAW		(5)

#define MARKER(x) ((flow_marker *)x)


static LIST (MarkersList);
static List *Markers = &MarkersList;

static int FlowMode = FLOW_NORMAL;
static int FlowDepth = 0;
static int FlowTotalDepth = 0;
static int FlowLoopTop = 0;
static int FlowLoopDepth = 0;

static int FlowCount = 0;       // decrements
static int FlowRepCount = 0;    // increments
static char FlowExpression[LINE_LENGTH] = "";


static int marker_pop (void);
static int marker_push (void);
static int ignore_commands (void);

static int parse_print (const char *c);
static void print_flow_state (void);


/******************************************************************************/
/******************************************************************************/

extern int
init_flow_parser (void)
{
  list_init (Markers);

  FlowMode = FLOW_NORMAL;
  FlowDepth = 0;
  FlowTotalDepth = 0;
  FlowLoopTop = 0;
  FlowLoopDepth = 0;
  FlowCount = 0;
  FlowRepCount = 0;
  FlowExpression[0] = '\0';

  return 0;
}

/******************************************************************************/

extern int
free_flow_parser (void)
{
  list_free (Markers);

  FlowMode = FLOW_NORMAL;
  FlowDepth = 0;
  FlowTotalDepth = 0;
  FlowLoopTop = 0;
  FlowLoopDepth = 0;
  FlowCount = 0;
  FlowRepCount = 0;
  FlowExpression[0] = '\0';

  return 0;
}

/******************************************************************************/
/******************************************************************************/

extern int
get_flowed_line (char *buf, int buflen, const char **pos)
{
  int tmp;
  int key;
  int line;
  char tmpbuf[LINE_LENGTH];
  double value;


  while (1) {

    line = get_line (buf, buflen);

    if (line == LINEBUF_ERROR) {
      return LINEBUF_ERROR;
    }
    else if (line == LINEBUF_EOF) {
      if (FlowMode == FLOW_NORMAL) {
        return LINEBUF_EOF;
      }
      else {
        error ("missing end\n");
        return LINEBUF_ERROR;
      }
    }

    if (FlowMode == FLOW_RAW) {
      char *c = buf;
      while ((*c != '\0') && isspace (*c)) {
        c++;
      }
      if (!keyword_is_end (c)) {
        // return a raw string starting at pos
        *pos = c;
        return FlowLoopTop;
      }
    }


    key = keyword_parse (buf, buflen, pos);
    if (key == -1) {
      return LINEBUF_ERROR;
    }


    if (keyword_raw_input (key)) {
      if (marker_push () != 0) {
        return LINEBUF_ERROR;
      }
      FlowMode = FLOW_RAW;
      FlowDepth = 1;
      FlowLoopTop = key;
      FlowLoopDepth++;
      FlowTotalDepth++;
      FlowExpression[0] = '\0';
      continue;
    }


    switch (key) {


      case quit_key:{
        if (FlowMode == FLOW_NORMAL) {
          printf ("Quitting early: line (%i)\n", current_line_number ());
          return LINEBUF_EOF;
        }
        else {
          error ("premature QUIT used\n");
          return LINEBUF_ERROR;
        }
      }


      case var_key:
      case const_key:{
        if (escan (*pos, "sfw", tmpbuf, &value) != 2) {
          error ("missing variable name or value\n");
          return LINEBUF_ERROR;
        }

        /*
         * The program can add variables that
         * start with '_', but the user can not.
         */

        if (!isalpha (tmpbuf[0])) {
          error ("variable names must start with a letter: %s\n", tmpbuf);
          return LINEBUF_ERROR;
        }

        if (key == const_key) {
          if (set_constant (tmpbuf, value) != 0)
            return LINEBUF_ERROR;
          DEBUG3 ("CONSTANT: %s <= %f\n", tmpbuf, value);
        }
        else {
          if (set_variable (tmpbuf, value) != 0)
            return LINEBUF_ERROR;
          DEBUG3 ("VARIABLE: %s <= %f\n", tmpbuf, value);
        }
        break;
      }


      case print_key:{
        if (parse_print (*pos) == -1) {
          return LINEBUF_ERROR;
        }
        break;
      }


      case include_key:{
        if (escan (*pos, "Qw", tmpbuf) != 1) {
          error ("parsing include\n");
          return LINEBUF_ERROR;
        }
        if (include_file (tmpbuf) == LINEBUF_ERROR) {
          return LINEBUF_ERROR;
        }
        if (FlowLoopDepth > 0) {
          unbuffer_line ();
        }
        break;
      }


      case while_key:{
        if (marker_push () != 0) {
          return LINEBUF_ERROR;
        }

        FlowMode = FLOW_WHILE;
        FlowDepth = 1;
        FlowLoopTop = line;
        FlowLoopDepth++;
        FlowTotalDepth++;

        strcpy (FlowExpression, *pos);

        if (escan (FlowExpression, "fw", &value) != 1) {
          return LINEBUF_ERROR;
        }

        if (value <= 0.0) {
          tmp = ignore_commands ();
          if (tmp != LINEBUF_OK)
            return tmp;
          else
            continue;
        }

        buffer_on ();

        break;
      }


      case repeat_key:{
        if (marker_push () != 0) {
          return LINEBUF_ERROR;
        }

        FlowMode = FLOW_REPEAT;
        FlowDepth = 1;
        FlowLoopTop = line;
        FlowLoopDepth++;
        FlowTotalDepth++;
        FlowRepCount = 0;

        tmp = escan (*pos, "isw", &FlowCount, FlowExpression);

        if (tmp < 1) {
          error ("repeat parse\n");
          return LINEBUF_ERROR;
        }

        if (tmp == 2) {
          if (!isalpha (FlowExpression[0])) {
            error ("variable names must start with a letter: %s\n",
                   FlowExpression);
            return LINEBUF_ERROR;
          }
          else if (set_variable (FlowExpression, 0.0) != 0) {
            return LINEBUF_ERROR;
          }
        }
        else {
          FlowExpression[0] = '\0';
        }

        if (FlowCount <= 0) {
          tmp = ignore_commands ();
          if (tmp != LINEBUF_OK)
            return tmp;
          else
            continue;
        }

        buffer_on ();

        break;
      }


      case if_key:{
        if (marker_push () != 0) {
          return LINEBUF_ERROR;
        }

        FlowMode = FLOW_IF;
        FlowDepth = 1;
        FlowTotalDepth++;

        strcpy (FlowExpression, *pos);

        if (escan (FlowExpression, "fw", &value) != 1) {
          return LINEBUF_ERROR;
        }

        if (value <= 0.0) {
          tmp = ignore_commands ();
          if (tmp != LINEBUF_OK) {
            return tmp;
          }
          else {
            continue;
          }
        }

        break;
      }


      case else_key:{
        if (FlowMode != FLOW_IF) {
          error ("misplaced ELSE\n");
          return LINEBUF_ERROR;
        }

        tmp = ignore_commands ();
        if (tmp != LINEBUF_OK) {
          return tmp;
        }
        else {
          continue;
        }

        break;
      }


      case end_key:{

        FlowDepth--;
        FlowTotalDepth--;

        if (FlowDepth < 0) {
          error ("extra end\n");
          return LINEBUF_ERROR;
        }
        else if (FlowDepth > 0) {
          return key;
        }
        else {
          if (get_debug () >= 9) {
            print_flow_state ();
          }

          switch (FlowMode) {
            case FLOW_NORMAL:{
              return key;
            }
            case FLOW_WHILE:{
              if (escan (FlowExpression, "fw", &value) != 1) {
                error ("parsing while\n");
                return LINEBUF_ERROR;
              }

              if (value <= 0.0) {
                buffer_off ();
                marker_pop ();
                FlowLoopDepth--;

                break;
              }
              else {
                FlowDepth++;
                FlowTotalDepth++;
                buffer_jump (FlowLoopTop);
              }

              break;
            }
            case FLOW_REPEAT:{
              FlowCount--;
              FlowRepCount++;

              if (FlowCount <= 0) {
                buffer_off ();
                marker_pop ();
                FlowLoopDepth--;
              }
              else {
                if (FlowExpression[0] != '\0') {
                  set_variable (FlowExpression, (double) FlowRepCount);
                }
                FlowDepth++;
                FlowTotalDepth++;
                buffer_jump (FlowLoopTop);
              }

              break;
            }
            case FLOW_IF:
            case FLOW_ELSE:
            case FLOW_RAW:{
              marker_pop ();
              break;
            }
          }                     /* switch */
        }                       /* else */

        break;
      }


      default:{
        if (keyword_ended (key))
          FlowDepth++;
        return key;
      }

    }                           /* switch */

  }                             /* while */

  return key;
}

/******************************************************************************/
/******************************************************************************/

static int
ignore_commands (void)
{
  int line;
  int key;
  int raw = 0;
  int depth = 1;
  char buf[LINE_LENGTH];
  const char *pos;

  while (depth > 0) {

    line = get_line (buf, LINE_LENGTH);

    if (line == LINEBUF_ERROR) {
      return LINEBUF_ERROR;
    }
    else if (line == LINEBUF_EOF) {
      return LINEBUF_EOF;
    }

    if (raw) {
      char *c = buf;
      while ((*c != '\0') && isspace (*c)) {
        c++;
      }
      if (keyword_is_end (c)) {
        raw = 0;
      }
      continue;
    }

    key = keyword_parse (buf, LINE_LENGTH, &pos);
    if (key == -1) {
      return LINEBUF_ERROR;
    }

    if (key == end_key) {
      depth--;
    }
    else if ((key == else_key) && (FlowMode == FLOW_IF) && (depth == 1)) {
      FlowMode = FLOW_ELSE;
      return LINEBUF_OK;        /* FIXME? -- check for stray ELSE's too? */
    }
    else if (key == include_key) {
      if (FlowLoopDepth > 0) {
        if (escan (pos, "Qw", buf) != 1) {
          error ("parsing include\n");
          return LINEBUF_ERROR;
        }
        if (include_file (buf) == LINEBUF_ERROR)
          return LINEBUF_ERROR;
        unbuffer_line ();
      }
    }
    else if (key == options_key) {
      raw = 1;
    }
    else if (keyword_ended (key)) {
      depth++;
    }
  }

  if (FlowMode == FLOW_NORMAL) {
    error ("ignore_commands() - bad mode\n");
    return LINEBUF_ERROR;
  }
  else {
    marker_pop ();
  }

  return LINEBUF_OK;
}

/******************************************************************************/

static int
marker_push (void)
{
  flow_marker *fm;

  fm = (flow_marker *) malloc (sizeof (flow_marker));
  if (fm == NULL) {
    perror ("marker_push(malloc)");
    return -1;
  }

  if (list_add (fm, Markers) != 0) {
    free (fm);
    return -1;
  }

  fm->mode = FlowMode;
  fm->depth = FlowDepth;
  fm->line = FlowLoopTop;
  fm->exp[0] = '\0';

  if ((FlowMode == FLOW_WHILE) || (FlowMode == FLOW_REPEAT)) {
    strcpy (fm->exp, FlowExpression);
  }
  FlowExpression[0] = '\0';

  if (FlowMode == FLOW_REPEAT) {
    fm->count = FlowCount;
    fm->repcount = FlowRepCount;
  }

  return 0;
}

/******************************************************************************/

static int
marker_pop (void)
{
  flow_marker *fm;

  if (Markers->count < 1)
    return -1;

  fm = (flow_marker *) list_pop (Markers);

  FlowMode = fm->mode;
  FlowDepth = fm->depth;
  FlowLoopTop = fm->line;
  FlowExpression[0] = '\0';

  if ((FlowMode == FLOW_WHILE) || (FlowMode == FLOW_REPEAT)) {
    strcpy (FlowExpression, fm->exp);
  }

  if (FlowMode == FLOW_REPEAT) {
    FlowCount = fm->count;
    FlowRepCount = fm->repcount;
  }

  free (fm);

  return 0;
}

/******************************************************************************/

static int
parse_print (const char *c)
{
  char buffer[LINE_LENGTH];

  if (escan (c, "eQw", buffer) != 1)
    return -1;

  printf ("[%03i]  %s\033[0m\n", current_line_number (), buffer);

  return 0;
}

/******************************************************************************/

static void
print_flow_state (void)
{
  printf ("FlowMode       %s= ", INDENT);
  switch (FlowMode) {
    case FLOW_NORMAL:
      printf ("FLOW_NORMAL\n");
      break;
    case FLOW_WHILE:
      printf ("FLOW_WHILE\n");
      break;
    case FLOW_REPEAT:
      printf ("FLOW_REPEAT\n");
      break;
    case FLOW_IF:
      printf ("FLOW_IF\n");
      break;
    case FLOW_ELSE:
      printf ("FLOW_ELSE\n");
      break;
    default:
      printf ("*** UNKNOWN ***\n");
  }

  printf (INDENT "FlowDepth      = %i\n", FlowDepth);
  printf (INDENT "FlowTotalDepth = %i\n", FlowTotalDepth);
  printf (INDENT "FlowLoopDepth  = %i\n", FlowLoopDepth);
  printf (INDENT "FlowLoopTop    = %i\n", FlowLoopTop);
  printf (INDENT "FlowCount      = %i\n", FlowCount);
  printf (INDENT "FlowRepCount   = %i\n", FlowRepCount);
  printf (INDENT "Markers->count = %i\n", Markers->count);
  printf (INDENT "FlowExpression = %s\n", FlowExpression);

  return;
}

/******************************************************************************/
/******************************************************************************/
