
#include <stdio.h>
#include <math.h>

#include "intersct.h"
#include "utils.h"              /* for M_PI */


#define USE_QUICK_TEST

#define OVERLAP_DIST	(0.1)   /* FIXME - do this somewhere else? */


#define DOT3_PRODUCT(x,y) \
  (((x)[0] * (y)[0]) + ((x)[1] * (y)[1]) + ((x)[2] * (y)[2]))

#define VEC3_SUB(dst,x,y)     \
  (dst)[0] = (x)[0] - (y)[0]; \
  (dst)[1] = (x)[1] - (y)[1]; \
  (dst)[2] = (x)[2] - (y)[2]; \

#define VEC3_SET(dst,x,y,z)    \
  (dst)[0] = (x); \
  (dst)[1] = (y); \
  (dst)[2] = (z); \


typedef double vec2[2];


/*
 * NOTE:
 *   I can guarantee that these algorithms aren't the fastest way of 
 *   testing these types of intersections. They were written in a 
 *   brute-force fashion, and certainly aren't optimized or elegant.
 */


static int quick_test (cvec3 pos1, cvec3 size1, cvec3 pos2, cvec3 size2);

static void top_and_bottom (double *top, double *bot, cvec3 pos,
                            cvec3 size);

static int crossing_line (vec2 point, vec2 normal, vec2 csize,
                          double size);

static int crossing_pyr_planes (cvec3 csize,
                                cvec3 pos2, cvec3 size2,
                                const double rot2);

/*
 * FIXME - csquare_to_square static int csquare_to_square (cvec3 size1,
 * cvec3 pos2, cvec3 size2, const double rot2); 
 */


/******************************************************************************/
/******************************************************************************/

int
point_to_square (cvec3 pos1, cvec3 pos2, cvec3 size2, const double rot2)
{
  double tmpx, tmpy, x, y;
  double xsize, ysize;
  double cos_val, sin_val;

  xsize = fabs (size2[0]);
  ysize = fabs (size2[1]);

  /*
   * Translate 
   */
  tmpx = pos1[0] - pos2[0];
  tmpy = pos1[1] - pos2[1];

  /*
   * Rotate
   */
  cos_val = cos (-rot2 * M_PI / 180.0);
  sin_val = sin (-rot2 * M_PI / 180.0);

  x = (cos_val * tmpx) - (sin_val * tmpy);
  y = (cos_val * tmpy) + (sin_val * tmpx);


  if ((x >= xsize) || (x <= -xsize) || (y >= ysize) || (y <= -ysize))
    return 0;

  return 1;
}

/******************************************************************************/

extern void
center_square (cvec3 cpos, const double crot, cvec3 tpos,
               const double trot, vec3 rpos, double *rrot)
{
  double tmpx, tmpy;
  double cos_val, sin_val;

  tmpx = tpos[0] - cpos[0];
  tmpy = tpos[1] - cpos[1];

  cos_val = cos (-crot * (M_PI / 180.0));
  sin_val = sin (-crot * (M_PI / 180.0));

  rpos[0] = (cos_val * tmpx) - (sin_val * tmpy);
  rpos[1] = (cos_val * tmpy) + (sin_val * tmpx);

  *rrot = trot - crot;

  return;
}

/******************************************************************************/

static int
csquare_to_square (cvec3 csize, cvec3 pos2, cvec3 size2, const double rot2)
{
  double cos_val, sin_val;
  double nvadj[2], nhadj[2];
  vec2 s1, s2;                  /* absolute size values */
  vec2 nh, nv;                  /* horizontal and vertical normals */
  vec2 point;                   /* starting piont of line */
  vec3 zeroes = { 0.0, 0.0, 0.0 };

  csquare_to_square (csize, pos2, size2, rot2); /* FIXME */

  /*
   * Is square1 centered within square2?
   */
  if (point_to_square (zeroes, pos2, size2, rot2)) {
    return 1;
  }

  s1[0] = fabs (csize[0]);
  s1[1] = fabs (csize[1]);
  s2[0] = fabs (size2[0]);
  s2[1] = fabs (size2[1]);

  /*
   * Is square2 centered within square1?
   */
  if ((pos2[0] >= -s1[0]) && (pos2[0] <= s1[0]) &&
      (pos2[1] >= -s1[1]) && (pos2[1] <= s1[1])) {
    return 1;
  }


  cos_val = cos (rot2 * (M_PI / 180.0));
  sin_val = sin (rot2 * (M_PI / 180.0));

  nh[0] = cos_val;
  nh[1] = sin_val;
  nv[0] = -sin_val;
  nv[1] = cos_val;

  nhadj[0] = nh[0] * s2[0];
  nhadj[1] = nh[1] * s2[0];
  nvadj[0] = nv[0] * s2[1];
  nvadj[1] = nv[1] * s2[1];


  if (IS_ZERO (nh[1])) {
    DEBUG9 ("csquare_to_square: testing 0/180 degrees\n");
    if (((fabs (pos2[0]) - s2[0]) < s1[0]) &&
        ((fabs (pos2[1]) - s2[1]) < s1[1])) {
      return 1;
    }
    else {
      return 0;
    }
  }
  if (IS_ZERO (nh[0])) {
    DEBUG9 ("csquare_to_square: testing 90/270 degrees\n");
    if (((fabs (pos2[0]) - s2[1]) < s1[0]) &&
        ((fabs (pos2[1]) - s2[0]) < s1[1])) {
      return 1;
    }
    else {
      return 0;
    }
  }

  point[0] = pos2[0] + nvadj[0];
  point[1] = pos2[1] + nvadj[1];
  if (crossing_line (point, nh, s1, s2[0])) {
    return 1;
  }

  point[0] = pos2[0] - nvadj[0];
  point[1] = pos2[1] - nvadj[1];
  if (crossing_line (point, nh, s1, s2[0])) {
    return 1;
  }

  point[0] = pos2[0] + nhadj[0];
  point[1] = pos2[1] + nhadj[1];
  if (crossing_line (point, nv, s1, s2[1])) {
    return 1;
  }

  point[0] = pos2[0] - nhadj[0];
  point[1] = pos2[1] - nhadj[1];
  if (crossing_line (point, nv, s1, s2[1])) {
    return 1;
  }

  return 0;
}

/******************************************************************************/

int
square_to_square (cvec3 pos1, cvec3 size1, const double rot1,
                  cvec3 pos2, cvec3 size2, const double rot2)
{
  double cos_val, sin_val;
  vec2 s1, s2;                  /* absolute size values */
  vec2 nh, nv;                  /* horizontal and vertical normals */
  vec2 point;                   /* starting piont of line */
  vec3 npos2;
  double nrot2;


#ifdef USE_QUICK_TEST
  if (quick_test (pos1, size1, pos2, size2))
    return 0;
#endif


  /*
   * Is square1 centered within square2?
   */
  point[0] = pos1[0];
  point[1] = pos1[1];
  if (point_to_square (pos1, pos2, size2, rot2)) {
    return 1;
  }

  center_square (pos1, rot1, pos2, rot2, npos2, &nrot2);

  cos_val = cos (nrot2 * (M_PI / 180.0));
  sin_val = sin (nrot2 * (M_PI / 180.0));

  nh[0] = cos_val;
  nh[1] = sin_val;
  nv[0] = -sin_val;
  nv[1] = cos_val;

  s1[0] = fabs (size1[0]);
  s1[1] = fabs (size1[1]);
  s2[0] = fabs (size2[0]);
  s2[1] = fabs (size2[1]);


  /*
   * Is square2 centered within square1?
   */
  if ((npos2[0] >= -s1[0]) && (npos2[0] <= s1[0]) &&
      (npos2[1] >= -s1[1]) && (npos2[1] <= s1[1])) {
    return 1;
  }

  if (IS_ZERO (nh[1])) {
    DEBUG9 ("square_to_square: testing 0/180 degrees\n");
    if (((fabs (npos2[0]) - s2[0]) < s1[0]) &&
        ((fabs (npos2[1]) - s2[1]) < s1[1])) {
      return 1;
    }
    else {
      return 0;
    }
  }
  if (IS_ZERO (nh[0])) {
    DEBUG9 ("square_to_square: testing 90/270 degrees\n");
    if (((fabs (npos2[0]) - s2[1]) < s1[0]) &&
        ((fabs (npos2[1]) - s2[0]) < s1[1])) {
      return 1;
    }
    else {
      return 0;
    }
  }

  point[0] = npos2[0] + (nv[0] * s2[1]);
  point[1] = npos2[1] + (nv[1] * s2[1]);
  if (crossing_line (point, nh, s1, s2[0])) {
    return 1;
  }

  point[0] = npos2[0] - (nv[0] * s2[1]);
  point[1] = npos2[1] - (nv[1] * s2[1]);
  if (crossing_line (point, nh, s1, s2[0])) {
    return 1;
  }

  point[0] = npos2[0] + (nh[0] * s2[0]);
  point[1] = npos2[1] + (nh[1] * s2[0]);
  if (crossing_line (point, nv, s1, s2[1])) {
    return 1;
  }

  point[0] = npos2[0] - (nh[0] * s2[0]);
  point[1] = npos2[1] - (nh[1] * s2[0]);
  if (crossing_line (point, nv, s1, s2[1])) {
    return 1;
  }

  return 0;
}

/******************************************************************************/

int
square_in_square (cvec3 inpos, cvec3 insize, const double inrot,
                  cvec3 outpos, cvec3 outsize, const double outrot)
{
  double tmpx, tmpy, xin, yin, rin;
  double cos_val, sin_val;
  double maxx, maxy;
  double sxin, syin, sxout, syout;  /* absolute size values */


#ifdef USE_QUICK_TEST
  /*
   * No quick test is used here.
   * This function is only called after a square_to_square 
   * in this program.
   */
#endif

  /*
   * Translate to center the outside square
   */
  tmpx = inpos[0] - outpos[0];
  tmpy = inpos[1] - outpos[1];

  /*
   * Rotate to center the outside square
   */
  cos_val = cos (-outrot * M_PI / 180.0);
  sin_val = sin (-outrot * M_PI / 180.0);
  xin = (cos_val * tmpx) - (sin_val * tmpy);
  yin = (cos_val * tmpy) + (sin_val * tmpx);

  rin = (inrot - outrot) * (M_PI / 180.0);
  cos_val = cos (rin);
  sin_val = sin (rin);

  /*
   * Get the absolute sizes
   */
  sxin = fabs (insize[0]);
  syin = fabs (insize[1]);
  sxout = fabs (outsize[0]);
  syout = fabs (outsize[1]);

  tmpx = fabs ((cos_val * sxin) - (sin_val * syin));
  maxx = fabs ((cos_val * sxin) + (sin_val * syin));
  if (tmpx > maxx) {
    maxx = tmpx;
  }

  tmpy = fabs ((cos_val * syin) - (sin_val * sxin));
  maxy = fabs ((cos_val * syin) + (sin_val * sxin));
  if (tmpy > maxy) {
    maxy = tmpy;
  }

  if (((xin + maxx) < (+sxout - ACCURACY)) &&
      ((xin - maxx) > (-sxout + ACCURACY)) &&
      ((yin + maxy) < (+syout - ACCURACY)) &&
      ((yin - maxy) > (-syout + ACCURACY))) {
    return 1;
  }

  return 0;
}

/******************************************************************************/

int
box_to_box (cvec3 pos1, cvec3 size1, const double rot1,
            cvec3 pos2, cvec3 size2, const double rot2)
{
  double top1, bot1;
  double top2, bot2;

  top_and_bottom (&top1, &bot1, pos1, size1);
  top_and_bottom (&top2, &bot2, pos2, size2);

  if ((bot1 >= top2) || (bot2 >= top1))
    return 0;

  return square_to_square (pos1, size1, rot1, pos2, size2, rot2);
}

/******************************************************************************/

int
box_to_pyr (cvec3 pos1, cvec3 size1, const double rot1,
            cvec3 pos2, cvec3 size2, const double rot2)
{
  vec3 adjsize;
  double adjscale = 1.0;
  double boxtop, boxbot;
  double pyrtop, pyrbot;


  top_and_bottom (&boxtop, &boxbot, pos1, size1);
  top_and_bottom (&pyrtop, &pyrbot, pos2, size2);

  if ((boxbot >= pyrtop) || (pyrbot >= boxtop))
    return 0;


  if ((pos2[2] >= boxbot) && (pos2[2] <= boxtop)) {
    return square_to_square (pos1, size1, rot1, pos2, size2, rot2);
  }
  else if (pos2[2] < boxbot) {
    adjscale = (fabs (size2[2]) - (boxbot - pyrbot)) / fabs (size2[2]);
  }
  else {
    adjscale = (fabs (size2[2]) - (pyrtop - boxtop)) / fabs (size2[2]);
  }

  adjsize[0] = size2[0] * adjscale;
  adjsize[1] = size2[1] * adjscale;

  return square_to_square (pos1, size1, rot1, pos2, adjsize, rot2);
}

/******************************************************************************/

int
pyr_to_pyr (cvec3 pos1, cvec3 size1, const double rot1,
            cvec3 pos2, cvec3 size2, const double rot2)
{
  vec3 adjsize1, adjsize2;
  double adjscale1, adjscale2;
  double tip1, base1;
  double tip2, base2;
  double top1, bot1;
  double top2, bot2;


  top_and_bottom (&top1, &bot1, pos1, size1);
  top_and_bottom (&top2, &bot2, pos2, size2);

  if ((bot1 >= top2) || (bot2 >= top1))
    return 0;


  tip1 = pos1[2] + size1[2];
  tip2 = pos2[2] + size2[2];
  base1 = pos1[2];
  base2 = pos2[2];

  /*
   * Test all limit cases.
   */

  if ((base1 >= bot2) && (base1 <= top2)) {
    double size = fabs (size2[2]);

    adjscale2 = (size - (fabs (base1 - base2))) / size;
    adjsize2[0] = size2[0] * adjscale2;
    adjsize2[1] = size2[1] * adjscale2;

    DEBUG9 ("ADJSCALE2 for BASE1 = %f\n", adjscale2);

    if (square_to_square (pos1, size1, rot1, pos2, adjsize2, rot2))
      return 1;
  }

  if ((base2 >= bot1) && (base2 <= top1)) {
    double size = fabs (size1[2]);

    adjscale1 = (size - (fabs (base2 - base1))) / size;
    adjsize1[0] = size1[0] * adjscale1;
    adjsize1[1] = size1[1] * adjscale1;

    DEBUG9 ("ADJSCALE1 for BASE2 = %f\n", adjscale1);

    if (square_to_square (pos1, adjsize1, rot1, pos2, size2, rot2))
      return 1;
  }


  if ((tip1 >= bot2) && (tip1 <= top2)) {
    double size = fabs (size2[2]);

    adjscale2 = (size - (fabs (tip1 - base2))) / size;
    adjsize2[0] = size2[0] * adjscale2;
    adjsize2[1] = size2[1] * adjscale2;

    DEBUG9 ("ADJSCALE2 for TIP1 = %f\n", adjscale2);

    if (point_to_square (pos1, pos2, adjsize2, rot2))
      return 1;
  }

  if ((tip2 >= bot1) && (tip2 <= top1)) {
    double size = fabs (size1[2]);

    adjscale1 = (size - (fabs (tip2 - base1))) / size;
    adjsize1[0] = size1[0] * adjscale1;
    adjsize1[1] = size1[1] * adjscale1;

    DEBUG9 ("ADJSCALE1 for TIP2 = %f\n", adjscale1);

    if (point_to_square (pos2, pos1, adjsize1, rot1))
      return 1;
  }

  if ((size1[2] * size2[2]) < 0.0) {
    vec3 npos;
    double nrot;

    /*
     * determination of normals requires an upwards pyramid
     * at the center.
     */

    if (size1[2] >= 0.0) {
      center_square (pos1, rot1, pos2, rot2, npos, &nrot);
      npos[2] = pos2[2] - pos1[2];  /* shift Z */
      return crossing_pyr_planes (size1, npos, size2, nrot);
    }
    else {
      center_square (pos2, rot2, pos1, rot1, npos, &nrot);
      npos[2] = pos1[2] - pos2[2];  /* shift Z */
      return crossing_pyr_planes (size2, npos, size1, nrot);
    }
  }

  return 0;
}

/******************************************************************************/
/******************************************************************************/

static int
quick_test (cvec3 pos1, cvec3 size1, cvec3 pos2, cvec3 size2)
{
  double x, y, tmp1, tmp2;

  x = (pos2[0] - pos1[0]) * (pos2[0] - pos1[0]);
  y = (pos2[1] - pos1[1]) * (pos2[1] - pos1[1]);
  tmp1 = x + y;

  x = (size1[0] * size1[0]) + (size1[1] * size1[1]);
  y = (size2[0] * size2[0]) + (size2[1] * size2[1]);
  tmp2 = x + y + 2.0 * sqrt (x * y);

  if (tmp1 > tmp2)
    return 1;

  return 0;
}

/******************************************************************************/

static void
top_and_bottom (double *top, double *bot, cvec3 pos, cvec3 size)
{
  if (size[2] >= 0.0) {
    *bot = pos[2];
    *top = pos[2] + size[2];
  }
  else {
    *bot = pos[2] + size[2];;
    *top = pos[2];
  }
  return;
}

/******************************************************************************/

static int
crossing_line (vec2 point, vec2 normal, vec2 csize, double size)
{
  double k;
  double cross;

  /*
   * Test the top and bottom origin lines
   */

  k = (csize[1] - point[1]) / normal[1];
  cross = point[0] + (k * normal[0]);
  if ((fabs (k) <= size) && (fabs (cross) < csize[0]))
    return 1;

  k = (-csize[1] - point[1]) / normal[1];
  cross = point[0] + (k * normal[0]);
  if ((fabs (k) <= size) && (fabs (cross) < csize[0]))
    return 1;

  /*
   * Test the right and left origin lines
   */

  k = (csize[0] - point[0]) / normal[0];
  cross = point[1] + (k * normal[1]);
  if ((fabs (k) <= size) && (fabs (cross) < csize[1]))
    return 1;

  k = (-csize[0] - point[0]) / normal[0];
  cross = point[1] + (k * normal[1]);
  if ((fabs (k) <= size) && (fabs (cross) < csize[1]))
    return 1;

  return 0;
}

/******************************************************************************/

static int
lines_to_wedge (cvec3 N, cvec3 S, cvec3 Pp, cvec3 Pc, cvec3 Pl, vec3 L[4],
                double cdist)
{
  vec3 P;                       /* intersection point */
  double k;                     /* scalar constant for line length */
  vec3 vtmp;
  double dot_N_L;
  double dotpoint;
  double dotcorner;
  double pdist;
  int i;


  VEC3_SUB (vtmp, Pc, Pp);
  dotcorner = DOT3_PRODUCT (S, vtmp);

  if (get_debug () >= 9) {
    printf ("P2P: \n");
    printf ("P2P: cdist = %f\n", cdist);
    printf ("P2P: N = {%f, %f, %f}\n", N[0], N[1], N[2]);
    printf ("P2P: S = {%f, %f, %f}\n", S[0], S[1], S[2]);
    printf ("P2P: Pp = {%f, %f, %f}\n", Pp[0], Pp[1], Pp[2]);
    printf ("P2P: Pl = {%f, %f, %f}\n", Pl[0], Pl[1], Pl[2]);
    printf ("P2P: Pc = {%f, %f, %f}\n", Pc[0], Pc[1], Pc[2]);
    printf ("P2P: \n");
  }

  for (i = 0; i < 4; i++) {

    DEBUG9 ("P2P: L[%i] = {%f, %f, %f}\n", i, L[i][0], L[i][1], L[i][2]);

    dot_N_L = DOT3_PRODUCT (N, L[i]);

    if (IS_ZERO (dot_N_L))
      continue;                 /* line is parallel to plane */


    VEC3_SUB (vtmp, Pp, Pl);
    k = DOT3_PRODUCT (N, vtmp) / dot_N_L;

    DEBUG9 ("P2P: k = %f\n", k);


    if ((k <= 0.0) || (k >= 1.0))
      continue;                 /* intersection is off of the line */


    P[0] = (k * L[i][0]) + Pl[0];
    P[1] = (k * L[i][1]) + Pl[1];   /* calculate the intersection point */
    P[2] = (k * L[i][2]) + Pl[2];

    DEBUG9 ("P2P: P = {%f, %f, %f}\n", P[0], P[1], P[2]);


    VEC3_SUB (vtmp, P, Pp);
    dotpoint = DOT3_PRODUCT (S, vtmp);

    if (dotpoint <= 0.0)
      continue;                 /* intersection is above the wedge */


    if (dotpoint >= dotcorner)
      continue;                 /* intersection is below the wedge */


    pdist =
        (vtmp[0] * vtmp[0]) + (vtmp[1] * vtmp[1]) + (vtmp[2] * vtmp[2]);

    if ((dotpoint * dotpoint * cdist) > (dotcorner * dotcorner * pdist)) {
      DEBUG9 ("P2P: INTERSECTION i=%i\n", i);
      return 1;
      continue;
    }
  }

  return 0;
}

/******************************************************************************/

static int
crossing_pyr_planes (cvec3 csize, cvec3 pos2, cvec3 size2,
                     const double rot2)
{
  vec3 N;                       /* normal to the plane */
  vec3 S;                       /* vector along the plane's slope */
  vec3 Pp;                      /* top of the center pyramid */
  vec3 Pc;                      /* corner point of the pyramid */
  vec3 Pl;                      /* beginning of the line */
  vec3 L[4];
  double hmdist;
  double vmdist;
  double crnrdist;
  double xsqu, ysqu, zsqu;
  double abssize2;
  double cos_val, sin_val;
  double cos_sx, cos_sy, sin_sx, sin_sy;

  DEBUG9 ("P2P: crossing_pyr_planes\n");

  /*
   * FIXME -- absolute SX and SY's? 
   */
  xsqu = (csize[0] * csize[0]); /* FIXME - not used */
  ysqu = (csize[1] * csize[1]);
  zsqu = (csize[2] * csize[2]);
  hmdist = xsqu + zsqu;
  vmdist = ysqu + zsqu;
  crnrdist = xsqu + ysqu + zsqu;

  Pp[0] = 0.0;
  Pp[1] = 0.0;
  Pp[2] = csize[2];

  Pl[0] = pos2[0];
  Pl[1] = pos2[1];
  Pl[2] = pos2[2] + size2[2];

  DEBUG9 ("P2P: POS2[2] = %f\n", pos2[2]);
  DEBUG9 ("P2P: SIZE2[2] = %f\n", size2[2]);


  cos_val = cos (rot2 * M_PI / 180.0);
  sin_val = sin (rot2 * M_PI / 180.0);
  cos_sx = cos_val * size2[0];
  cos_sy = cos_val * size2[1];
  sin_sx = sin_val * size2[0];
  sin_sy = sin_val * size2[1];
  VEC3_SET (L[0], +(cos_sx) - (sin_sy), +(cos_sy) + (sin_sx), -size2[2]);
  VEC3_SET (L[1], -(cos_sx) + (sin_sy), -(cos_sy) - (sin_sx), -size2[2]);
  VEC3_SET (L[2], +(cos_sx) + (sin_sy), -(cos_sy) + (sin_sx), -size2[2]);
  VEC3_SET (L[3], -(cos_sx) - (sin_sy), +(cos_sy) - (sin_sx), -size2[2]);

  abssize2 = fabs (csize[2]);

  DEBUG9 ("P2P:\n");
  DEBUG9 ("P2P: RIGHT\n");
  VEC3_SET (N, abssize2, 0.0, csize[0]);    /* right plane */
  VEC3_SET (S, csize[0], 0.0, -csize[2]);
  VEC3_SET (Pc, csize[0], csize[1], 0.0);

  if (lines_to_wedge (N, S, Pp, Pc, Pl, L, crnrdist)) {
    return 1;
  }

  DEBUG9 ("P2P:\n");
  DEBUG9 ("P2P: LEFT\n");
  VEC3_SET (N, -abssize2, 0.0, csize[0]);   /* left plane */
  VEC3_SET (S, -csize[0], 0.0, -csize[2]);
  VEC3_SET (Pc, -csize[0], -csize[1], 0.0);

  if (lines_to_wedge (N, S, Pp, Pc, Pl, L, crnrdist)) {
    return 1;
  }

  DEBUG9 ("P2P:\n");
  DEBUG9 ("P2P: TOP\n");
  VEC3_SET (N, 0.0, abssize2, csize[1]);    /* top plane */
  VEC3_SET (S, 0.0, csize[1], -csize[2]);
  VEC3_SET (Pc, -csize[0], csize[1], 0.0);

  if (lines_to_wedge (N, S, Pp, Pc, Pl, L, crnrdist)) {
    return 1;
  }

  DEBUG9 ("P2P:\n");
  DEBUG9 ("P2P: BOTTOM\n");
  VEC3_SET (N, 0.0, -abssize2, csize[1]);   /* bottom plane */
  VEC3_SET (S, 0.0, -csize[1], -csize[2]);
  VEC3_SET (Pc, csize[0], -csize[1], 0.0);

  if (lines_to_wedge (N, S, Pp, Pc, Pl, L, crnrdist)) {
    return 1;
  }


  return 0;
}

/******************************************************************************/

extern int
check_box_overlap (cvec3 pos1, cvec3 size1, const double rot1,
                   cvec3 pos2, cvec3 size2, const double rot2)
{
  double angle, tmp;
  double top1, bot1;
  double top2, bot2;
  vec3 npos2;
  double nrot2;
  vec2 s1, s2;


  /*
   * 90 degree angle?
   */

  angle = rot1 - rot2;
  if (!IS_ZERO (fmod (angle, 90.0))) {
    return 0;
  }


  /*
   * Do they intersect at all?
   */

  top_and_bottom (&top1, &bot1, pos1, size1);
  top_and_bottom (&top2, &bot2, pos2, size2);

  if ((bot1 >= top2) || (bot2 >= top1)) {
    return 0;
  }

  // FIXME - what to do?

  if (((bot1 + ACCURACY) >= top2) || ((bot2 + ACCURACY) >= top1)) {
    return 0;
  }

  center_square (pos1, rot1, pos2, rot2, npos2, &nrot2);

  s1[0] = fabs (size1[0]);
  s1[1] = fabs (size1[1]);
  s2[0] = fabs (size2[0]);
  s2[1] = fabs (size2[1]);

  if (!IS_ZERO (fmod (nrot2, 180.0))) {
    tmp = s2[0];
    s2[0] = s2[1];
    s2[1] = tmp;
  }

  if (((fabs (npos2[0]) - s2[0]) > (s1[0] - OVERLAP_DIST)) ||
      ((fabs (npos2[1]) - s2[1]) > (s1[1] - OVERLAP_DIST))) {
    /*
     * FIXME - use this? (s1[0] < OVERLAP_DIST) || (s1[1] < OVERLAP_DIST)
     * || (s2[0] < OVERLAP_DIST) || (s2[1] < OVERLAP_DIST) ||
     * (fabs(size1[2]) < OVERLAP_DIST) || (fabs(size2[2]) <
     * OVERLAP_DIST)) { 
     */
    return 0;
  }


  /*
   * Check for the different overlap types
   */

  if (IS_EQUAL (top1, top2)) {
    DEBUG3 ("Top Overlap\n");
    return 1;
  }
  if (IS_EQUAL (bot1, bot2) && !IS_ZERO (bot1)) {
    DEBUG3 ("Bot Overlap\n");
    return 1;
  }

  if (IS_EQUAL (s1[0], npos2[0] + s2[0])) {
    DEBUG3 ("West Overlap\n");
    return 1;
  }
  if (IS_EQUAL (-s1[0], npos2[0] - s2[0])) {
    DEBUG3 ("East Overlap\n");
    return 1;
  }

  if (IS_EQUAL (s1[1], npos2[1] + s2[1])) {
    DEBUG3 ("North Overlap %f = %f + %f\n", s1[1], npos2[1], s2[1]);
    DEBUG3 ("North Overlap s1 (%f, %f)\n", s1[0], s1[1]);
    DEBUG3 ("North Overlap s2 (%f, %f)\n", s2[0], s2[1]);
    DEBUG3 ("North NPos2 (%f, %f, %f)\n", npos2[0], npos2[1], npos2[2]);
    DEBUG3 ("North NRot2 (%f)\n", nrot2);
    return 1;
  }
  if (IS_EQUAL (-s1[1], npos2[1] - s2[1])) {
    DEBUG3 ("South Overlap %f = %f - %f\n", -s1[1], npos2[1], s2[1]);
    DEBUG3 ("South Overlap s1 (%f, %f)\n", s1[0], s1[1]);
    DEBUG3 ("South Overlap s2 (%f, %f)\n", s2[0], s2[1]);
    DEBUG3 ("South NPos2 (%f, %f, %f)\n", npos2[0], npos2[1], npos2[2]);
    DEBUG3 ("South NRot2 (%f)\n", nrot2);
    return 1;
  }

  return 0;
}

/******************************************************************************/

extern int
check_pyr_overlap (cvec3 ppos, cvec3 psize, const double prot,
                   cvec3 bpos, cvec3 bsize, const double brot,
                   double *factor)
{
  double psa, bsa;

  if (IS_ZERO (psize[2])) {
    return 0;
  }

  /*
   * Check for a base planar intersection
   */

  if ((psize[2] * bsize[2]) >= 0.0) {
    if (!IS_EQUAL (ppos[2], bpos[2])) {
      return 0;
    }
  }
  else {
    if (!IS_EQUAL (ppos[2], bpos[2] + bsize[2])) {
      return 0;
    }
  }

  /*
   * Do they really intersect?
   */

  if (!square_to_square (ppos, psize, prot, bpos, bsize, brot)) {
    return 0;
  }

  if (!square_in_square (ppos, psize, prot, bpos, bsize, brot)) {
    return 0;
  }

  psa = fabs (psize[2]);
  bsa = fabs (bsize[2]);

  if (bsa >= psa) {
    *factor = -1.0;
  }
  else {
    *factor = (psa - bsa) / psa;
  }

  return 1;
}

/******************************************************************************/
/******************************************************************************/
