
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "bzmapper.h"
#include "linebuf.h"
#include "utils.h"
#include "list.h"


#define COMMENTS_NEED_SPACE

#define NO_WHITE_BUFFER         /* marignal speedup */


typedef struct _filedef {
  FILE *file;
  char *abspath;
  char *relpath;
  int linenum;
  struct _filedef *prev;
} filedef;

typedef struct {
  char *line;
  int linenum;
  int runcount;
  filedef *file;
} linebuf;

typedef struct {
  int line;
  const filedef *file;
} debug_info;



static LIST (FilesList);
static List *Files = &FilesList;
static filedef *CurrentFile = NULL;
static int IncludeDepth = 0;

static LIST (BuffersList);
static List *Buffers = &BuffersList;
static int BufferLine = -2;
static int BufferLevel = 0;
static int LastUsed = -1;

static LIST (DebugInfoList);
static List *DebugInfo = &DebugInfoList;

static int RunCountMax = 0;
static time_t TimeOut = 0;
static time_t StartTime;

static int get_non_ended (char *buf, int buflen);
static int get_non_white (char *buf, int buflen);
static int clean_line (char *buf);
static int buffer_line (const char *line);
static void free_filedef (void *fd);
static void free_linebuf (void *lb);
static int make_filenames (const char *filename, filedef * f);
static filedef *find_file (const char *name);

static void print_file_debug (linebuf_fd fd);


#define FILEDEF(x) ((filedef *) FilesList.items[x])
#define BUFFER(x) ((linebuf *) BuffersList.items[x])


/*
 * Here is a quick description of the buffering algorithm
 * and control.
 *
 * Number coming into nad out of the 'get_line()' function.
 * Numbers before the '|' symbol are before the call, and 
 * those after are the results after thee function.
 *
 *
 *   App# : value returned by get_line()
 *   BL#  : BufferLine
 *   CNT  : Buffers->count 
 *   ALB  : Actual Buffers->items[x] number for that line
 *
 *
 *    TEXT         App#   BL#    CNT   ABL
 *    ----         ----   ---    ---   ---
 *    ...           |0   -2|-2   0|0   N/A
 *    while (x)     |0   -2|-1   0|0   N/A       <- sets BufferLine to -1
 *      cmd1        |1   -1| 0   0|1    0
 *      while (y)   |2    0| 1   1|2    1
 *        cmd2      |3    1| 2   2|3    2
 *      end         |4    2| 3   3|4    3        <- buffer_jump (2)
 *    end           |5    3| 4   4|5    4        <- buffer_jump (0)
 *
 */


/****************************************************************************/
/****************************************************************************/

void
init_line_buffer (void)
{
  list_init (Files);
  list_init (Buffers);
  list_init (DebugInfo);
  IncludeDepth = 0;
  CurrentFile = NULL;
  BufferLine = -2;
  BufferLevel = 0;
  LastUsed = -1;
  StartTime = time (NULL);

  return;
}

/****************************************************************************/

void
free_line_buffer (void)
{
  list_nuke (free_filedef, Files);
  list_nuke (free_linebuf, Buffers);
  list_nuke (free, DebugInfo);
  CurrentFile = NULL;
  BufferLine = -2;
  BufferLevel = 0;
  LastUsed = -1;
  StartTime = time (NULL);

  return;
}

/****************************************************************************/
/****************************************************************************/

int
get_line (char *buf, int buflen)
{
  int tmp;

  DEBUG4 ("get_line = [Blev=%2i, BL=%2i, BC=%2i]\n",
          BufferLevel, BufferLine, Buffers->count);

  if ((BufferLine > -2) && ((BufferLine + 1) < Buffers->count)) {
    BufferLine++;
    strcpy (buf, BUFFER (BufferLine)->line);
    BUFFER (BufferLine)->runcount++;
    LastUsed = BufferLine;

    if ((RunCountMax > 0) &&
        (BUFFER (BufferLine)->runcount >= RunCountMax)) {
      error ("Run count for line exceeded (%i)\n", RunCountMax);
      return -1;
    }
    if ((TimeOut > 0) && ((time (NULL) - StartTime) > TimeOut)) {
      error ("Parsing time exceeded (%i sec)\n", TimeOut);
      return -1;
    }
    DEBUG5 ("get_line (executed) [BL=%2i, BC=%2i] <<%s>>\n",
            BufferLine, Buffers->count, buf);

    return BufferLine + 1;
  }
  else {
    tmp = get_non_ended (buf, buflen);

    if (tmp == LINEBUF_ERROR) {
      return LINEBUF_ERROR;
    }
    else if (tmp == LINEBUF_EOF) {
      return LINEBUF_EOF;
    }

    if (BufferLine == (Buffers->count - 1)) {
      buffer_line (buf);
      BufferLine++;
      LastUsed = -1;
      DEBUG5 ("get_line (buffered) [BL=%2i, BC=%2i] <<%s>>\n",
              BufferLine, Buffers->count, buf);

      return BufferLine + 1;
    }
    else {
      DEBUG5 ("get_line (eaten-up) [BL=%2i, BC=%2i] <<%s>>\n",
              BufferLine, Buffers->count, buf);
      return LINEBUF_OK;
    }
  }

  return LINEBUF_OK;
}

/****************************************************************************/

static int
get_non_ended (char *buf, int buflen)
{
  int tmp;
  int done = 0;

  while (!done) {

    tmp = get_non_white (buf, buflen);

    if (tmp == LINEBUF_ERROR) {
      return LINEBUF_ERROR;
    }
    else if (tmp == LINEBUF_EOF) {

      fclose (CurrentFile->file);
      CurrentFile->file = NULL;
      IncludeDepth--;

      if (CurrentFile->prev != NULL) {
        CurrentFile = CurrentFile->prev;
        if (get_debug () >= 2) {
          print_file_debug (CurrentFile);
        }
        continue;
      }
      else {
        return LINEBUF_EOF;
      }
    }
    else {
      done = 1;
    }
  }

  return LINEBUF_OK;
}

/****************************************************************************/

static int
get_non_white (char *buf, int buflen)
{
  int tmp;

  while (1) {

    if (fgets (buf, buflen, CurrentFile->file) == NULL) {

      if (feof (CurrentFile->file) != 0)
        return LINEBUF_EOF;
      else
        return LINEBUF_ERROR;
    }

    CurrentFile->linenum++;

    if (buf[0] != '\0') {       /* check the length */
      int len;
      len = (int) strlen (buf);
      if ((len == (buflen - 1)) && (buf[buflen - 1] != '\n')) {
        error ("Line exceeds maximum length: %i\n", buflen);
        return LINEBUF_ERROR;
      }
    }

    tmp = clean_line (buf);     /* 0 is a blank/commented line */

    if (tmp == 1) {
      return LINEBUF_OK;
    }
    else if (tmp == LINEBUF_ERROR) {
      return LINEBUF_ERROR;
    }
  }

  return LINEBUF_OK;
}

/****************************************************************************/

static int
clean_line (char *buf)
{
  char *c;
  int text = 0;
  int quote = 0;
  int escape = 0;

  c = buf;

  while (*c != '\0') {

    if (*c == '"') {
      text = 1;
      if (quote) {
        if (escape)
          escape = 0;
        else
          quote = 0;
      }
      else {
        quote = 1;
        escape = 0;
      }
    }
    else if ((*c == '\\') && quote) {
      if (escape)
        escape = 0;
      else
        escape = 1;
    }

#ifdef COMMENTS_NEED_SPACE

    else if ((*c == '#') && !quote) {
      if ((c == buf) || isspace (*(c - 1))) {
        *c = '\0';              /* clip the comments and newlines */
      }
      else {
        error ("comments need to be spaced\n");
        return LINEBUF_ERROR;
      }
      break;
    }
    else if (((*c == '\n') || (*c == '\r')) && !quote) {
      *c = '\0';                /* clip the newlines */
      break;
    }

#else                           /* COMMENTS_NEED_SPACE */

    else if (((*c == '#') || (*c == '\n') || (*c == '\r')) && !quote) {
      *c = '\0';                /* clip the comments and newlines */
      break;
    }

#endif                          /* COMMENTS_NEED_SPACE */

    else {
      if (!isspace (*c))
        text = 1;
      escape = 0;
    }

    c++;
  }

  if (quote) {
    error ("single \", unterminated string.\n");
    return LINEBUF_ERROR;
  }
  else if (text) {
    return 1;
  }
  else {
    return 0;
  }
}

/****************************************************************************/

int
buffer_on (void)
{
  BufferLevel++;

  if (BufferLine == -2) {
    BufferLine = -1;
    LastUsed = -1;
  }

  DEBUG3 ("buffer_on (%d)\n", BufferLevel);

  return LINEBUF_OK;
}

/****************************************************************************/

int
buffer_off (void)
{
  BufferLevel--;


  if (BufferLevel < 0) {
    DEBUG3 ("buffer_off (%d)\n", BufferLevel);
    return LINEBUF_ERROR;
  }
  else if (BufferLevel == 0) {
    list_nuke (free_linebuf, Buffers);
    BufferLine = -2;
    LastUsed = -1;
    DEBUG3 ("buffer_off (%d)\n", BufferLevel);
  }

  return LINEBUF_OK;
}

/****************************************************************************/

int
buffer_jump (int line)
{
  if ((line < 0) || (line > Buffers->count)) {
    DEBUG3 ("buffer_jump ERROR [line=%d, BC=%d]\n", line, Buffers->count);
    return LINEBUF_ERROR;
  }

  BufferLine = line - 1;
  DEBUG3 ("buffer_jump (%d)\n", BufferLine);

  return LINEBUF_OK;
}

/****************************************************************************/

int
unbuffer_line (void)
{
  if (BufferLevel == 0) {
    return LINEBUF_ERROR;
  }
  else {
    free_linebuf (list_pop (Buffers));
    BufferLine--;
  }

  return LINEBUF_OK;
}

/****************************************************************************/
/****************************************************************************/

int
include_file (const char *filename)
{
  filedef *f;

  /*
   * Create and Initialize the File Descriptor
   */

  IncludeDepth++;

  if (IncludeDepth >= LINEBUF_INC_DEPTH) {
    error ("File include depth exceeded (%i)\n", LINEBUF_INC_DEPTH);
    return -1;
  }

  f = (filedef *) malloc (sizeof (filedef));
  if (f == NULL) {
    perror ("include_file(malloc)");
    return LINEBUF_ERROR;
  }
  f->file = NULL;
  f->abspath = NULL;
  f->relpath = NULL;
  f->linenum = 0;

  if (make_filenames (filename, f) == LINEBUF_ERROR) {
    free_filedef (f);
    return LINEBUF_ERROR;
  }

  if (list_add (f, Files) == -1) {
    free_filedef (f);
    return LINEBUF_ERROR;
  }


  /*
   * Check for file recursion
   */

  if (find_file (f->abspath) != NULL) {
    error ("File Recursion\n");
    free_filedef (f);
    list_pop (Files);
    return LINEBUF_ERROR;
  }


  /*
   * Open the file
   */

  f->file = fopen (f->abspath, "rb");   /* the 'b' is for Windows */
  if (f->file == NULL) {
    perror ("fopen");
    free_filedef (f);
    list_pop (Files);
    return LINEBUF_ERROR;
  }

  f->prev = CurrentFile;

  CurrentFile = f;


  if (get_debug () >= 2) {
    print_file_debug (CurrentFile);
  }

  return LINEBUF_OK;
}

/****************************************************************************/
/****************************************************************************/

static int
buffer_line (const char *line)
{
  linebuf *b;

  b = (linebuf *) malloc (sizeof (linebuf));
  if (b == NULL) {
    perror ("buffer_line(malloc)");
    return LINEBUF_ERROR;
  }

#ifdef NO_WHITE_BUFFER
  while ((*line != '\0') && isspace (*line))
    line++;
#endif                          /* NO_WHITE_BUFFER */

  b->line = strdup (line);
  if (b == NULL) {
    perror ("buffer_line(strdup)");
    free (b);
    return LINEBUF_ERROR;
  }

  if (list_add (b, Buffers) == -1) {
    free (b->line);
    free (b);
    return LINEBUF_ERROR;
  }

  b->file = CurrentFile;
  b->linenum = CurrentFile->linenum;
  b->runcount = 0;

  return LINEBUF_OK;
}

/****************************************************************************/

static void
free_filedef (void *fd)
{
  filedef *f;

  f = (filedef *) fd;

  if (f->file != NULL)
    fclose (f->file);

  free (f->abspath);
  free (f->relpath);
  free (f);

  return;
}

/****************************************************************************/

static void
free_linebuf (void *lb)
{
  linebuf *b;

  b = (linebuf *) lb;

  free (b->line);
  free (b);

  return;
}

/****************************************************************************/

static int
make_filenames (const char *filename, filedef * f)
{
  char *abscurr;
  char *relcurr;
  char startpath[2] = { '.', '\0' };    /* constant problems... */

  if (CurrentFile == NULL)
    abscurr = startpath;
  else
    abscurr = CurrentFile->abspath;

  if (get_absolute_name (abscurr, filename, &f->abspath) == LINEBUF_ERROR) {
    error ("couldn't find file: %s\n", filename);
    return LINEBUF_ERROR;
  }

  if (CurrentFile == NULL)
    relcurr = startpath;
  else
    relcurr = CurrentFile->relpath;

  if (get_relative_name (relcurr, filename, &f->relpath) == LINEBUF_ERROR) {
    error ("couldn't find file: %s\n", filename);
    free (f->abspath);
    return LINEBUF_ERROR;
  }

  return LINEBUF_OK;
}

/****************************************************************************/

static filedef *
find_file (const char *name)
{
  filedef *f;

  f = CurrentFile;

  while (f != NULL) {
#ifndef _WIN32
    if (strcmp (f->abspath, name) == 0) {
#else
    if (stricmp (f->abspath, name) == 0) {
#endif
      return f;
    }

    f = f->prev;
  }

  return NULL;
}

/****************************************************************************/
/****************************************************************************/

extern int
current_line_number (void)
{
  if (DebugInfo->count > 0) {
    return ((debug_info *) DebugInfo->items[DebugInfo->count - 1])->line;
  }
  else if (LastUsed >= 0) {
    return BUFFER (LastUsed)->linenum;
  }
  else {
    if (CurrentFile == NULL)
      return -1;
    else
      return CurrentFile->linenum;
  }
}

/****************************************************************************/

extern linebuf_fd
current_file (void)
{
  if (DebugInfo->count > 0) {
    return ((debug_info *) DebugInfo->items[DebugInfo->count - 1])->file;
  }
  else if (LastUsed >= 0)
    return BUFFER (LastUsed)->file;
  else
    return CurrentFile;
}

/****************************************************************************/

extern const char *
current_filename (void)
{
  if (LastUsed >= 0)
    return BUFFER (BufferLine)->file->abspath;
  else if (CurrentFile != NULL)
    return CurrentFile->abspath;
  else
    return NULL;
}

/****************************************************************************/

extern const char *
top_filename (linebuf_fd fd)
{
  if (fd == NULL)
    return NULL;

  return fd->relpath;
}

/****************************************************************************/

extern void
set_runcount (int max)
{
  if (max > 0)
    RunCountMax = max;
  else
    RunCountMax = 0;

  return;
}

/****************************************************************************/

extern void
set_timeout (int secs)
{
  if (secs > 0)
    TimeOut = secs;
  else
    TimeOut = 0;

  return;
}

/****************************************************************************/
/****************************************************************************/

extern void
print_file_stack (linebuf_fd fd)
{
  if (fd == NULL) {
    printf ("error:\n");
    return;
  }
  else {
    printf ("error in: %s\n", fd->relpath);

    if (fd->prev == NULL)
      return;

    while (fd->prev != NULL) {
      fd = fd->prev;
      printf (INDENT INDENT "from: %s\n", fd->relpath);
    }
  }
  return;
}

/****************************************************************************/

static void
print_file_debug (linebuf_fd fd)
{
  int i = 0;

  printf ("FILENAME STACK:\n");

  while (fd != NULL) {
    printf (INDENT "ABSNAME(%i) %s\n", i, fd->abspath);
    printf (INDENT "RELNAME(%i) %s\n", i, fd->relpath);
    fd = fd->prev;
    i++;
  }

  return;
}

/****************************************************************************/

extern void
push_debug_info (int linenum, linebuf_fd fd)
{
  debug_info *di;

  di = (debug_info *) malloc (sizeof (debug_info));
  di->line = linenum;
  di->file = fd;

  list_add (di, DebugInfo);

  return;
}

/****************************************************************************/

extern void
pop_debug_info (void)
{
  if (DebugInfo->count == 0) {
    printf ("Internal error, pop_debug_info() in list.c\n");
  }

  free (list_pop (DebugInfo));

  return;
}

/****************************************************************************/
/****************************************************************************/
