
#include <stdio.h>
#include <stdlib.h>

#include "loadbmp.h"
#include "utils.h"


#define RED_WEIGHT 	 (0.299)
#define GREEN_WEIGHT (0.587)
#define BLUE_WEIGHT  (0.114)


typedef struct {
  u8 red;
  u8 green;
  u8 blue;
  u8 alpha;
} RgbaBit;


#define WINBMP	(0)
#define OS2BMP	(1)

#define COMP_MODE_NONE		(0)
#define COMP_MODE_RLE8		(1)
#define COMP_MODE_RLE4		(2)
#define COMP_MODE_BFIELDS	(3)

typedef struct {

  u8 BmpType;                   /* WINBMP or OS2BMP */

  /*
   * BMP File Header (14 bytes)
   */

  u16 FileType;
  u32 FileSize;
  u16 Reserved1;                /* always 0 */
  u16 Reserved2;                /* always 0 */
  u32 DataOffset;

  /*
   * BMP Info/Core Header (16 bytes)
   */

  s32 HeaderSize;
  s32 Width;
  s32 Height;
  u16 Planes;                   /* always 1 */
  u16 BitCount;

  /*
   * Only for Windows BMPs (extra 24 bytes)
   */

  u32 Compression;
  u32 ImageSize;
  u32 xPelsPerMeter;
  u32 yPelsPerMeter;
  u32 ColorUsed;
  u32 ColorImportant;

  /*
   * The Color Palette 
   */
  u16 colors;
  u8 red[256];
  u8 green[256];
  u8 blue[256];
  u8 grey[256];
  u32 bitfield_red;
  u32 bitfield_green;
  u32 bitfield_blue;

  u32 rowbytes;

} BmpHeader;


static int read_value (void *ptr, u32 size, FILE * file);
#define READVAL(f,x) (read_value (&x, sizeof(x), f))

static int bmp_header (FILE * file, BmpHeader * bh);
static int bmp_palette (FILE * file, BmpHeader * bh);
static void greyscale_palette (BmpHeader * bh);

static int load_1bit_image (FILE * file, BmpHeader * bh, u8 * image);
static int load_4bit_image (FILE * file, BmpHeader * bh, u8 * image);
static int load_8bit_image (FILE * file, BmpHeader * bh, u8 * image);
static int load_16bit_image (FILE * file, BmpHeader * bh, u8 * image);
static int load_24bit_image (FILE * file, BmpHeader * bh, u8 * image);
static int load_32bit_image (FILE * file, BmpHeader * bh, u8 * image);


/****************************************************************************/

#define BYTE(x,b) (*((u8 *)(x) + b))

static int
read_value (void *ptr, u32 size, FILE * file)
{
#if (BYTE_ORDER != LITTLE_ENDIAN)
  u8 tmp;
#endif

  if (fread (ptr, size, 1, file) != 1)
    return 0;

#if (BYTE_ORDER != LITTLE_ENDIAN)
  if (size == 4) {
    tmp = BYTE (ptr, 0);
    BYTE (ptr, 0) = BYTE (ptr, 3);
    BYTE (ptr, 3) = tmp;
    tmp = BYTE (ptr, 1);
    BYTE (ptr, 1) = BYTE (ptr, 2);
    BYTE (ptr, 2) = tmp;
  }
  else if (size == 2) {
    tmp = BYTE (ptr, 0);
    BYTE (ptr, 0) = BYTE (ptr, 1);
    BYTE (ptr, 1) = tmp;
  }
#endif

  return 1;
}

/****************************************************************************/

static int
bmp_header (FILE * file, BmpHeader * bh)
{
  int bytes;

  if (!READVAL (file, bh->FileType) ||
      !READVAL (file, bh->FileSize) ||
      !READVAL (file, bh->Reserved1) ||
      !READVAL (file, bh->Reserved2) || !READVAL (file, bh->DataOffset)) {
    perror ("fread(BmpFileHeader)");
    return -1;
  }

  if (bh->FileType != ('B' | ('M' << 8))) {
    error ("Not a BMP file.\n");
    error ("VAL: (%04X)\n", bh->FileType);
    error ("CHAR0: %c (%02X)\n",
           bh->FileType & 0x00FF, bh->FileType & 0x00FF);
    error ("CHAR1: %c (%02X)\n",
           (bh->FileType & 0xFF00) >> 8, (bh->FileType & 0xFF00) >> 8);
    return -1;
  }

  DEBUG7 ("FileSize       = %i\n", bh->FileSize);
  DEBUG7 ("Reserved1      = %i\n", bh->Reserved1);
  DEBUG7 ("Reserved2      = %i\n", bh->Reserved2);
  DEBUG7 ("OffseBits      = %i\n", bh->DataOffset);


  if (!READVAL (file, bh->HeaderSize) ||
      !READVAL (file, bh->Width) ||
      !READVAL (file, bh->Height) ||
      !READVAL (file, bh->Planes) || !READVAL (file, bh->BitCount)) {
    perror ("fread(BmpInfoHeader)");
    return -1;
  }

  bytes = bh->Width * bh->BitCount;
  bytes = (bytes + 7) / 8;
  bytes = (bytes + 3) / 4;
  bytes = bytes * 4;
  bh->rowbytes = bytes;

  DEBUG7 ("Header Size    = %i\n", bh->HeaderSize);
  DEBUG7 ("Width          = %i\n", bh->Width);
  DEBUG7 ("Height         = %i\n", bh->Height);
  DEBUG7 ("Planes         = %i\n", bh->Planes);
  DEBUG7 ("BitCount       = %i\n", bh->BitCount);
  DEBUG7 ("Rowbytes       = %i\n", bh->rowbytes);

  if (bh->HeaderSize >= 40) {
    int i;

    bh->BmpType = WINBMP;

    if (!READVAL (file, bh->Compression) ||
        !READVAL (file, bh->ImageSize) ||
        !READVAL (file, bh->xPelsPerMeter) ||
        !READVAL (file, bh->yPelsPerMeter) ||
        !READVAL (file, bh->ColorUsed) ||
        !READVAL (file, bh->ColorImportant)) {
      perror ("fread(BmpWinHeader)");
      return -1;
    }

    for (i = 40; i < bh->HeaderSize; i++) {
      if (fgetc (file) == EOF) {
        return -1;
      }
    }

    DEBUG7 ("Compression    = %i\n", bh->Compression);
    DEBUG7 ("ImageSize      = %i\n", bh->ImageSize);
    DEBUG7 ("xPelsPerMeter  = %i\n", bh->xPelsPerMeter);
    DEBUG7 ("yPelsPerMeter  = %i\n", bh->yPelsPerMeter);
    DEBUG7 ("ColorUsed      = %i\n", bh->ColorUsed);
    DEBUG7 ("ColorImportant = %i\n", bh->ColorImportant);
  }
  else if (bh->HeaderSize == 12) {
    bh->BmpType = OS2BMP;
  }
  else {
    error ("file header size error (%i)\n", bh->HeaderSize);
    return -1;
  }

  return 0;
}

/****************************************************************************/

static int
bmp_palette (FILE * file, BmpHeader * bh)
{
  u32 i;
  u32 eaten = 0;

  bh->colors = 0;

  switch (bh->BitCount) {
    case 1:
    case 4:
    case 8:{

      if (bh->ColorUsed == 0) {
        bh->colors = (1 << bh->BitCount);
      }
      else {
        bh->colors = bh->ColorUsed;
      }

      for (i = 0; i < bh->colors; i++) {
        u8 junk;

        if (!READVAL (file, bh->blue[i]) ||
            !READVAL (file, bh->green[i]) || !READVAL (file, bh->red[i])) {
          perror ("fread(color palette");
          return -1;
        }
        else if ((bh->BmpType == WINBMP) && !READVAL (file, junk)) {
          perror ("fread(color palette");
          return -1;
        }

        DEBUG9 ("colpal[%i] R-%-3i G-%-3i B-%-3i\n", i,
                bh->red[i], bh->green[i], bh->blue[i]);
      }

      for (i = bh->colors; i < (u32) (1 << bh->BitCount); i++) {
        bh->red[i] = 0;
        bh->green[i] = 0;
        bh->blue[i] = 0;
      }

      if (bh->BmpType == WINBMP)
        eaten = bh->colors * 4;
      else
        eaten = bh->colors * 3;

      break;
    }
    case 16:
    case 32:{
      if (bh->Compression == COMP_MODE_BFIELDS) {
        DEBUG9 ("BitFields!!!\n");
        if (!READVAL (file, bh->bitfield_red) ||
            !READVAL (file, bh->bitfield_green) ||
            !READVAL (file, bh->bitfield_blue)) {
          perror ("fread(bitfields)");
          return -1;
        }
      }

      eaten = 4 * 3;

      break;
    }
    case 24:{
      break;
    }
    default:{
      error ("Non-standard bit depth (%i)\n", bh->BitCount);
      return -1;
    }
  }


  /*
   * Suck up any extraneous characters 
   */

  for (i = (14 + bh->HeaderSize + eaten); i < bh->DataOffset; i++) {
    if (fgetc (file) == EOF) {
      return -1;
    }
  }

  return 0;
}

/****************************************************************************/

static void
greyscale_palette (BmpHeader * bh)
{
  int i;

  for (i = 0; i < bh->colors; i++) {
    bh->grey[i] = (u8)
        ((RED_WEIGHT * (float) bh->red[i]) +
         (GREEN_WEIGHT * (float) bh->green[i]) +
         (BLUE_WEIGHT * (float) bh->blue[i]) + 0.5);
  }

  if (get_debug () >= 8) {
    for (i = 0; i < bh->colors; i++) {
      printf ("grey[%i] = %i\n", i, bh->grey[i]);
    }
  }

  return;
}

/****************************************************************************/

static int
load_1bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  int byte = 0;
  u32 count;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {
      if ((col % 8) == 0) {
        byte = fgetc (file);
        count++;
        if (byte == EOF) {
          perror ("load_1bit_image");
          return -1;
        }
      }

      if ((byte & 0x80) == 0)
        *(image + (row * bh->Width + col)) = bh->grey[0];
      else
        *(image + (row * bh->Width + col)) = bh->grey[1];

      byte = byte << 1;

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

static int
load_4bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  int byte = 0;
  u32 count;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {
      if ((col % 2) == 0) {
        byte = fgetc (file);
        if (byte == EOF) {
          perror ("load_4bit_image");
          return -1;
        }
        count++;

        *(image + (row * bh->Width + col)) = bh->grey[(byte >> 4) & 0x0F];
      }
      else {
        *(image + (row * bh->Width + col)) = bh->grey[byte & 0x0F];
      }

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

static int
load_8bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  int byte;
  u32 count;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {
      byte = fgetc (file);
      if (byte == EOF) {
        perror ("load_8bit_image");
        return -1;
      }
      count++;

      *(image + (row * bh->Width + col)) = bh->grey[byte];

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

static int
load_16bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  u32 count;
  u16 value;
  u8 red, green, blue;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {

      if (!READVAL (file, value)) {
        perror ("load_16bit_image");
        return -1;
      }

      blue = (value & 0x01F) << 3;
      value = value >> 5;
      green = (value & 0x01F) << 3;
      value = value >> 5;
      red = (value & 0x01F) << 3;


      count = count + 2;

      value = (u8)
          ((RED_WEIGHT * (float) red) +
           (GREEN_WEIGHT * (float) green) +
           (BLUE_WEIGHT * (float) blue) + 0.5);

      *(image + (row * bh->Width + col)) = (u8) value;

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

static int
load_24bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  u32 count;
  int blue, green, red;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {
      blue = fgetc (file);
      green = fgetc (file);
      red = fgetc (file);

      count = count + 3;

      if ((red == EOF) || (green == EOF) || (blue == EOF)) {
        perror ("load_24bit_image");
        return -1;
      }

      *(image + (row * bh->Width + col)) = (u8)
          ((RED_WEIGHT * (float) red) +
           (GREEN_WEIGHT * (float) green) +
           (BLUE_WEIGHT * (float) blue) + 0.5);

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

static int
load_32bit_image (FILE * file, BmpHeader * bh, u8 * image)
{
  int row, col;
  u32 count;
  int blue, green, red, junk;

  for (row = 0; row < bh->Height; row++) {

    count = 0;

    for (col = 0; col < bh->Width; col++) {
      blue = fgetc (file);
      green = fgetc (file);
      red = fgetc (file);
      junk = fgetc (file);

      count = count + 4;

      if ((red == EOF) || (green == EOF) || (blue == EOF)) {
        perror ("load_32bit_image");
        return -1;
      }

      *(image + (row * bh->Width + col)) = (u8)
          ((RED_WEIGHT * (float) red) +
           (GREEN_WEIGHT * (float) green) +
           (BLUE_WEIGHT * (float) blue) + 0.5);

      if (col == (bh->Width - 1)) {
        while (count < bh->rowbytes) {
          fgetc (file);
          count++;
        }
      }
    }
  }

  return 0;
}

/****************************************************************************/

int
load_bmp (const char *filename, int *w, int *h, u8 ** image)
{
  int tmp;
  FILE *file;
  BmpHeader bh;

  file = fopen (filename, "rb");    /* the 'b' is for Windows */
  if (file == NULL) {
    error ("opening (%s) ", filename);
    perror ("");
    return -1;
  }

  if ((bmp_header (file, &bh) != 0) || (bmp_palette (file, &bh) != 0)) {
    fclose (file);
    return -1;
  }

  if (bh.Compression != COMP_MODE_NONE) {
    error ("Compression not currently supported.\n");
    fclose (file);
    return -1;
  }

  *w = bh.Width;
  *h = bh.Height;

  if (bh.colors > 0)
    greyscale_palette (&bh);

  DEBUG7 ("Rowbytes = %i\n", bh.rowbytes);

  *image = (u8 *) malloc (sizeof (u8) * bh.Width * bh.Height);
  if (*image == NULL) {
    perror ("malloc(image)");
    fclose (file);
    return -1;
  }

  if (bh.Compression != COMP_MODE_NONE) {
    error ("Compression not currently supported.\n");
  }

  switch (bh.BitCount) {
    case 1:{
      tmp = load_1bit_image (file, &bh, *image);
      break;
    }
    case 4:{
      tmp = load_4bit_image (file, &bh, *image);
      break;
    }
    case 8:{
      tmp = load_8bit_image (file, &bh, *image);
      break;
    }
    case 16:{
      tmp = load_16bit_image (file, &bh, *image);
      break;
    }
    case 24:{
      tmp = load_24bit_image (file, &bh, *image);
      break;
    }
    case 32:{
      tmp = load_32bit_image (file, &bh, *image);
      break;
    }
    default:{
      error ("Currently unsupported bit depth (%i)\n", bh.BitCount);
      fclose (file);
      return -1;
    }
  }

  if (tmp != 0)
    error ("Loading of BMP unsuccessful.\n");


  fclose (file);
  return 0;
}

/****************************************************************************/
