
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <string.h>

#include "bzmapper.h"
#include "utils.h"


#define IN_EXTENSION    ".bzg"
#define IN_EXTENSION2   ".map"
#define OUT_EXTENSION   ".bzw"

#define INDENT "  "
#define MAX_DIGITS      (6)
#define MAX_PRECISION   (0.0000005)

#define MEMSTAT         "/proc/self/status"
#define MEMLINE         "VmSize:"
#define MEMLOOPS        (4096 * 1)


/*
 * Function prototypes
 */
static int parse_options (int argc, char **argv);
static void print_help (const char *execname);
static int copy_header (FILE * infile, FILE * outfile);
static char *make_outname (const char *inname);
static int process_map (const char *infile, const char *outfile);
static void fprintf_double (double value);
static void print_basics (vec3 pos, vec3 size, double rot,
                          int drivethrough, int shootthrough);
static void print_name (const char *name);

static int mesh_box (const char *name, vec3 pos, vec3 size, double rot,
                     int drivethrough, int shootthrough);
static int mesh_pyr (const char *name, vec3 pos, vec3 size, double rot,
                     int drivethrough, int shootthrough, int flipz);

/*
 * Static variables
 */
static FILE *OutFile;
static int Stats = 0;
static int Output = 1;
static int Loop = 0;
static int Comments = 0;
static int OutputNames = 1;
static int OutputMeshes = 0;
static int PrintDefaults = 0;

static double WorldSize = 400.0f; // half width

static const char TagString[] = "# Parsed by " BZMAPPER_NAME;


/******************************************************************************/
/******************************************************************************/

int
main (int argc, char **argv)
{
  int tmp;
  int code;
  int args_left;
  char *inname;
  char *outname;
  int count = 0;


  printf ("\n%s %s\n\n", BZMAPPER_NAME, BZMAPPER_VERSION);


  args_left = parse_options (argc, argv);

  if ((args_left < 0) || (args_left > 2)) {
    print_help (argv[0]);
    return EXIT_FAILURE;
  }
  else if (args_left == 0) {
    if (PrintDefaults) {
      load_map (NULL);
    }
    else {
      print_help (argv[0]);
    }
    return EXIT_FAILURE;
  }
  else {
    argv = argv + (argc - args_left);
    inname = argv[0];

    if (args_left == 1) {
      outname = make_outname (inname);
      if (outname == NULL) {
        printf ("couldn't make output filename.\n");
        return EXIT_FAILURE;
      }
    }
    else {                      /* args_left == 2 */
      outname = argv[1];
    }
  }


  printf ("input file  : %s\n", inname);
  if (Output)
    printf ("output file : %s\n\n", outname);
  else
    printf ("output file : << none >>\n\n");


  for (tmp = 0; tmp < 72; tmp++) {
    putchar ('-');
  }
  putchar ('\n');



  do {

    if (Loop && (count % (1024) == 0)) {
      fprintf (stderr, "COUNT: %i\n", count);
    }


    /*
     * Process the map
     */
    code = process_map (inname, outname);

#ifdef _WIN32
    if (code != 0) {
      if (!Loop) {
        fflush (stdout);
        printf ("Press RETURN to continue.\n");
        getchar ();
      }
    }
#endif


    if (!Loop) {
      break;
    }

    count++;
  }
  while (Loop);

  if (outname != argv[1]) {
    free (outname);
  }

  return code;
}

/******************************************************************************/
/******************************************************************************/

static int
parse_options (int argc, char **argv)
{
  int tmp;
  char *ctmp;
  char **end;

  end = argv + argc;
  argv++;

  while (argv != end) {

    if (argv[0][0] != '-') {
      return (end - argv);
    }
    else {
      switch (argv[0][1]) {

        case '-':{
          argv++;
          return (end - argv);
        }
        case 'h':{
          return -1;
        }
        case 'u':{
          break;
        }
        case 'c':{
          Comments = 1;
          break;
        }
        case 'n':{
          Output = 0;
          break;
        }
        case 'i':{
          switch (argv[0][2]) {
            case 's':
              set_input_size_style (BZMAPPER_FLIPZ);
              break;
            case 'n':
              set_ignore_names (0);
              break;
            case 'c':
              set_convert_old ();
              break;
            default:
              printf ("Unknown option flag: \"-i%c\"\n\n", argv[0][2]);
              return -1;
          }
          break;
        }
        case 'o':{
          switch (argv[0][2]) {
            case 's':
              set_output_size_style (BZMAPPER_NEGSIZE);
              break;
            case 'v':
              set_overlaps ();
              break;
            case 'm':
              set_merge_boxes ();
              break;
            case 'z':
              set_z_ordering (BZMAPPER_UNORDERED);
              break;
            case 'Z':
              set_z_ordering (BZMAPPER_DESCENDING);
              break;
            case 'n':
              OutputNames = 0;
              break;
            case 'M':
              OutputMeshes = 1;
              break;
            default:
              printf ("Unknown option flag: \"-o%c\"\n\n", argv[0][2]);
              return -1;
          }
          break;
        }
        case 't':
        case 'r':{
          char c = argv[0][1];

          if ((end - argv) == 1) {
            return -1;          /* ran out of arguments */
          }

          argv++;

          tmp = strtol (argv[0], &ctmp, 10);
          if (ctmp == argv[0]) {
            return -1;          /* bad conversion */
          }

          if (c == 't') {
            set_timeout (tmp);
          }
          else {
            set_runcount (tmp);
          }

          break;
        }
        case 'd':{
          switch (argv[0][2]) {
            case 't':
              set_debug_teleporters ();
              break;
            case 'o':
              set_debug_objects ();
              break;
            case 's':
              Stats = 1;
              break;
            case 'd':
              PrintDefaults = 1;
              break;
            case 'l':
              Loop = 1;
              break;
            default:{
              if ((argv[0][2] >= '0') && (argv[0][2] <= '9')) {
                set_debug (argv[0][2] - '0');
              }
              else {
                printf ("Unknown option flag: \"-d%c\"\n\n", argv[0][2]);
                return -1;
              }
            }
          }
          break;
        }
        default:{
          printf ("Unknown option flag: \"-%c\"\n\n", argv[0][1]);
          return -1;
        }
      }                         /* end switch */
    }                           /* end 'is option' */
    argv++;
  }                             /* end WHILE */

  return 0;
}

/******************************************************************************/

static void
print_help (const char *execname)
{
  printf ("Usage:  %s [OPTIONS] <infile> [outfile]\n\n"
          "  --   : end of options (for files that start with '-')\n"
          "  -h   : print this help message\n"
          "  -is  : use new style negz sizes (for 1.10.x compatibility)\n"
          "  -ic  : convert old 1.7g2 style maps\n"
          "  -in  : don't ignore 'name' parameters\n"
          "  -os  : output old negative size style\n"
          "  -ov  : remove planar overlaps\n"
          "  -om  : merge boxes\n"
          "  -oM  : output boxes and pyramids as arcs and cones\n"
          "  -oz  : don't order the output by height\n"
          "  -oZ  : order the output by ascending height\n"
          "  -on  : don't output name tags\n"
          "  -n   : don't produce any output\n"
          "  -c   : add comments for object names\n"
          "  -t N : set a time-out limit (seconds)\n"
          "  -r N : set a run count limit (line executions)\n"
          "  -dN  : N is the debugging level (0-9) [ex: -d2]\n"
          "  -do  : debug objects\n"
          "  -dt  : debug teleporters\n"
          "  -ds  : debug statistics\n"
          "  -dd  : print default variables and functions\n"
          "  -dl  : continually loop (for debugging)\n\n", execname);

  return;
}

/******************************************************************************/

#ifdef _WIN32
#  define my_strcmp stricmp
#else
#  define my_strcmp strcmp
#endif

static char *
make_outname (const char *inname)
{
  char *outname;
  char *ext;
  char *tmp;
  int length;

  outname = strdup (inname);
  if (outname == NULL) {
    perror ("make_outname(strdup)");
    return NULL;
  }

  length = strlen (outname);

  ext = outname + (length - strlen (IN_EXTENSION));

  if (my_strcmp (IN_EXTENSION, ext) == 0) {
    *ext = 0;
  }

  if (my_strcmp (IN_EXTENSION2, ext) == 0) {
    *ext = 0;
  }

  length = strlen (outname);

  tmp = (char *) realloc (outname, sizeof (char) *
                          (length + strlen (OUT_EXTENSION) + 1));
  if (tmp == NULL) {
    free (outname);
    perror ("make_outname(realloc)");
    return NULL;
  }
  else {
    outname = tmp;
  }

  strcat (outname, OUT_EXTENSION);

  return outname;
}

/******************************************************************************/

static int
copy_header (FILE * infile, FILE * outfile)
{
  int i, whitelines;
  int already_tagged = 0;
  char *c;
  char buffer[LINE_LENGTH];

  buffer[0] = '\0';
  c = buffer;
  whitelines = 0;


  while (!feof (infile)) {

    fgets (buffer, LINE_LENGTH, infile);

    if (strncmp (buffer, TagString, strlen (TagString)) == 0) {
      already_tagged = 1;
      sprintf (buffer, "%s %s #\n", TagString, BZMAPPER_VERSION);
    }

    c = buffer;

    while ((*c != '\0') && isspace (*c) && (*c != '#')) {
      c++;
    }

    if (*c == '\0') {
      whitelines++;
    }
    else if (*c == '#') {
      if (whitelines > 0) {
        for (i = 0; i < whitelines; i++) {
          fputc ('\n', outfile);
        }
      }
      whitelines = 0;
      fputs (buffer, outfile);
    }
    else {
      break;
    }
  }

  if (!already_tagged) {
    fprintf (outfile, "\n");
    fprintf (outfile, "###########################\n");
    fprintf (outfile, "%s %s #\n", TagString, BZMAPPER_VERSION);
    fprintf (outfile, "###########################\n\n");
  }
  else {
    fprintf (outfile, "\n");
  }

  return 0;
}

/******************************************************************************/

static int
process_map (const char *infile, const char *outfile)
{
  if (Output) {
    FILE *file;

    file = fopen (infile, "rb");    /* the 'b' is for Windows */
    if (file == NULL) {
      perror ("input file");
      return -1;
    }

    OutFile = fopen (outfile, "w");
    if (OutFile == NULL) {
      perror ("output file");
      fclose (file);
      return -1;
    }

    copy_header (file, OutFile);

    fclose (file);
  }


  /*
   * Load a custom map
   */
  if (load_map (infile) != 0) {
    if (Output) {
      fclose (OutFile);
      remove (outfile);
    }
    return -1;
  }


  if (Output) {
    fclose (OutFile);
  }


  if (Stats) {
    printf ("\nStatistics:\n");
    printf (INDENT "Boxes       : %i\n", stats_boxes ());
    printf (INDENT "Pyramids    : %i\n", stats_pyramids ());
    printf (INDENT "Tetras      : %i\n", stats_tetras ());
    printf (INDENT "Bases       : %i\n", stats_bases ());
    printf (INDENT "Teleporters : %i\n", stats_teleporters ());
    printf (INDENT "Links       : %i\n", stats_links ());
    printf (INDENT "Weapons     : %i\n", stats_weapons ());
    printf (INDENT "Zones       : %i\n\n", stats_zones ());
  }


  return 0;
}

/******************************************************************************/

static void
fprintf_double (double value)
{
  int i;
  double d, m;
  char format[] = " %.?f";      /* '?' in format[3] */

  d = 1.0;

  for (i = 0; i < MAX_DIGITS; i++) {
    m = fabs (fmod (value, d));
    if ((m < MAX_PRECISION) || (fabs (m - d) < MAX_PRECISION))
      break;
    d = d / 10.0;
  }

  format[3] = '0' + i;

  if (Output) {
    fprintf (OutFile, format, value);
  }

  return;
}

/******************************************************************************/
/******************************************************************************/

static void


print_basics (vec3 pos, vec3 size, double rot,
              int drivethrough, int shootthrough)
{
  fprintf (OutFile, INDENT "position");
  fprintf_double (pos[0]);
  fprintf_double (pos[1]);
  fprintf_double (pos[2]);
  fprintf (OutFile, "\n");

  fprintf (OutFile, INDENT "size");
  fprintf_double (size[0]);
  fprintf_double (size[1]);
  fprintf_double (size[2]);
  fprintf (OutFile, "\n");

  fprintf (OutFile, INDENT "rotation");
  fprintf_double (rot);
  fprintf (OutFile, "\n");

  if (drivethrough) {
    fprintf (OutFile, INDENT "drivethrough\n");
  }

  if (shootthrough) {
    fprintf (OutFile, INDENT "shootthrough\n");
  }

  return;
}

/******************************************************************************/

static void
print_name (const char *name)
{
  char *str, *c;

  if (Comments) {
    fprintf (OutFile, "#!name /# %s\n", name);
  }

  if (OutputNames) {

    str = strdup (name);
    if (str == NULL) {
      printf ("Memory error in print_name() [main.c]\n");
      exit (EXIT_FAILURE);
    }

    c = str;
    while (*c != '\0') {
      if (*c == '/') {
        *c = '|';
      }
      c++;
    }

    fprintf (OutFile, INDENT "name <%s>\n", str);   /* the '<' matters */

    free (str);
  }

  return;
}

/******************************************************************************/

int
make_world (double size, double flag_height)
{
  WorldSize = size;

  if (!Output)
    return 0;

  fprintf (OutFile, "world\n");

  fprintf (OutFile, INDENT "size");
  fprintf_double (size);
  fprintf (OutFile, "\n");

  fprintf (OutFile, INDENT "flagHeight");
  fprintf_double (flag_height);
  fprintf (OutFile, "\n");

  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_box (const char *name, vec3 pos, vec3 size, double rot,
          int drth, int shth)
{
  if (!Output)
    return 0;
    
  if (!OutputMeshes) {
    fprintf (OutFile, "box\n");
    print_name (name);
    print_basics (pos, size, rot, shth, drth);
    fprintf (OutFile, "end\n\n");
  } else {
    return mesh_box (name, pos, size, rot, drth, shth);
  }

  return 0;
}

/******************************************************************************/

int
make_pyr (const char *name, vec3 pos, vec3 size, double rot,
          int drth, int shth, int flipz)
{
  if (!Output)
    return 0;

  if (!OutputMeshes || flipz) {
    fprintf (OutFile, "pyramid\n");
    print_name (name);
    print_basics (pos, size, rot, drth, shth);
    if (flipz) {
      fprintf (OutFile, INDENT "flipz\n");
    }
    fprintf (OutFile, "end\n\n");
  } else {
    return mesh_pyr (name, pos, size, rot, drth, shth, flipz);
  }

  return 0;
}

/******************************************************************************/

int
make_tetra (const char *name, const vec3 *vertices, unsigned char visible,
            unsigned char colored, const unsigned char (*colors)[4],
            int drivethrough, int shootthrough)
{
  static int warning = 0;
  int v;
  
  if (!Output)
    return 0;
    
  //FIXME
  if (warning) {
    fprintf (stderr, "Warning: bzmapper tetra's are disabled.\n");
    warning = 0;
  }
  return 0;
  

  fprintf (OutFile, "tetra\n");
  print_name (name);
  
  // vertices
  for (v = 0; v < 4; v++) {
    fprintf (OutFile, INDENT "vertex");
    fprintf_double (vertices[v][0]);
    fprintf_double (vertices[v][1]);
    fprintf_double (vertices[v][2]);
    fprintf (OutFile, "\n");
    if (colored & (1 << v)) {
      fprintf (OutFile, INDENT "color %i %i %i %i\n",
               colors[v][0], colors[v][1], colors[v][2], colors[v][3]);
    }
  }
  
  // visibility
  fprintf (OutFile, INDENT "visible");
  for (v = 0; v < 4; v++) {
    if (visible & (1 << v)) {
      fprintf (OutFile, " 1");
    }
    else {
      fprintf (OutFile, " 0");
    }
  }
  fprintf (OutFile, "\n");

  // don't forget the basics
  if (drivethrough) {
    fprintf (OutFile, INDENT "drivethrough\n");
  }
  if (shootthrough) {
    fprintf (OutFile, INDENT "shootthrough\n");
  }

  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_tele (const char *name, vec3 pos, vec3 size, double rot,
           double border, int drivethrough, int shootthrough)
{
  if (!Output)
    return 0;

  fprintf (OutFile, "teleporter\n");
  print_name (name);
  print_basics (pos, size, rot, drivethrough, shootthrough);
  fprintf (OutFile, INDENT "border");
  fprintf_double (border);
  fprintf (OutFile, "\n");
  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_base (const char *name, vec3 pos, vec3 size, double rot,
           int color, int drivethrough, int shootthrough)
{
  if (!Output)
    return 0;

  fprintf (OutFile, "base\n");
  print_name (name);
  print_basics (pos, size, rot, drivethrough, shootthrough);
  fprintf (OutFile, INDENT "color %i\n", color);
  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_weapon (const char *name, vec3 pos, vec3 size, double rot,
             int drivethrough, int shootthrough, double initdelay,
             double *delays, int delay_count, const char *type)
{
  int i;

  if (!Output)
    return 0;

  fprintf (OutFile, "weapon\n");
  print_name (name);
  print_basics (pos, size, rot, drivethrough, shootthrough);

  if (type != NULL)
    fprintf (OutFile, INDENT "type %s\n", type);

  fprintf (OutFile, INDENT "initdelay");
  fprintf_double (initdelay);
  fprintf (OutFile, "\n");

  fprintf (OutFile, INDENT "delay");
  for (i = 0; i < delay_count; i++) {
    fprintf_double (delays[i]);
  }
  fprintf (OutFile, "\n");

  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_zone (const char *name, vec3 pos, vec3 size, double rot,
           const char *flags, int teams, int safety)
{
  if (!Output)
    return 0;

  fprintf (OutFile, "zone\n");
  print_name (name);
  print_basics (pos, size, rot, 0, 0);

  if (flags != NULL) {
    fprintf (OutFile, INDENT "flag%s\n", flags);
  }

  if (teams != 0) {
    int i;
    fprintf (OutFile, INDENT "team");
    for (i = 0; i < 8; i++) {
      if (teams & (1 << i)) {
        fprintf (OutFile, " %i", i);
      }
    }
    fprintf (OutFile, "\n");
  }

  if (safety != 0) {
    int i;
    fprintf (OutFile, INDENT "safety");
    for (i = 0; i < 8; i++) {
      if (safety & (1 << i)) {
        fprintf (OutFile, " %i", i);
      }
    }
    fprintf (OutFile, "\n");
  }

  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_link (const char *name, int from, int to)
{
  if (!Output)
    return 0;

  fprintf (OutFile, "link\n");
  print_name (name);

  fprintf (OutFile, INDENT "from %i\n", from);
  fprintf (OutFile, INDENT "to   %i\n", to);

  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

int
make_options (const char *options)
{
  if (!Output || (options == NULL))
    return 0;

  fprintf (OutFile, "options\n");
  fprintf (OutFile, "%s", options);
  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/

static void random_color (int indent)
{
  int i;
  for (i = 0; i < indent; i++) {
    fprintf (OutFile, INDENT);
  }
  fprintf (OutFile, "color ");
  fprintf_double ((double)rand() / (double)RAND_MAX);
  fprintf_double ((double)rand() / (double)RAND_MAX);
  fprintf_double ((double)rand() / (double)RAND_MAX);
  fprintf_double (1.0);
  fprintf (OutFile, "\n");
  
  return;    
}

/******************************************************************************/

static void location_color (int indent, vec3 pos)
{
  int i;
  for (i = 0; i < indent; i++) {
    fprintf (OutFile, INDENT);
  }
  fprintf (OutFile, "color ");
  if (pos[2] > 0.0f) {
    fprintf_double (0.0);
    fprintf_double (0.0);
    fprintf_double (1.0);
  } else {
  double factor = (pos[0] + pos[1]) / (4.0 * WorldSize);
    // red is top right
    fprintf_double (factor + 0.5);
    fprintf_double (0.5 - factor);
    fprintf_double (0.0);
  }
  fprintf (OutFile, "\n");
  
  return;    
}

/******************************************************************************/

static void location_color2 (int indent, vec3 pos)
{
  int i;
  for (i = 0; i < indent; i++) {
    fprintf (OutFile, INDENT);
  }
  fprintf (OutFile, "color ");
  fprintf_double ((pos[0] + WorldSize) / (2.0 * WorldSize));
  fprintf_double ((pos[1] + WorldSize) / (2.0 * WorldSize));
  fprintf_double (1.0f - (pos[2] / 80.0));
  fprintf_double (1.0);
  fprintf (OutFile, "\n");
  
  return;    
}

/******************************************************************************/

static int
mesh_box (const char *name, vec3 pos, vec3 size, double rot,
          int drivethrough, int shootthrough)
{
  double angle = rot * (M_PI / 180.0); // radians;
  const double cosV = cos(angle);
  const double sinV = sin(angle);

  vec3 c[8]; // the corners
  c[0][2] = c[1][2] = c[2][2] = c[3][2] = pos[2];
  c[4][2] = c[5][2] = c[6][2] = c[7][2] = pos[2] + size[2];
  
  c[0][0] = c[4][0] = pos[0] + ((cosV * +size[0]) - (sinV * +size[1]));
  c[0][1] = c[4][1] = pos[1] + ((cosV * +size[1]) + (sinV * +size[0]));

  c[1][0] = c[5][0] = pos[0] + ((cosV * -size[0]) - (sinV * +size[1]));
  c[1][1] = c[5][1] = pos[1] + ((cosV * +size[1]) + (sinV * -size[0]));

  c[2][0] = c[6][0] = pos[0] + ((cosV * -size[0]) - (sinV * -size[1]));
  c[2][1] = c[6][1] = pos[1] + ((cosV * -size[1]) + (sinV * -size[0]));

  c[3][0] = c[7][0] = pos[0] + ((cosV * +size[0]) - (sinV * -size[1]));
  c[3][1] = c[7][1] = pos[1] + ((cosV * -size[1]) + (sinV * +size[0]));
  
  fprintf (OutFile, "mesh\n");
  print_name (name);
  if (drivethrough) {
    fprintf (OutFile, INDENT "drivethrough\n");
  }
  if (shootthrough) {
    fprintf (OutFile, INDENT "shootthrough\n");
  }

  fprintf (OutFile, INDENT "inside ");
  fprintf_double (pos[0]);
  fprintf_double (pos[1]);
  fprintf_double (pos[2] + (0.5 * size[2]));
  fprintf (OutFile, INDENT "\n");
  
  int i;
  for (i = 0; i < 8; i++) {
    fprintf (OutFile, INDENT "vertex ");
    fprintf_double (c[i][0]);
    fprintf_double (c[i][1]);
    fprintf_double (c[i][2]);
    fprintf (OutFile, INDENT "\n");
  }
  
  fprintf (OutFile, INDENT "texture roof\n");

//  random_color(1);
  fprintf (OutFile, INDENT "color 0.2 0.2 0.2\n");
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 0 3 2 1\n");
  fprintf (OutFile, INDENT "endface\n");

//  random_color(1);
//  fprintf (OutFile, INDENT "color 0.8 0.8 0.8\n");
  double tmp = (pos[2] / 80.0) + 0.2;
//  tmp = 0.2 + (tmp * ((double)rand() / (double)RAND_MAX));
  fprintf (OutFile, INDENT "color ");
  fprintf_double (tmp);
  fprintf_double (tmp);
  fprintf_double (tmp);
  fprintf (OutFile, "\n");
  
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 4 5 6 7\n");
  fprintf (OutFile, INDENT "endface\n");

  location_color(1, pos);

  fprintf (OutFile, INDENT "texture boxwall\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 0 1 5 4\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 1 2 6 5\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 2 3 7 6\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 3 0 4 7\n");
  fprintf (OutFile, INDENT "endface\n");
  
  fprintf (OutFile, "end\n\n");
  
  return 0;
}                     
                     

/******************************************************************************/

static int
mesh_pyr (const char *name, vec3 pos, vec3 size, double rot,
          int drivethrough, int shootthrough, int flipz)
{
  double angle = rot * (M_PI / 180.0); // radians;
  const double cosV = cos(angle);
  const double sinV = sin(angle);
  
  double sz2 = fabs(size[2]);
  if (flipz) {
    sz2 = -fabs(sz2);
  }

  vec3 c[5]; // the corners
  
  c[0][0] = pos[0] + ((cosV * +size[0]) - (sinV * +size[1]));
  c[0][1] = pos[1] + ((cosV * +size[1]) + (sinV * +size[0]));

  c[1][0] = pos[0] + ((cosV * -size[0]) - (sinV * +size[1]));
  c[1][1] = pos[1] + ((cosV * +size[1]) + (sinV * -size[0]));

  c[2][0] = pos[0] + ((cosV * -size[0]) - (sinV * -size[1]));
  c[2][1] = pos[1] + ((cosV * -size[1]) + (sinV * -size[0]));

  c[3][0] = pos[0] + ((cosV * +size[0]) - (sinV * -size[1]));
  c[3][1] = pos[1] + ((cosV * -size[1]) + (sinV * +size[0]));

  c[0][2] = c[1][2] = c[2][2] = c[3][2] = pos[2];
  
  c[4][0] = pos[0];
  c[4][1] = pos[1];
  c[4][2] = pos[2] + sz2;
  
  fprintf (OutFile, "mesh\n");
  print_name (name);
  if (drivethrough) {
    fprintf (OutFile, INDENT "drivethrough\n");
  }
  if (shootthrough) {
    fprintf (OutFile, INDENT "shootthrough\n");
  }

  fprintf (OutFile, INDENT "inside ");
  fprintf_double (pos[0]);
  fprintf_double (pos[1]);
  fprintf_double (pos[2] + (0.5 * sz2));
  fprintf (OutFile, INDENT "\n");

  int i;
  const int posOrder[5] = {0, 1, 2, 3, 4};
  const int negOrder[5] = {3, 2, 1, 0, 4};
  const int* order = posOrder;
  if (sz2 < 0.0) {
    order = negOrder;
  }
  
  for (i = 0; i < 5; i++) {
    int j = order[i];
    fprintf (OutFile, INDENT "vertex ");
    fprintf_double (c[j][0]);
    fprintf_double (c[j][1]);
    fprintf_double (c[j][2]);
    fprintf (OutFile, INDENT "\n");
  }
  
  fprintf (OutFile, INDENT "texture mesh\n");
//  fprintf (OutFile, INDENT "texture pyrwall\n");
  
  location_color (1, pos);
  
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 0 1 4\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 1 2 4\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 2 3 4\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 3 0 4\n");
  fprintf (OutFile, INDENT "endface\n");
//  random_color(1);
  fprintf (OutFile, INDENT "face\n");
  fprintf (OutFile, INDENT INDENT "vertices 0 3 2 1\n");
  fprintf (OutFile, INDENT "endface\n");
  
  fprintf (OutFile, "end\n\n");

  return 0;
}

/******************************************************************************/
/******************************************************************************/
