/*
 *  U T I L S . C
 *
 *  Helper file for the bzmapper application.
 *
 *  Author:     Dave R      (trepan)
 *  Date:       Feb 2003
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <string.h>
#include <errno.h>

#include "linebuf.h"

#ifndef _WIN32
#  define REALPATH realpath
#else
#  define REALPATH realpath_win
#  include <windows.h>
#  include <ctype.h>
static char *realpath_win (const char *path, char *absolute);
#endif


#include "utils.h"


static int Debug = 0;
static int FirstError = 0;


static int is_rooted (const char *path);


/******************************************************************************/
/******************************************************************************/

static int
is_rooted (const char *path)
{
  if (path == NULL) {
    return 0;
  }

#ifndef _WIN32
  if (strncmp (PATH_SEPARATOR, path, strlen (PATH_SEPARATOR)) == 0) {
    return 1;
  }
#else
  if ((strncmp (PATH_SEPARATOR, path, strlen (PATH_SEPARATOR)) == 0) ||
      ((strlen (path) >= 3) && isalpha (path[0]) && (path[1] == ':')
       && (path[2] == PATH_SEPARATOR[0]))) {
    return 1;
  }
#endif

  return 0;
}

/******************************************************************************/

#ifdef _WIN32
static char *
realpath_win (const char *path, char *absolute)
{
  char *junk;

  if (GetFullPathName (path, MAX_PATH, absolute, &junk) == 0) {
    printf ("Couldn't find file %s\n", path);
    return NULL;
  }
  else if (GetLongPathName (absolute, absolute, MAX_PATH) == 0) {
    printf ("GetLongPathName failed.\n");
    return NULL;
  }

  return absolute;
}
#endif


/******************************************************************************/

/*
 * current  : the full absolute path of the current file, or '.'
 * target   : the filename in the include statement
 * absolute : the full absolute verified path of the target filename
 */

extern int
get_absolute_name (const char *current, const char *target,
                   char **absolute)
{
  char *c;
  char *tmpname;

#ifndef _WIN32
  char path[PATH_MAX];
#else
  char path[MAX_PATH];
#endif


  if ((current[0] == '.') && (current[1] == '\0')) {
    if (REALPATH (target, path) == NULL) {
      perror ("realpath");
      return -1;
    }

    *absolute = strdup (path);
    if (*absolute == NULL) {
      perror ("get_absolute_name(strdup)");
      return -1;
    }
  }
  else if (is_rooted (target)) {
    if (REALPATH (target, path) == NULL) {
      perror ("realpath");
      return -1;
    }

    *absolute = strdup (path);
    if (*absolute == NULL) {
      perror ("get_absolute_name(strdup)");
      return -1;
    }
  }
  else {
    tmpname = strdup (current);
    if (tmpname == NULL) {
      perror ("get_absolute_name(strdup)");
      return -1;
    }

    /*
     * find the basname 
     */
    c = strrchr (tmpname, PATH_SEPARATOR[0]);

    if (*c == PATH_SEPARATOR[0])
      *(c + 1) = '\0';
    else
      return -1;

    tmpname = (char *) realloc (tmpname,
                                sizeof (char) * (strlen (tmpname) +
                                                 strlen (target) + 1));
    if (tmpname == NULL) {
      perror ("get_absolute_name (realloc)");
      return -1;
    }

    strcat (tmpname, target);

    if (REALPATH (tmpname, path) == NULL) {
      perror ("realpath");
      free (tmpname);
      return -1;
    }

    *absolute = strdup (path);
    if (*absolute == NULL) {
      perror ("get_absolute_name(strdup)");
      free (tmpname);
      return -1;
    }

    free (tmpname);
  }

  return 0;
}

/******************************************************************************/

/*
 * current  : the full relative path of the current file, or '.'
 * target   : the filename in the include statement
 * relative : the full relative verified path of the target filename
 */

extern int
get_relative_name (const char *current, const char *target,
                   char **relative)
{
  char *c;


  if ((current[0] == '.') && (current[1] == '\0')) {
    *relative = strdup (target);
    if (*relative == NULL) {
      perror ("get_relative_name(strdup)");
      return -1;
    }
  }
  else if (is_rooted (target)) {
    *relative = strdup (target);
    if (*relative == NULL) {
      perror ("get_relative_name(strdup)");
      return -1;
    }
  }
  else {
    c = strrchr (current, PATH_SEPARATOR[0]);

    if (c == NULL) {
      *relative = strdup (target);
      if (*relative == NULL) {
        perror ("get_relative_name(strdup)");
        return -1;
      }
      else {
        return 0;
      }
    }
    else {
      char *tmpname;

      tmpname = strdup (current);
      if (tmpname == NULL) {
        perror ("get_relative_name(strdup)");
        return -1;
      }

      tmpname[(c - current) + 1] = '\0';    /* clip the basename */
      *relative = (char *) realloc (tmpname, sizeof (char) *
                                    (strlen (tmpname) + strlen (target) +
                                     1));
      if (*relative == NULL) {
        perror ("get_relative_name(malloc)");
        free (tmpname);
        return -1;
      }

      strcat (*relative, target);
    }
  }

  return 0;
}

/******************************************************************************/

static int Seed = 0;

int
my_rand (void)
{
  Seed = Seed * 1103515245 + 12345;
  return (unsigned int) (Seed / 65536) % 32768;
}

/******************************************************************************/

void
my_srand (unsigned int seed)
{
  Seed = seed;
  return;
}

/******************************************************************************/

extern void
set_debug (int debug)
{
  Debug = debug & DEBUG_MASK;
}

/******************************************************************************/

extern int
get_debug (void)
{
  return (Debug & DEBUG_MASK);
}

/******************************************************************************/

extern void
set_debug_objects (void)
{
  Debug = Debug | DEBUG_OBJS_BIT;
  return;
}

/******************************************************************************/

extern void
set_debug_teleporters (void)
{
  Debug = Debug | DEBUG_TELES_BIT;
  return;
}

/******************************************************************************/

extern int
get_debug_objects (void)
{
  return ((Debug & DEBUG_OBJS_BIT) != 0);
}

/******************************************************************************/

extern int
get_debug_teleporters (void)
{
  return ((Debug & DEBUG_TELES_BIT) != 0);
}

/******************************************************************************/

extern void
clear_error (void)
{
  FirstError = 0;
  return;
}

/******************************************************************************/

extern void
error (const char *format, ...)
{
  int linenum;
  va_list arglist;

  if (!FirstError) {
    for (linenum = 0; linenum < 72; linenum++)
      putchar ('-');
    putchar ('\n');

    print_file_stack (current_file ());
    FirstError = 1;
  }

  linenum = current_line_number ();
  if (linenum >= 0) {
    printf ("line (%i) ", linenum);
  }

  va_start (arglist, format);
  vprintf (format, arglist);
  va_end (arglist);

  return;
}

/******************************************************************************/
/******************************************************************************/
