/*
 *  B Z I N D E N T . C
 *
 *  An indenting program for BZFlag maps
 *
 *  Author:  Dave Rodgers
 *  Date:    Oct 23, 2003
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "../utils.h"
#include "../keywords.h"


#define BZINDENT_VERSION    "BZINDENT V0.8"
#define LINE_LENGTH		    (512)
#define INDENT_SPACES	    (2)
#define TEMP_CHAR		    '~'


static char *CurrentFile;

static int indent_file (const char *filename);
static char *out_file_name (const char *basename);

/******************************************************************************/
/******************************************************************************/

int
main (int argc, char **argv)
{
  if ((argc < 2) ||
      (strcmp ("-help", argv[1]) == 0) ||
      (strcmp ("--help", argv[1]) == 0) ||
      (strcmp ("-?", argv[1]) == 0) ||
      (strcmp ("-h", argv[1]) == 0) || (strcmp ("-H", argv[1]) == 0)) {
    printf ("\n%s\n\n", BZINDENT_VERSION);
    printf ("Usage: %s [input-files]\n\n", argv[0]);
    return -1;
  }


  argv++;

  while (*argv != NULL) {
    CurrentFile = *argv;
    indent_file (*argv);
    argv++;
  }

  return 0;
}

/******************************************************************************/

static int
indent_file (const char *filename)
{
  FILE *infile;
  FILE *outfile;
  char *outname = NULL;
  char buffer[LINE_LENGTH];
  const char *c, *dummy;
  int i, keyword;
  int line = 0;
  int err = 0;
  int indent = 0;
  int indent_real = 0;
  int raw_input = 0;

  outname = out_file_name (filename);
  if (outname == NULL) {
    return -1;
  }

  infile = fopen (filename, "r");

  if (infile == NULL) {
    fprintf (stderr, "fopen (\"%s\"): ", filename);
    perror ("");
    free (outname);
    return -1;
  }

  outfile = fopen (outname, "w");

  if (outfile == NULL) {
    fprintf (stderr, "fopen (\"%s\"): ", outname);
    perror ("");
    free (outname);
    return -1;
  }


  while (!feof (infile)) {

    if (fgets (buffer, LINE_LENGTH, infile) == NULL) { 
      break;
    }

    line++;

    c = buffer;
    while ((*c != '\0') && isspace (*c)) {
      c++;
    }

    if ((*c == '\0') || (*c == '#')) {
      fputs (buffer, outfile);
    }
    else {
      if (!raw_input) {
        keyword = keyword_parse (buffer, LINE_LENGTH, &dummy);
      }
      else {
        if (keyword_is_end (buffer)) {
          keyword = end_key;
          raw_input = 0;
        }
        else {
          keyword = var_key; // fake value that won't affect indenting
        }
      }

      if (keyword < 0) {
        err = 1;
      }

      if (keyword == end_key) {
        indent--;
        indent_real--;
        if (indent < 0) {
          printf ("Extra end at line %i\n", line);
          indent = 0;
          err = 1;
        }
      }
      else if (keyword == else_key) {
        indent--;
        if (indent < 0) {
          printf ("Early 'else' command at line %i\n", line);
          indent = 0;
          err = 1;
        }
      }


      for (i = 0; i < (indent * INDENT_SPACES); i++) {
        fputc (' ', outfile);
      }

      if (keyword == else_key) {
        indent++;
      }


      if (keyword_ended (keyword)) {
        indent++;
        indent_real++;
      }
      
      if (keyword_raw_input (keyword)) {
        raw_input = 1;
      }

      fputs (c, outfile);
    }
  }

  fclose (infile);
  fclose (outfile);

  if (indent_real > 0) {
    printf ("%s: Missing %i 'end' keywords.\n", filename, indent_real);
    err = 1;
  }
  else if (indent_real < 0) {
    printf ("%s: %i extra 'end' keywords.\n", filename, -indent_real);
    err = 1;
  }

  if (!err) {
    if (rename (outname, filename) == -1) {
      perror ("rename");
    }
  }

  free (outname);

  if (err) {
    return -1;
  } else {
    return 0;
  }
}

/******************************************************************************/

static char *
out_file_name (const char *basename)
{
  int len;
  char *c;

  len = strlen (basename);

  c = (char *) malloc ((size_t) len + 2);
  if (c == NULL) {
    perror ("tmpname(malloc)");
    return NULL;
  }

  strcpy (c, basename);
  c[len] = TEMP_CHAR;
  c[len + 1] = '\0';

  return c;
}

/******************************************************************************/

/*
 * These functions are required by the keyparse.c file
 */

extern int
get_debug (void)
{
  return 0;
}

extern void
error (const char *format, ...)
{
  va_list ap;

  printf ("[%s] ", CurrentFile);
  va_start (ap, format);
  vprintf (format, ap);
  va_end (ap);

  return;
}

/******************************************************************************/
/******************************************************************************/
